//
// Assert.h - vOS asserts
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vOS_Assert_h
#define vOS_Assert_h

#include <cassert>

namespace vos
{
	/*!
		\def ASSERT(condition)

		Usual assert functionality - if \a condition is false, then
		in debug builds the program will complain to the user.

		\def vosERROR()

		Simple way of reporting a debug problem in the library - currently
		evaluates to ASSERT(false).
	*/
#ifdef _DEBUG
	// Complain in debug builds
	#ifndef ASSERT
		#define ASSERT(cond) assert(cond)
	#endif
	#define vosERROR() ASSERT(false)
#else
	// Ignore in release builds
	#ifndef ASSERT
		#define ASSERT(cond)
	#endif
	#define vosERROR()
#endif

	/*!
		\def VERIFY(statement)
		
		Macro evaluates the \a statement always - in debug builds it
		will call ASSERT, so that if \a statement returns false, the
		program will complain.
		
	    Use instead of ASSERT when you want the statement to always
		be evaluated, even in non-debug builds.
	*/
#ifdef _DEBUG
	// In debug builds, execute the statement, and assert that it succeeds.
	#ifndef VERIFY
		#define VERIFY(stmnt) ASSERT(stmnt)
	#endif
#else
	// In release builds, we still execute the statement, but ignore the return code.
	#ifndef VERIFY
		#define VERIFY(stmnt) stmnt
	#endif
#endif

};

#endif // !vOS_Assert_h
