//
// Date.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_Date_h
#define vos_Date_h

#include "vos.h"

namespace vos
{
	class String;

	/*!
		\enum DayName

		DayName specifies a day of the week (e.g. Monday, Tuesday, etc)
	*/
	typedef enum
	{
		Sunday = 0,
		Monday,
		Tuesday,
		Wednesday,
		Thursday,
		Friday,
		Saturday
	}
	DayName;

	/*!
		\enum MonthName

		MonthName specifies a name of a month (e.g. January, February, etc)
	*/
	typedef enum
	{
		January = 0,
		February,
		March,
		April,
		May,
		June,
		July,
		August,
		September,
		October,
		November,
		December
	}
	MonthName;

	/**************************************************************************
	
	>	class Date
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Encapsulate a date, that is, a particular day, such
					as 4th July 1978.

					The class does not include any information about time
					of day - use vos::Time for Date+Time in one object,
					or use vos::TimeOfDay if you don't care about the
					date.
	
	**************************************************************************/
	
	class Date
	{
		public:
			/// Constructor
			Date();

			/// Copy constructor
			Date(const Date& rhs);

			/// Assignment operator
			Date& operator=(const Date& rhs);

			/// Set a date object to be a particular date
			bool Set(int DayOfMonth,
					 MonthName Month,
					 int Year);

			/// Set a date object to be a particular date
			bool Set(int DayOfMonth,	// 1-31
					 int Month,			// 1-12
					 int Year);			// 1980 for 1980, not just 80

			// Check to see if a given date is valid
			static bool IsValid(int Year, int Month, int Day);

		public:
			//
			// Date arithmetic
			//

			// Add a number of days onto a Date
			void AddDays(int nDays);

			Date& operator+=(int nDays);

			Date& operator-=(int nDays);

			//
			// Date comparison
			//
			bool operator==(const Date& rhs);
			bool operator!=(const Date& rhs);
			bool operator>(const Date& rhs);
			bool operator>=(const Date& rhs);
			bool operator<=(const Date& rhs);
			bool operator<(const Date& rhs);

		public:
			//
			// Date component access
			//
			// Get the day of the month from this date (1-31)
			int GetDay() const { return m_Day; }

			// Get the month name from this date
			MonthName GetMonth() const { return m_Month; }

			// Get the full year from this date (not just the last two digits)
			int GetYear() const { return m_Year; }

			/// Get cardinal index of this date's month; Jan = 1, Dec = 12
			int GetCardinalMonth() const { return ((int) (m_Month - January)) + 1; }

			/// Test to see if date falls within a leap year
			bool IsLeapYear();

			/// Get the number of days in the month in which the date falls
			int GetDaysInMonth();

			/// Get today's date
			static Date GetToday();

			/// Get deterministic string representation - unaffected by locales etc.
			String MakeString();

		private:
			// The date
			int m_Day;
			MonthName m_Month;
			int m_Year;
	};
};

#endif // !vos_Date_h
