//
// ErrorManager.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Errors.h"

#include "ErrorManager.h"

#include "FileSpec.h"

#include "Debugger.h"

#if VOS_ERROR_TYPESAFE
	static vos::Error RaiseErrorObject(vos::Error::Error_RaiseError);
	static vos::Error NoErrorObject(vos::Error::Error_NoError);
#else
	const bool RaiseErrorObject	= false;
	const bool NoErrorObject	= true;
#endif

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/**************************************************************************
							 
	>	ErrorInfo::ErrorInfo(int Code, 
							 TErrorType Type,
							 char* szSourceFile,
							 int LineNumber)

		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		Code - the code of the error
					Type - type of the error
					szSourceFile - name of the source file that registered
								   the error.
					LineNumber - line number in the source file where the
								 error was registered.

		Purpose:	Constructor

	**************************************************************************/

	ErrorInfo::ErrorInfo(int Code, 
						 TErrorType Type,
						 char* szSourceFile,
						 int LineNumber) : m_Code(Code),
										   m_Type(Type),
										   m_SourceLine(LineNumber)
	{
		if (szSourceFile != NULL)
		{
#ifndef VOS_MACOS
			// Convert filename to leafname
			FileSpec SourceFileSpec;
			SourceFileSpec.SetFromString(szSourceFile);
			m_SourceFile = SourceFileSpec.GetLeafname();
#else
			m_SourceFile.assign(szSourceFile);
#endif
		}
	}

	/**************************************************************************

	>	ErrorInfo::ErrorInfo(int Code, 
							 TErrorType Type, 
							 const String& Text,
							 char* szSourceFile,
							 int LineNumber)

		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		Code - the code of the error
					Type - type of the error
					Text - the textual error message describing the error
					szSourceFile - name of the source file that registered
								   the error.
					LineNumber - line number in the source file where the
								 error was registered.

		Purpose:	Constructor

	**************************************************************************/

	ErrorInfo::ErrorInfo(int Code, 
						 TErrorType Type, 
						 const String& Text,
						 char* szSourceFile,
						 int LineNumber) : m_Code(Code),
										   m_Type(Type),
										   m_Text(Text),
										   m_SourceLine(LineNumber)
	{
		if (szSourceFile != NULL)
		{
#ifndef VOS_MACOS
			// Convert filename to leafname
			FileSpec SourceFileSpec;
			SourceFileSpec.SetFromString(szSourceFile);
			m_SourceFile = SourceFileSpec.GetLeafname();
#else
			m_SourceFile.assign(szSourceFile);
#endif
		}
	}

	/**************************************************************************
	
	>	bool ErrorInfo::GetPrevError(const ErrorInfo** ppErrorInfo) const
	
		Author:		Tim Browse
		Date:		13/4/2002
		Outputs:	ppErrorInfo - the previous error
		Returns:	true => the previous error is in ppErrorInfo
					false => there is no previous error
	
		Purpose:	Get the error that was reported before this one.
	
	**************************************************************************/
	
	bool ErrorInfo::GetPrevError(const ErrorInfo** ppErrorInfo) const
	{
		// NB direction is inverted as the most recent error is at the head of the list.
		ErrorInfo *pPreviousError = (ErrorInfo *) GetNext();

		if ((pPreviousError != NULL) && (ppErrorInfo != NULL))
		{
			*ppErrorInfo = pPreviousError;
		}

		return (pPreviousError != NULL);
	}

	/**************************************************************************
	
	>	bool ErrorInfo::GetNextError(const ErrorInfo** ppErrorInfo) const
	
		Author:		Tim Browse
		Date:		13/4/2002
		Outputs:	ppErrorInfo - the next error
		Returns:	true => the next error is in ppErrorInfo
					false => there is no next error
	
		Purpose:	Get the error that was reported after this one.
					(NB. You probably want to use GetPrevError())
	
	**************************************************************************/
	
	bool ErrorInfo::GetNextError(const ErrorInfo** ppErrorInfo) const
	{
		// NB direction is inverted as the most recent error is at the head of the list.
		ErrorInfo *pNextError = (ErrorInfo *) GetPrev();

		if ((pNextError != NULL) && (ppErrorInfo != NULL))
		{
			*ppErrorInfo = pNextError;
		}

		return (pNextError != NULL);
	}

	/**************************************************************************
	
	>	ErrorStack::~ErrorStack()
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Destructor (deletes all ErrorInfo objects on the stack)
	
	**************************************************************************/
	
	ErrorStack::~ErrorStack()
	{
		DeleteAll();
	}

	/**************************************************************************
	
	>	void ErrorStack::Push(ErrorInfo* pInfo)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		pInfo - the error to push onto the stack 
							(see purpose for details)
	
		Purpose:	Push an error onto the error stack.

					NB. The ErrorInfo object is now owned by the ErrorStack
						object; a copy of it is not used - the actual object
						passed in is used by the stack.
	
	**************************************************************************/
	
	void ErrorStack::Push(ErrorInfo* pInfo)
	{
		AddHead(pInfo);
	}

	/**************************************************************************
	
	>	ErrorInfo* ErrorStack::Pop()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	The ErrorInfo object popped by the stack.
					(or NULL if the stack is empty - this is considered an
					 logic error, so a COMPLAIN error will be issued - use
					 ErrorsOutstanding() to avoid this).
	
		Purpose:	Pop an ErrorInfo object from the stack.

					NB. The caller is now responsible for disposing of the
						returned ErrorInfo object.
	
	**************************************************************************/
	
	ErrorInfo* ErrorStack::Pop()
	{
		ErrorInfo* pInfo = (ErrorInfo*) GetHead();
		if (pInfo != NULL)
		{
			Remove(pInfo);
		}
		else
		{
			COMPLAIN("ErrorStack: Pop() called on empty stack.");
		}

		return pInfo;
	}

	/**************************************************************************
	
	>	ErrorInfo* ErrorStack::Peek()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	The top error on the stack (or NULL if stack is empty).
	
		Purpose:	Get the top error on the stack, but do not remove it.
	
	**************************************************************************/
	
	ErrorInfo* ErrorStack::Peek()
	{
		return (ErrorInfo*) GetHead();
	}


	/**************************************************************************
	
	>	bool ErrorStack::ErrorsOutstanding()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	true => one or more errors are outstanding
					false => no errors (Stack is empty)
	
		Purpose:	Find out if there are any errors outstanding on the
					error stack.
	
	**************************************************************************/
	
	bool ErrorStack::ErrorsOutstanding()
	{
		return (Length() > 0);
	}

	/**************************************************************************
	
	>	void ErrorStack::Clear()
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Remove (and delete) all errors from the stack.
	
	**************************************************************************/
	
	void ErrorStack::Clear()
	{
		DeleteAll();
	}


	/**************************************************************************
	
	>	void ErrorStateStack::Push(ErrorStack* pNewState)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		pNewState - the error stack to push onto this state stack.
								(NB. This object is then owned by the stack
									 until it is removed via Pop()).
	
		Purpose:	Push a complete error stack onto this state stack.
	
	**************************************************************************/
	
	void ErrorStateStack::Push(ErrorStack* pNewState)
	{
		AddHead(pNewState);
	}

	/**************************************************************************
	
	>	ErrorStack* ErrorStateStack::Pop()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	Pointer to the ErrorStack object that has been removed
					or NULL if the stack is empty (which also generates a
					COMPLAIN error).

					NB. The caller is responsible for disposing of this object.
	
		Purpose:	Pop an error stack from the stack.
	
	**************************************************************************/
	
	ErrorStack* ErrorStateStack::Pop()
	{
		ErrorStack *pState = (ErrorStack*) GetHead();
		if (pState != NULL)
		{
			Remove(pState);
		}
		else
		{
			COMPLAIN("ErrorStateStack: Pop() called on empty stack!");
		}

		return pState;
	}


	//
	/// The error object returned by GetLastError() when there are no errors outstanding.
	//
	ErrorInfo ErrorManager::s_NoError(0, Error_None);

	/**************************************************************************
	
	>	ErrorManager::ErrorManager() 
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	ErrorManager::ErrorManager() : m_pReporter(NULL),
								   m_pCurrentErrors(NULL)
	{
		// Create a new errors list.
		m_pCurrentErrors = new ErrorStack;
	}

	/**************************************************************************
	
	>	ErrorManager::~ErrorManager()
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	ErrorManager::~ErrorManager()
	{
		// Auto-delete the installed error reporter
		if (m_pReporter!=NULL)
		{
			delete m_pReporter;
			m_pReporter = NULL;
		}

		// Delete outstanding errors
		delete m_pCurrentErrors;
	}

	// Register a new error that has just occurred.
	/**************************************************************************

	>	Error ErrorManager::RegisterError(int Code, 
										  TErrorType Type, 
										  const char *szReason,
										  char* szSourceFile,
										  int LineNumber)

		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		Code - the error code of the error being registered
					Type - the error type of the error being registered
					szReason - Textual error message describing the error
					szSourceFile - the source file that is registering the error
					LineNumber - the source line where the error is registered
		Returns:	Error (so you can pass it back to your caller)

		Purpose:	Register an error with the error manager.

	**************************************************************************/

	Error ErrorManager::RegisterError(int Code, 
									  TErrorType Type, 
									  const char *szReason,
									  char* szSourceFile,
									  int LineNumber)
	{
		// Make a new ErrorInfo object for this error.
		ErrorInfo* pNewError = new ErrorInfo(Code,
											 Type,
											 szReason,
											 szSourceFile,
											 LineNumber);

		// Add this error to the error stack
		m_pCurrentErrors->Push(pNewError);

		return RaiseErrorObject;
	}

	//
	/// s_bReportingError is used to prevent re-entrant calls to error manager.
	//
	static bool s_bReportingError = false;

	/**************************************************************************
	
	>	void ErrorManager::ReportError()
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Report an error to the user, using the installed error
					handler.

					NB. If you have not installed an error handler, then
						reporting an error has no effect in release builds.
						(In debug builds, the debugger is invoked)
	
	**************************************************************************/
	
	void ErrorManager::ReportError()
	{
		if (s_bReportingError)
		{
			// TODO: use trace output to record error
#ifdef VOS_MSWINDOWS
			// If you hit this line, then an error has occurred while reporting an error.
			_asm int 3;
#elif (defined VOS_MACOS)
			Debugger();
#else
			#error ErrorManager::ReportError Needs porting
#endif
			return;
		}

		if (!m_pCurrentErrors->ErrorsOutstanding())
		{
			// Error - trying to report an error when no errors
			// are outstanding.
			RegisterError(0, Error_Normal, 
						  "ReportError() called when no errors have been encountered.");
			ReportError();
			return;
		}

		if (m_pReporter != NULL)
		{
			s_bReportingError = true;

			// Remove the current error stack, and get the currently
			// installed error reporter to report the error.
			//
			// NB. We remove the errors from the stack before reporting,
			//     because the error reporter may be persistent (e.g. a
			//     dialog that the user must acknowledge), and async events
			//     may cause interference (e.g. a function may call NoError()
			//     while the dialog is still visible, which will cause an
			//     assert to go off incorrectly).
			//

			// Ok, remove the current error stack, and make a fresh one.
			ErrorStack* pCurrentErrors = m_pCurrentErrors;
			m_pCurrentErrors = new ErrorStack;

			// Report this error state
			m_pReporter->ReportError(pCurrentErrors);

			// Delete the error stack - it is no longer needed
			delete pCurrentErrors;

			s_bReportingError = false;
		}
		else
		{
			// If debugging, then always drop into debugger when an error
			// is reported with no installed error handler.
			ClearErrors();		// This is considered "reported" now

#ifdef _DEBUG
			Debugger::Interrupt();
#endif

		}
	}

	/**************************************************************************
	
	>	void ErrorManager::ClearErrors()
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Clear all errors - used when the caller has recovered from
					the outstanding errors.
	
	**************************************************************************/
	
	void ErrorManager::ClearErrors()
	{
		// Delete all outstanding errors
		m_pCurrentErrors->Clear();
	}

	// Get details of last error reported.

	/**************************************************************************
	
	>	const ErrorInfo* ErrorManager::GetLastError() const
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	The info about the last registered error.
					If no errors are outstanding, a null error object (with an
					error code of 0 and type of Error_None) is returned.
	
		Purpose:	Get information about the last error registered with the
					ErrorManager.
	
	**************************************************************************/

	const ErrorInfo* ErrorManager::GetLastError() const
	{
		if (m_pCurrentErrors->ErrorsOutstanding())
		{
			return m_pCurrentErrors->Peek();
		}
		else
		{
			// No error - return null error object
			return &s_NoError;
		}
	}

	
	/**************************************************************************
	
	>	Error ErrorManager::LastError() const
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	Error that can be returned to caller of the caller.
	
		Purpose:	Get the last error registered, so that it can be returned
					to the caller's caller.

					e.g. when you encounter a function that returns an error,
						 you can just return the error like this:

						  if (!MyFunction())
							  return ErrorManager::LastError();

						 NB. It is recommended that you use the LASTERROR macro
							 from the vOS\Errors.h include file, rather than
							 call this function directly.
	
	**************************************************************************/
	
	Error ErrorManager::LastError() const
	{
		if (m_pCurrentErrors->ErrorsOutstanding())
		{
			return RaiseErrorObject;
		}
		else
		{
			return NoErrorObject;
		}
	}

	/**************************************************************************
	
	>	Error ErrorManager::NoError() const
	
		Author:		Tim Browse
		Date:		14/4/2002
		Returns:	Error value signifying no error occurred.
	
		Purpose:	Used to provide an Error value that signifies that no
					error occurred.  Used by any function that returns an
					Error value, and wants to tell the caller that the
					function completely successfully.

					It is strongly recommended that you use the VNOERROR macro
					from vOS\Errors.h rather than calling this function
					directly (as this practice will make release builds 
					much more efficient).
  
					NB. In debug builds, if this function is called when
						errors are outstanding, a COMPLAIN error is issued.
						This is so that when a function ignores the error
						return value from another function it calls, the
						program will COMPLAIN when that function actually
						returns an error.

						Basically, the logic is: you're saying no error 
						occurred, but you just called an error that
						returned an error that you did not handle (by
						using the ERROR_HANDLED macro).  This is an 
						error - you should handle that error properly.

	**************************************************************************/
	
	Error ErrorManager::NoError() const
	{
		// If there are no outstanding errors, or we are currently
		// reporting the errors, return NoError.
		if (!m_pCurrentErrors->ErrorsOutstanding() || s_bReportingError)
		{
			return NoErrorObject;
		}

		// There are outstanding errors which have not been cleared.
		// (see function comment above for details)
		COMPLAIN("A function has attempted to return the VNOERROR value after an "
				 "error has been reported but not cleared.  The function should "
				 "either return the error using LASTERROR, or notify the Error "
				 "Manager that it has been handled with ERROR_HANDLED().");

		return LastError();
	}

	/**************************************************************************
	
	>	void ErrorManager::PushState()
	
		Author:		Tim Browse
		Date:		14/4/2002
	
		Purpose:	Push the current error state (stack of errors) onto
					a stack, and make the current error state empty (of 
					errors).  Use this (along wioth PopState) in error handling 
					code which might otherwise corrupt the error state.
	
	**************************************************************************/
	
	void ErrorManager::PushState()
	{
		// Push the current state onto the state stack
		m_StateStack.Push(m_pCurrentErrors);

		// Make a new state
		m_pCurrentErrors = new ErrorStack;
	}

	/**************************************************************************
	
	>	void ErrorManager::PopState()
	
		Author:		Tim Browse
		Date:		14/4/2002
	
		Purpose:	Pop the last error state from the stack (see PushState).
	
	**************************************************************************/
	
	void ErrorManager::PopState()
	{
		// Delete the current state
		delete m_pCurrentErrors;

		// Pop the last state and install it as the current state
		m_pCurrentErrors = m_StateStack.Pop();

		// Check for errors
		if (m_pCurrentErrors == NULL)
		{
			COMPLAIN("ErrorManager: Unbalanced calls to PushState() and PopState()!");

			// Recover by creating a new state
			m_pCurrentErrors = new ErrorStack;
		}
	}

	/**************************************************************************
	
	>	void ErrorManager::SetErrorReporter(ErrorReporter *pReporter)
	
		Author:		Tim Browse
		Date:		14/4/2002
		Inputs:		pReporter - the error reporter to use
	
		Purpose:	Install an error reporter to be called when ReportError()
					is called.

					By default, no error reporter is installed.  When ReportError()
					is called with no error reporter available, in debug builds the
					debugger is invoked, and in release builds, the errors
					are silently discarded.  It is therefore recommended that
					you install an appropriate error reporter.

					NB. The currently installed error reporter will be
						deleted on program shutdown (so you don't have to).
	
	**************************************************************************/
	
	void ErrorManager::SetErrorReporter(ErrorReporter *pReporter)
	{
		if (m_pReporter)
		{
			delete m_pReporter;
			m_pReporter=NULL;
		}
		// Install the error reporter
		m_pReporter = pReporter;
	}

	/**************************************************************************
	
	>	ErrorManager& TheErrorManager()
	
		Author:		Tim Browse
		Date:		14/4/2002
	
		Purpose:	Get the singleton ErrorManager object.
					
					It is recommended that you use the ERRMGR() macro
					from vOS\Errors.h, rather than call this function
					directly.  (It will allow dynamic libraries to use 
					the same error manager as the host program).
	
	**************************************************************************/
	
	ErrorManager& TheErrorManager()
	{
		static ErrorManager TheErrorManager;

		return TheErrorManager;
	}

	/**************************************************************************
	
	>	const ErrorInfo* GetLastError()
	
		Author:		Tim Browse
		Date:		14/4/2002
		Returns:	ErrorInfo object representing the last registered error
					(you do not own this object - do NOT delete it)
	
		Purpose:	Get the ErrorInfo object for the last error that was
					registered - use this in error handling code.
	
	**************************************************************************/
	
	const ErrorInfo* GetLastError()
	{
		return TheErrorManager().GetLastError();
	}

	/**************************************************************************
	
	>	int GetLastErrorCode()
	
		Author:		Tim Browse
		Date:		14/4/2002
		Returns:	Error code for last error registered
	
		Purpose:	Get the error code for the last error that was registered.

					Use this function in error handling code.

	**************************************************************************/
	
	int GetLastErrorCode()
	{
		return TheErrorManager().GetLastError()->GetCode();
	}

	/**************************************************************************
	
	>	TErrorType GetLastErrorType()
	
		Author:		Tim Browse
		Date:		14/4/2002
		Returns:	Type of the error; one of:
						Error_None 
						Error_Complaint
						Error_Normal
						Error_Critical
						Error_Terminal
	
		Purpose:	Get the type of the last error that was registered.
	
					Use this function in error handling code.

	**************************************************************************/
	
	TErrorType GetLastErrorType()
	{
		return TheErrorManager().GetLastError()->GetType();
	}

};


#ifdef VOS_LOG_ERRORS

	/**************************************************************************
	
	>	bool LogReport(char *Report, int Line, char *File, bool ReturnCode)
	
		Author:		Tim Browse
		Date:		14/4/2002
		Inputs:		Report - the error message
					Line - line number where the error occurred
					File - source file in which the error occurred
					ReturnCode - value to return from this function
		Returns:	Same value as ReturnCode
	
		Purpose:	Log a vOS error to an error log file.
	
	**************************************************************************/
	
	bool LogReport(char *Report, int Line, char *File, bool ReturnCode)
	{
		FILE *Log = fopen("c:\\vOSErrorLog.txt", "a");

		if (Log != NULL)
		{
			fprintf(Log, "%s from line %d, %s\n", Report, Line, File);
			fclose(Log);
		}

		return(ReturnCode);
	}

#endif
