//
// FileInAFile.cpp - a class to support a file within another file
//

#include "StdHdr.h"

#include "FileInAFile.h"

#include "Errors.h"
#include "Msgs.h"

//
// Macros for CR/LF handling
//
#define CR 0x0D
#define LF 0x0A


namespace vos
{
	
	/**************************************************************************

	>	FileInAFile::FileInAFile()

		Author:		Tim Browse
		Date:		02/06/2001

		Purpose:	Constructor

	**************************************************************************/

	FileInAFile::FileInAFile() : m_FileStart(0),
								 m_FileLength(0),
								 m_pParentFile(NULL),
								 m_Modes((vos::FileMode) 0)
	{
	}

	
	/**************************************************************************

	>	FileInAFile::~FileInAFile()

		Author:		Tim Browse
		Date:		02/06/2001

		Purpose:	Destructor

	**************************************************************************/

	FileInAFile::~FileInAFile()
	{
		delete m_pParentFile;
		m_pParentFile = NULL;
	}

	
	/**************************************************************************

	>	void FileInAFile::Create(vos::File* pParentFile, uint FileStart, int FileLength,
								 vos::FileMode ModesPermitted)

		Author:		Tim Browse
		Date:		02/06/2001
		Inputs:		pParentFile - the file we are using the data from to create
								  the new subfile.  This file object should NOT
								  currently be 'open' as this object will open
								  it when Open() is called on this object.

					FileStart - byte offset of the start of the subfile's data
								(in the file specified by pParentFile).

					FileLength - length (in bytes) of the file data to use.

					ModesPermitted - access mode permitted to this underlying
									 file data.  At present, only read ops are
									 allowed.

		Purpose:	Initialise the FileInAFile object, so that Open(), Read()
					etc can subsequently be called on it.  You *must* call this
					function before trying any vos::File operation on this object.

	**************************************************************************/

	void FileInAFile::Create(vos::File* pParentFile, uint FileStart, int FileLength,
							 vos::FileMode ModesPermitted)
	{
		// Check parameters
		if (pParentFile == NULL)
		{
			COMPLAIN("FileInAFile::Create() - pParentFile is NULL");
			return;
		}

		// Remember file details
		m_pParentFile = pParentFile;
		m_FileStart = FileStart;
		m_FileLength = FileLength;
		m_Modes = ModesPermitted & (vos::fam_Read | vos::fam_Write);
	}

	
	/**************************************************************************

	>	vos::Error FileInAFile::Open(vos::FileMode Mode)

		Author:		Tim Browse
		Date:		02/06/2001
		Inputs:		Mode - the file access mode to use when opening the file.
		Returns:	Error if the file cannot be opened, or if the requested
						  access permission is not available.

		Purpose:	Open the FileInAFile, ready for action.

	**************************************************************************/

	vos::Error FileInAFile::Open(vos::FileMode Mode)
	{
		// Check file access permissions
		if ((Mode & m_Modes) != Mode)
		{
			// Requested permission was not granted by creator of this object.
			return ERROR(vos::Err_File_AccessDenied, msgFileInAFile_OpenFailedAccess);
		}

		// Ok, try to open
		if (m_pParentFile->Open(Mode))
		{
			// Seek to start of file
			if (m_pParentFile->SeekFromStart(m_FileStart))
			{
				// Opened and positioned successfully
				return VNOERROR;
			}
			else
			{
				m_pParentFile->Close();
			}
		}

		// Error occurred
		return LASTERROR;
	}

	
	/**************************************************************************

	>	void FileInAFile::Close()

		Author:		Tim Browse
		Date:		02/06/2001

		Purpose:	Close the file object.

	**************************************************************************/

	void FileInAFile::Close()
	{
		if (m_pParentFile != NULL)
		{
			m_pParentFile->Close();
		}
		else
		{
			COMPLAIN("FileInAFile::Close() called on NULL parent file.");
		}
	}


	
	/**************************************************************************

	>	bool FileInAFile::IsOpen() const

		Author:		Tim Browse
		Date:		02/06/2001
		Returns:	true => file is open
					false => file is closed

		Purpose:	Find out if the file is currently open.

	**************************************************************************/

	bool FileInAFile::IsOpen() const
	{
		if (m_pParentFile != NULL)
		{
			return m_pParentFile->IsOpen();
		}
		else
		{
			COMPLAIN("FileInAFile::IsOpen() called on NULL parent file.");
			return false;
		}
	}

	
	/**************************************************************************

	>	vos::Error FileInAFile::GetSize(uint* pSize)

		Author:		Tim Browse
		Date:		02/06/2001
		Outputs:	pSize - size of the file, in bytes.
		Returns:	Error - e.g. file not open or initialised.

		Purpose:	Determine the size of the file, in bytes.

	**************************************************************************/

	vos::Error FileInAFile::GetSize(uint* pSize)
	{
		if (m_pParentFile != NULL)
		{
			*pSize = m_FileLength;
			return VNOERROR;
		}
		else
		{
			COMPLAIN("FileInAFile::GetSize() called on NULL parent file.");
			return ERROR(vos::Err_File_GetPosFail, msgFile_GetPos_Fail);
		}
	}

	
	/**************************************************************************

	>	vos::Error FileInAFile::Read(void* pData, uint cBytes, uint* pcBytes)

		Author:		Tim Browse
		Date:		02/06/2001
		Inputs:		cBytes - number of bytes to read.
		Outputs:	pData - the bytes read from the file
					pcBytes - the number of bytes actually read (may be NULL)
		Returns:	Error - e.g. eof, file not open, etc.

		Purpose:	Read a number of bytes from a file.

					If there are not enough bytes in the file to satisfy the
					read request, as many bytes as possible are read, pcBytes
					is updated (if non-NULL), and vos::Err_File_EOF is
					returned.

	**************************************************************************/

	vos::Error FileInAFile::Read(void* pData, uint cBytes, uint* pcBytes)
	{
		if (m_pParentFile != NULL)
		{
			// Check range of read operation
			uint Pos;
			if (GetPos(&Pos))
			{
				// Can we read the requested number of bytes from the file?
				int cBytesLeft = m_FileLength - Pos;

				if (cBytes > cBytesLeft)
				{
					// No, so read as much as we can, then return an error
					if (m_pParentFile->Read(pData, cBytesLeft, pcBytes))
					{
						// Fake the error, as Read() worked as we expected, but they
						// still asked for too much data.
						return ERROR(vos::Err_File_EOF, msgFile_Read_EOF);
					}
					else
					{
						// Could not even read what we wanted, so just pass
						// on the error.
						return LASTERROR;
					}
				}
				else
				{
					// Nominal case...
					//
					// Read operation meets all our limits, so it's a simple 
					// pass-through
					return m_pParentFile->Read(pData, cBytes, pcBytes);
				}
			}
		}

		// If we get here, then a problem occured reading the data
		return ERROR(vos::Err_File_ReadFail, msgFile_Read_Fail);
	}

	
	/**************************************************************************

	>	vos::Error FileInAFile::Write(const void* pData, uint cBytes, uint* pcBytes)

		Author:		Tim Browse
		Date:		02/06/2001

		Purpose:	Write a number of bytes to the file.  
		
					NB. Write operations are currently NOT supported by this class.

	**************************************************************************/

	vos::Error FileInAFile::Write(const void* pData, uint cBytes, uint* pcBytes)
	{
		COMPLAIN("Writing to FileInAFile objects not yet supported.");
		return ERROR(vos::Err_File_WriteFail, "Writing to FileInAFile objects not yet supported.");
	}

	/**************************************************************************

	>	vos::Error FileInAFile::SeekFromStart(uint Offset)

		Author:		Tim Browse
		Date:		2/6/01
		Inputs:		Offset - the number of bytes from the start of the file
							 to position the file read/write pointer at.
		Returns:	Standard error

		Purpose:	Seek to a position in the file specified as a number of 
					bytes from the start of the file.

	**************************************************************************/

	vos::Error FileInAFile::SeekFromStart(uint Offset)
	{
		if (m_pParentFile != NULL)
		{
			// Check validity of seek request...

			// Note special case - we always allow seeking to byte 0, 
			// even in a zero-length file.  Otherwise, the seek must be to 
			// a byte within our file range.
			if ((Offset == 0) || (Offset < m_FileLength))
			{
				return m_pParentFile->SeekFromStart(m_FileStart + Offset);
			}
			else
			{
				// Seek requested is outside limits
				return ERROR(vos::Err_File_SeekFail, msgFile_Access_Fail);
			}
		}
		else
		{
			return ERROR(vos::Err_File_NotOpen, msgFile_Access_Fail);
		}
	}

	/**************************************************************************

	>	vos::Error FileInAFile::SeekFromEnd(uint Offset)

		Author:		Tim Browse
		Date:		2/6/01
		Inputs:		Offset - the number of bytes from the end of the file
							 to position the file read/write pointer at.
		Returns:	Standard error

		Purpose:	Seek to a position in the file specified as a number of 
					bytes from the end of the file.

	**************************************************************************/

	vos::Error FileInAFile::SeekFromEnd(uint Offset)
	{
		if (m_pParentFile != NULL)
		{
			// Check validity of seek request...
			if (Offset < m_FileLength)
			{
				return m_pParentFile->SeekFromStart(m_FileStart + (m_FileLength - (Offset + 1)));
			}
			else
			{
				// Seek requested is outside limits
				return ERROR(vos::Err_File_SeekFail, msgFile_Access_Fail);
			}
		}
		else
		{
			return ERROR(vos::Err_File_NotOpen, msgFile_Access_Fail);
		}
	}

	/**************************************************************************

	>	vos::Error FileInAFile::SeekRelative(int Offset)

		Author:		Tim Browse
		Date:		2/6/01
		Inputs:		Offset - the offset, in bytes, to seek to, relative to the
							 current file read/write position.

							 Positive values seek forward, negative values
							 seek backwards.

		Returns:	Standard error

		Purpose:	Seek to a position in the file, relative to the current
					read/write position of the file.

	**************************************************************************/

	vos::Error FileInAFile::SeekRelative(int Offset)
	{
		if (m_pParentFile != NULL)
		{
			// Check validity of seek request...
			uint Pos;
			if (GetPos(&Pos))
			{
				if ((Pos + Offset) < m_FileLength)
				{
					// Nominal case
					return m_pParentFile->SeekFromStart(m_FileStart + Pos + Offset);
				}
				else
				{
					// Seek requested is outside limits
					return ERROR(vos::Err_File_SeekFail, msgFile_Access_Fail);
				}
			}
			else
			{
				// Unable to get file position for parent file - layer our
				// error on top of this.
				return ERROR(vos::Err_File_SeekFail, msgFile_Access_Fail);
			}
		}
		else
		{
			return ERROR(vos::Err_File_NotOpen, msgFile_Access_Fail);
		}
	}

	/**************************************************************************

	>	vos::Error FileInAFile::GetPos(uint* pPos) const

		Author:		Tim Browse
		Date:		2/6/01
		Outputs:	pPos - the current read/write position of the file.
		Returns:	Standard error

		Purpose:	Get the current read/write position of the file.

	**************************************************************************/

	vos::Error FileInAFile::GetPos(uint* pPos) const
	{
		uint Pos;
		if (m_pParentFile->GetPos(&Pos))
		{
			if ((Pos < m_FileStart) || (Pos > (m_FileStart + m_FileLength + 1)))
			{
				// Someone has moved the file pointer to a position before
				// the start of our file, or past the EOF position.
				COMPLAIN("Illegal file position in FileInAFile::Read()");
				return ERROR(vos::Err_File_SeekFail, msgFile_Access_Fail);
			}

			// Check for EOF position (which is when the pointer is 1 byte
			// past the last byte in the file)
			if (Pos == (m_FileStart + m_FileLength + 1))
			{
				return ERROR(vos::Err_File_EOF, msgFile_Access_Fail);
			}

			// Ok pointer is within valid section of the file, so adjust the 
			// position to be relative to our file.
			Pos = Pos - m_FileStart;

			// and return normalised position to caller
			*pPos = Pos;
			return VNOERROR;
		}
		else
		{
			// Seek in parent file failed.
			return LASTERROR;
		}
	}

	/**************************************************************************

	>	bool FileInAFile::eof() const

		Author:		Tim Browse
		Date:		2/6/01
		Returns:	true => eof encountered
					false => eof not encountered

		Purpose:	Find out if the end-of-file has been reached for this
					file yet.

	**************************************************************************/

	bool FileInAFile::eof() const
	{
		bool eof = false;

		if (m_pParentFile != NULL)
		{
			uint Pos;
			if (GetPos(&Pos))
			{
				// If we obtained the position ok, then we're not at EOF
			}
			else
			{
				eof = (vos::GetLastErrorCode() == vos::Err_File_EOF);
				ERROR_HANDLED();
			}
		}
		else
		{
			COMPLAIN("FileInAFile::eof() called on NULL parent file.");
		}

		return eof;
	}

	/**************************************************************************

	>	vos::Error FileInAFile::ReadLine(vos::String* pData)

		Author:		Tim Browse
		Date:		2/6/01
		Outputs:	pData - the line of text read from the file.
		Returns:	Standard error

		Purpose:	Read a line of text from the file, terminating at any
					type of line-end combination or EOF.  If a newline
					character is read, it is included in the string.

					There is no limit to the length of line that can be
					read from a text file with this function.

	**************************************************************************/

	vos::Error FileInAFile::ReadLine(vos::String* pData)
	{
		if (m_pParentFile == NULL)
		{
			COMPLAIN("FileInAFile::ReadLine() called on NULL parent file.");
			return ERROR(vos::Err_File_ReadFail, msgFile_Access_Fail);
		}

		// Before we start, are we already at EOF?
		if (eof())
		{
			// Error occurred.
			return ERROR(vos::Err_File_EOF, msgFile_Read_EOF);
		}

		// Clear the string first
		pData->Clear();

		// Buffer to hold line fragments
		const int kLineBufSize = 128;
		char szBuffer[kLineBufSize+1];
		szBuffer[0] = 0;

		uint Pos;

		for(;;)
		{
			// Remember where we are
			if (!GetPos(&Pos))
				return LASTERROR;

			// Read as much as we can into the buffer.
			uint cBytesRead = 0;
			if (!Read(szBuffer, kLineBufSize, &cBytesRead))
			{
				// Read failed - was it eof?
				if (vos::GetLastErrorCode() == vos::Err_File_EOF)
				{
					// Yes - have we hit EOF without reading *any* data?
					if ((cBytesRead == 0) && (pData->IsEmpty()))
					{
						// Yes, so just pass on EOF
						return LASTERROR;
					}

					// Otherwise, we've finished reading this line - simulate
					// a line ending (in case the line doesn't contain one)
					szBuffer[cBytesRead] = CR;

					// Clear the error, and let the code below sort out this
					// line as though we had found a real line ending.
					ERROR_HANDLED();
				}
				else
				{
					// Read error occured - return error to caller
					return LASTERROR;
				}
			}

			// OK, we read a chunk of data - is there a line-terminator
			// char sequence in it?
			uint iChar = 0;

			while ((iChar < kLineBufSize) &&
				   (szBuffer[iChar] != CR) &&
				   (szBuffer[iChar] != LF))
			{
				// Look at next character
				iChar++;
			}

			// Copy the characters we want to use to the destination string
			szBuffer[iChar] = 0;
			*pData += szBuffer;

			// Check to see if we found a line terminator
			if (iChar < kLineBufSize)
			{
				// We did, so seek to start of terminator in file
				if (!SeekFromStart(Pos + iChar))
					return LASTERROR;

				// Read next two bytes, and check for CRLF or LFCR combinations
				if (!ConsumeLineEnding())
					return LASTERROR;

				// Ok, so we've reached the end of the line, and we've positioned
				// the file pointer at the next line, so we're done.
				return VNOERROR;
			}

			// Otherwise, we keep going, looking for a line terminator...
		}
	}

	
	/**************************************************************************

	>	vos::Error FileInAFile::ConsumeLineEnding()

		Author:		Tim Browse
		Date:		02/06/2001
		Returns:	File errors.

		Purpose:	Move on in the file, so that the current line-ending is
					consumed.  Line endings can be single characters (CR or LF)
					or double-characters (CRLF or LFCR).

					- If a double-character sequence is found, we move past it.

					- If only one character is available from the file, we
					  treat that as legal (as it may be a line-ending followed
					  by EOF)

					- If no characters can be read, we return an error.  If the
					  first character is not CR or LF, we also return an error.

					    (i.e. there *must* be a line ending to consume)


					If the function returns with no error, the next data read
					from the file will be from the next line in the file.

					NB. Blank lines are preserved by this function, e.g. given
						the data stream:

							Hello<CR><LF><CR><LF>There<CR><LF>

						If a function reads "Hello", and then calls this function,
						the remaining data in the file stream will be:
		
							<CR><LF>There<CR><LF>

						i.e. only one line ending is ever consumed by this function.

						Similarly, this stream:

							Hello<CR><CR>There<CR>

						in the same situation would end up as:

							<CR>There<CR>
			
						and so on.

	**************************************************************************/

	vos::Error FileInAFile::ConsumeLineEnding()
	{
		// Read first byte and look at the next one...
		byte Ch1 = 0;

		if (!Read_byte(&Ch1))
		{
			// We should be able to read at least one character!
			return LASTERROR;
		}

		// Should be a line terminator...
		if ((Ch1 != CR) && (Ch1 != LF))
		{
			return ERROR(vos::Err_File_ReadFail, msgFile_Read_Fail);
		}

		// Look at the next character
		byte Ch2 = 0;

		if (!Read_byte(&Ch2))
		{
			// Ok, if this is EOF, then this is fine...
			if (vos::GetLastErrorCode() == vos::Err_File_EOF)
			{
				// EOF => we're done.
				ERROR_HANDLED();
				return VNOERROR;
			}
			else
			{
				// Otherwise, pass the error back to the caller
				return LASTERROR;
			}
		}

		// Check for double-character line-ending
		if (((Ch1 == CR) && (Ch2 == LF)) ||
			((Ch1 == LF) && (Ch2 == CR)))
		{
			// Found a double-char line ending, so we are now
			// in the correct position...
		}
		else
		{
			// Only a single-char ending - back up one byte so
			// the file pointer is in the correct position.
			SeekRelative(-1);
		}

		// All done
		return VNOERROR;
	}
}

