//
// FileLock.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "FileLock.h"

#include "Msgs.h"
#include "File.h"
#include "Errors.h"

#include "OS_Headers.h"

namespace vos
{
	/**************************************************************************

		FileLock::FileLock()

		Author:		Tim Browse
		Date:		10/9/00

		Purpose:	Ctor

	**************************************************************************/

	FileLock::FileLock() : m_Lock(0)
	{
	}

	/**************************************************************************

		void FileLock::SetLockFile(const FileSpec& LockFile)

		Author:		Tim Browse
		Date:		10/9/00
		Inputs:		LockFile - name of the lock file to use.

		Purpose:	Set the filename to use for creating the lock file.

	**************************************************************************/

	void FileLock::SetLockFile(const FileSpec& LockFile)
	{
		m_LockFile = LockFile;
	}

	/**************************************************************************

		Error FileLock::Lock()

		Author:		Tim Browse
		Date:		10/9/00
		Returns:	Error_FileLock_Fail - failed to lock the file.

		Purpose:	Lock the specified lock file.

	**************************************************************************/

	Error FileLock::Lock()
	{
		// If we don't have a lock filename, we can't do anything.
		if (m_LockFile.IsEmpty())
		{
			return ERROR(Err_Uninitialised, msgvos_Generic);
		}

		if (m_Lock == 0)
		{
			// Try to create lock file
#if defined VOS_MSWINDOWS
			String szLockFile = m_LockFile.GetString();

			m_hFile = ::CreateFile(szLockFile.c_str(),
								   GENERIC_WRITE,
								   0,							// deny sharing
								   NULL,						// default security
								   CREATE_NEW,					// Fail if file exists
								   FILE_FLAG_DELETE_ON_CLOSE,	// Delete the file when we close it
								   NULL);

			if (m_hFile == INVALID_HANDLE_VALUE)
			{
				// Unable to create the lock file
				return ERROR(Err_FileLock_Fail, msgFileLock_Fail);
			}
#elif defined VOS_MACOS
			String szLockFile = m_LockFile.GetString();
			Str255 LockFile;
			FSMakeFSSpec(0, 0, szLockFile.p_str( LockFile ), &m_FileSpec);
			OSErr Err = FSpCreate(&m_FileSpec, '????', 'TEXT', smSystemScript);

#else			
#error FileLock::Lock() needs porting
#endif
		}

		// File locked ok - increase semaphore
		m_Lock++;

		return VNOERROR;
	}

	/**************************************************************************

		void FileLock::Unlock()

		Author:		Tim Browse
		Date:		10/9/00

		Purpose:	Unlock the lock file.  Calls can be nested, so this may
					not cause the file to be unlocked immediately.

	**************************************************************************/

	void FileLock::Unlock()
	{
		// Decrease lock semaphore
		m_Lock--;

		// Delete lock file if all locks have been released
		if (m_Lock == 0)
		{
#if defined VOS_MSWINDOWS

			// Close Win32 file object - OS will delete it automatically.
			::CloseHandle(m_hFile);
			m_hFile = INVALID_HANDLE_VALUE;
			
#elif defined VOS_MACOS
			// Delete the file. We didn't open it.
			::FSpDelete(&m_FileSpec);
#else
#error FileLock::Lock() needs porting
#endif

		}
	}

	/**************************************************************************

		bool FileLock::IsLocked()

		Author:		Tim Browse
		Date:		10/9/00
		Returns:	true => File lock is active
					false => file lock is not in use

		Purpose:	Find out if the FileLock object currently has a valid
					lock on the lockfile.

	**************************************************************************/

	bool FileLock::IsLocked()
	{
		return (m_Lock > 0);
	}

	/**************************************************************************

		bool FileLock::IsLockedByOther()

		Author:		Tim Browse
		Date:		10/9/00
		Returns:	true => File is locked by another object
					false => file is not locked by another object

		Purpose:	Find out if the file this FileLock object refers to is
					currently locked by another object.

	**************************************************************************/

	bool FileLock::IsLockedByOther()
	{
		if (m_Lock > 0)
		{
			// We have a lock on this object; not anyone else
			return false;
		}

		// See if file exists
		return (m_LockFile.GetFileType() != FileType_NotFound);
	}

};


