//
// FileMatches.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_FileMatches_h
#define vos_FileMatches_h

#include "vos.h"

#include "FileSpec.h"

#ifdef VOS_MSWINDOWS
#include "Win32\Win32_Headers.h"
#endif


namespace vos
{
	class FileInfo;

	/***************************************************************************

	>	class FileMatches

		Author:		Tim Browse

		Purpose:	Scans for files on disk whose names match a certain pattern.

					Construct a FileMatches object by passing in the FileSpec
					of the filename pattern to match. This has 2 parts - the full
					path of the folder to search, and a wild-carded leaf-name.

					Note that the search is *not* recursive - it only looks in a
					single folder.

					(JCW) The wild-carded specifier is currently given in a Windows
					specific non portable form (e.g. to scan all files use "*.*")
					Possibly a better approach would be to have an optional filter
					that you can add to the returned files, but by default to
					simply return all matching files (i.e. effectively "*.*").

					Anyway, to use this, at present, do something like:

					\code

						FileSpec SearchFilter("c:\\Folder\\*.*");
						FileSpec FoundSpec;

						FileMatches Search(SearchFilter);

						while (Search.GetNextFile(&FoundSpec))
						{
							if (FoundSpec.GetFileType() == FileType_File)
								DealWithFile(FoundSpec);
						}

					\endcode

	***************************************************************************/

	class FileMatches
	{
		public:
			FileMatches();
			FileMatches(const FileSpec&);
			~FileMatches();

			void StartScan(const FileSpec&);
			
		public:
			bool GetNextFile(FileSpec* pFileSpec, FileInfo* pInfo = NULL);

		private:
			void CleanUp();
			void SkipSysDirEntries();

		private:
			// The parent folder - i.e. the one that contains the matching files
			FileSpec m_ParentFolder;

		#ifdef VOS_MSWINDOWS

			private:
				WIN32_FIND_DATA m_FindData;
				HANDLE m_hFile;

		#elif (defined VOS_MACOS)
		
		private:
			// Set up the search parameter block
			bool InitializeFileMatches(const FileSpec& TheFileSpec);
		
			private:
				Str255			m_SearchName;
				CInfoPBRec		m_SearchRec;
				int32			m_SearchIndex;
				int32			m_SearchDir;
		#endif
	};

};

#endif // !vos_FileMatches_h
