//
// FileSpec.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_FileSpec_h
#define vos_FileSpec_h

#include "vos.h"

#include "String.h"
#include "ErrorManager.h"

#include "StdHdr.h"

namespace vos
{
	class DiskFile;

	/*!
		Enum: FileType

		The FileType enum is used when finding out if a named file 
		exists.  The named file is either a file, a folder, or it does 
		not exist.
	*/
	typedef enum
	{
		FileType_NotFound,
		FileType_File,
		FileType_Folder
	}
	FileType;

	/*!
		Enum: FileCreation

		The FileCreation enum is used to control creation of files when
		calling the FileSpec::Copy() function.  You can optionally force
		the Copy function to autmatically create any necessary parent 
		folders before creating the copy of the file.
	*/
	typedef enum
	{
		// Normal file creation
		FileCreation_Normal,					

		// Create parent folder to hold file if it does not exist.
		FileCreation_ForceParentCreation		
	}
	FileCreation;

	class FileMatches;

	/**************************************************************************

	>	class CreateFolderHandler

		Author:		Tim Browse

		Date:		28/09/99

		Purpose:	Callback class (interface) for creating folders.

					When calling FileSpec::CreateFolder() or FileSpec::CreateFolderTree(), 
					you can pass in a pointer to an object with this interface to 
					receive notification of any folder that is actually created.

	**************************************************************************/

	class CreateFolderHandler
	{
		public:
			// Over-ride this function to receive notifications.
			virtual void NotifyFolderCreated(const FileSpec& NewFolderName) {}
	};

	/**************************************************************************

	>	class CopyFileHandler : public CreateFolderHandler

		Author:		Tim Browse

		Date:		28/09/99

		Purpose:	Callback class (interface) for copying files.

					When calling FileSpec::Copy(), you can pass in a pointer
					to an object with this interface to receive regular
					updates on the copy process, and the option to cancel
					the copy.

	**************************************************************************/

	class CopyFileHandler : public CreateFolderHandler
	{
		public:
			// Used to 'pulse' UI (or do whatever you want), so a progress
			// display can be used, and/or provide the user with a 'Cancel'
			// button to cancel the file copy.
			// If an error is returned, the copy is cancelled, otherwise it
			// continues as normal.
			virtual Error CopyFileCallback(uint cBytesCopied) = 0;
	};



	/**************************************************************************

	>	class FileSpec

		Author:		Tim Browse, Rob Myers

		Date:		15/08/99, 13/07/00

		Purpose:	The FileSpec class is used to provide a platform 
					independent way of specifying a file/folder object.

					On most platforms this is simply a string - a filename.

					On MacOS, this can be a real FileSpec object.  Even on
					platforms that just use strings, the folder separator
					character varies - it can be '\', '/' or even '.'
					characters.

					Strings and FileSpecs
					~~~~~~~~~~~~~~~~~~~~~

					All platforms must support the flattening of a FileSpec
					object into a String - and the reversal of this, which
					is to convert a String into a FileSpec object.

					If you convert FileSpec F into a string object S, and
					then initialise another FileSpec object, F' using the
					string S, then the two FileSpec objects F and F' *must*
					refer to the same file object.

					When converting between FileSpecs and String objects, 
					you use the SetFromString() and GetString() functions.
					Both functions allow you to specify what character is
					used to separate a pathname into folders in the Strings.
					You cannot control what character is used to separate
					the folders internally in a FileSpec object - and it
					should not matter anyway.

					These two functions default to using the character stored
					in FileSpec::kStringFolderSeparator - if you want to
					globally change the behaviour of these functions (rather 
					than specifying the folder separator everytime you
					call them) then you can simply change the contents of
					this variable.

					Absolute and Relative FileSpecs
					~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

					FileSpec objects can hold full or partial pathnames - the
					class does not enforce that a FileSpec must describe an
					existing file.

					You can append one filespec onto another using the
					function AppendRelativeFileSpec().  You can extract
					the tail part of a FileSpec by passing in the parent
					FileSpec to GetRelativeFileSpec().  For example...

					\code

						// double \\'s omitted for clarity...
						FileSpec F;
						F.SetFromString("c:\games\files");

						FileSpec F2;
						F2.SetFromString("c:\games");

						FileSpec F3;
						F1.GetRelativeFileSpec(F2, &F3);

						String S = F3.GetString();

						// S now equals "files"

					\endcode

					Of course, you can do all this with strings if you want,
					but the FileSpec class takes care of all the fiddly messing
					about with folder separators, and knowing when to add/remove
					them etc.  You know it makes sense.
						

	**************************************************************************/

	class FileSpec
	{
		public:
			// The character to be used to separate folders when dealing
			// with strings.
			static char kStringFolderSeparator;

		public:
#ifdef VOS_MACOS
			FileSpec() : m_FSSpecDirty(true) {}
			FileSpec(const String& Pathname) :	m_FileName(Pathname), m_FSSpecDirty(true) {}
			FileSpec(const ::FSSpec& MacFileSpec);
#else
			FileSpec() {}
			FileSpec(const String& Pathname) : m_FileName(Pathname) {}
#endif
			// Append a folder name on to the end of a filespec.
			void AppendFolder(const String& FolderName);

			// Append a filename (leafname) on to the end of a filespec.
			void AppendFilename(const String& Filename);

			// Append a partial pathname on to the end of a filespec.
			void AppendRelativeFileSpec(const FileSpec& RelativeFileSpec);

			// Extract a partial pathname from a filespec.
			bool GetRelativeFileSpec(const FileSpec& Prefix, FileSpec* pRelativeFileSpec) const;

			// Standard comparison operators.
			bool operator==(const FileSpec& rhs) const;
			bool operator!=(const FileSpec& rhs) const;

			// Get next level up from this FileSpec
			virtual bool GetParent(FileSpec* pParent) const;

			// Is the filespec empty?
			bool IsEmpty() const { return m_FileName.IsEmpty(); }

			// Clear the filespec to an empty filename
			void Clear();
			
			// Is the FileSpec absolute or relative?
			bool IsAbsolute() const;

			// Convert between FileSpec objects and String objects.
			String GetString(char FolderSeparator = kStringFolderSeparator) const;
			void SetFromString(const String& Filename, char FolderSeparator = kStringFolderSeparator);

			// Extract the leafname from a filespec.
			String GetLeafname() const;

			// Find out if a filespec names a file, a folder, or nothing at all.
			FileType GetFileType() const;

			// Determine the size of the referenced file
			uint32 GetFileSize(void) const;

			// Get last-modified time for the referenced file
			Error GetFileTime(Time* pTime) const;

			// Check to see if file is read only
			bool IsReadOnly() const;

			//
			// Utility functions : create/copy/delete
			//
			Error CreateFolder(CreateFolderHandler *pHandler = NULL) const;
			Error CreateFolderTree(CreateFolderHandler *pHandler = NULL) const;

			Error Delete() const;
			Error Rename(const FileSpec& NewName) const;
			Error Copy(const FileSpec& DestinationFile,
					   FileCreation CreateRule = FileCreation_Normal,
					   CopyFileHandler* pCopyFileHandler = NULL) const;

			//
			// Comparison
			//
			bool operator<(const FileSpec& rhs) const;

		protected:
			void AppendFolderSeparator();

		protected:
			String m_FileName;

			// The DiskFile class needs access to our data
			friend class DiskFile;
			
#ifdef VOS_MACOS
		private:
		
			// Get a full path from a MacOS folder ref
			static String GetFullPath (int16 VRefNum, int32 DirID );
			
			// Rebuild the FSSpec if needed
			void UpdateFSSpec () const;
			
		private:
			
			// The cached MacOS FSSpec
			FSSpec 	m_FSSpec;
			
			// A flag set by string manipulation routines to indicate
			// whether the FSSpec needs rebuilding to reflect changes to the path
			bool	m_FSSpecDirty;
#endif
	};
};


#endif // !vos_FileSpec_h
