//
// MemoryFile.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "MemoryFile.h"

#include "Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/**************************************************************************
	
	>	MemoryFile::MemoryFile()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	MemoryFile::MemoryFile() : m_pData(NULL),
							   m_cbBufferLength(0),
							   m_Pos(0),
							   m_IOwnTheBuffer(false),
							   m_FileOpen(false)
	{
	}

	/**************************************************************************
	
	>	MemoryFile::~MemoryFile()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	MemoryFile::~MemoryFile()
	{
		Close();

		// Vape any existing buffer
		DeleteBuffer();
	}


	/**************************************************************************

	>	Error MemoryFile::Create(byte* pBuffer, uint cbBufferLength)

		Author:		Tim Browse/Jason Williams
		Date:		19/1/2001

		Purpose:	Creates a memory file from a RAM buffer

		Notes:		The MemoryFile SHARES the data in the buffer, so the buffer
					needs to remain available for the lifetime of the MemoryFile.

					It is still your responsibility to delete the buffer yourself
					after the MemoryFile has been destroyed.

	**************************************************************************/

	Error MemoryFile::Create(byte* pBuffer, uint cbBufferLength)
	{
		if (pBuffer == NULL)
		{
			// Can't open a file with no memory
			return ERROR(Err_File_OpenFailed, "Failed to create memory-based file.");
		}

		if (IsOpen())
		{
			return ERROR(Err_File_OpenFailed, "MemoryFile::Create - file is already open");
		}

		// Vape any existing buffer
		DeleteBuffer();

		// Remember file data
		m_pData = pBuffer;
		m_cbBufferLength = cbBufferLength;
		m_Pos = 0;

		// Worked ok
		return VNOERROR;
	}



	/**************************************************************************

	>	Error MemoryFile::Create(File *pSourceFile)

		Author:		Jason Williams
		Date:		15/1/2001

		Inputs:		pSourceFile	- The other file to cache in RAM

		Purpose:	Creates a memory file from another file, effectively
					caching the fiel in memory.

		Notes:		pSourceFile must not be Open
		
					A MemoryFile Create'd from another MemoryFile will make
					A COMPLETE COPY of the other file - they will not share
					one buffer between them.

	**************************************************************************/

	Error MemoryFile::Create(File *pSourceFile)
	{
		if (pSourceFile == NULL)
		{
			// Can't open a file with no source file
			return ERROR(Err_File_OpenFailed, "Failed to create memory-based file.");
		}

		if (IsOpen())
		{
			return ERROR(Err_File_OpenFailed, "MemoryFile::Create - file is already open");
		}

		// Vape any existing buffer
		DeleteBuffer();

		// Try to open the file
		RETURN_ON_FAILURE(pSourceFile->Open(fam_Read));

		// Try to create a buffer & load the file into it
		RETURN_ON_FAILURE(pSourceFile->GetSize(&m_cbBufferLength));
		m_Pos = 0;

		if (m_cbBufferLength > 0)
		{
			m_pData = new uint8[m_cbBufferLength];

			if (m_pData == NULL)
			{
				// Memory allocation failure
				pSourceFile->Close();
				return ERROR(Err_File_OpenFailed, "Failed to create memory-based file.");
			}

			// Remember that we are responsible for deleting the data buffer on exit
			m_IOwnTheBuffer = true;
		}

		if (m_pData != NULL)
		{
			// Read the other file into our buffer
			RETURN_ON_FAILURE(pSourceFile->Read(m_pData, m_cbBufferLength));
		}

		pSourceFile->Close();

		// Worked ok
		return VNOERROR;
	}


	/**************************************************************************

	>	Error MemoryFile::Open(byte* pBuffer, uint cbBufferLength, FileMode Mode)

		Author:		Jason
		Date:		15/1/2001

		Purpose:	Creates and Opens a memory file from a static buffer

		Notes:		The MemoryFile SHARES the data in the buffer, so the buffer
					needs to remain available for the lifetime of the MemoryFile.

					It is still your responsibility to delete the buffer yourself
					after the MemoryFile has been destroyed.

	**************************************************************************/

	Error MemoryFile::Open(byte* pBuffer, uint cbBufferLength, FileMode Mode)
	{
		RETURN_ON_FAILURE(Create(pBuffer, cbBufferLength));
		RETURN_ON_FAILURE(Open(Mode));

		return(VNOERROR);
	}



	/**************************************************************************

	>	Error MemoryFile::Open(File *pSourceFile, FileMode Mode)

		Author:		Jason
		Date:		15/1/2001

		Inputs:		pSourceFile	- The other file to cache in RAM

		Purpose:	Creates and Opens a memory file from another file,
					effectively caching the file in memory.

		Notes:		pSourceFile must not be Open
		
					A MemoryFile Create'd from another MemoryFile will make
					A COMPLETE COPY of the other file - they will not share
					one buffer between them.

	**************************************************************************/

	Error MemoryFile::Open(File *pSourceFile, FileMode Mode)
	{
		RETURN_ON_FAILURE(Create(pSourceFile));
		RETURN_ON_FAILURE(Open(Mode));

		return(VNOERROR);
	}



	/**************************************************************************

	>	virtual 	Error MemoryFile::Open(FileMode Mode)

		Author:		Tim Browse/Jason Williams
		Date:		19/1/2001

		Inputs:		Mode	- The access mode for the file

		Purpose:	Creates and Opens a memory file from another file,
					effectively caching the file in memory.

	**************************************************************************/

	Error MemoryFile::Open(FileMode Mode)
	{
		if (IsOpen())
		{
			return ERROR(Err_File_OpenFailed, "MemoryFile::Open - file is already open");
		}

		if ((Mode & fam_Write) != 0)
		{
			// Can't open a file with no memory
			return ERROR(Err_File_OpenFailed, "MemoryFile::Open - Memory Files are read-only");
		}

		if (m_pData == NULL)
		{
			return ERROR(Err_File_OpenFailed, "MemoryFile::Open - You must call Create() before Open()");
		}

		m_FileOpen = true;
		m_Pos = 0;

		return(VNOERROR);
	}



	/**************************************************************************

	>	virtual void MemoryFile::Close()

		Author:		Tim Browse/Jason Williams
		Date:		19/1/2001

		Purpose:	Closes the file.

					NOTE that closing the file does not release any buffers
					allocated by the MemoryFile - they are only deallocated
					when you delete the MemoryFile object (or assign a new
					buffer). This allows you to re-open the MemoryFile and
					still find data in it!

	**************************************************************************/

	void MemoryFile::Close()
	{
		m_FileOpen = false;
	}

	bool MemoryFile::IsOpen() const
	{
		return (m_FileOpen);
	}


	/**************************************************************************

	>	virtual Error MemoryFile::GetSize(uint* pSize)

		Author:		Jason
		Date:		15/1/2001

		Outputs:	pSize	- The size, in bytes, of this file

		Purpose:	Determine the size (in bytes) of this file

	**************************************************************************/

	Error MemoryFile::GetSize(uint* pSize)
	{
		COMPLAIN_IF(pSize == NULL, "Illegal NULL param");

		if (pSize != NULL)
			*pSize = m_cbBufferLength;

		return(VNOERROR);
	}


	/**************************************************************************
	
	>	Error MemoryFile::Read(void* pData, uint cBytes, uint* pcBytes)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		pData - the buffer to store the bytes in
					cBytes - the number of bytes to read (pData must be
							 big enough to hold this number of bytes)
		Outputs:	pcBytes - the number of bytes that were actually read.

		Returns:	Errors: Err_File_NotOpen
							Err_File_EOF
	
		Purpose:	Read a number of bytes from the MemoryFile.
	
	**************************************************************************/
	
	Error MemoryFile::Read(void* pData, uint cBytes, uint* pcBytes)
	{
		// Check the file is open
		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, 
						 "An attempt was made to read from a memory-based file that has not been opened.");
		}

		// Check we're not at EOF
		if (eof())
		{
			return ERROR(Err_File_EOF, "End of file reached in memory-based file.");
		}

		// Read as much as we can
		uint cbBytesLeft = m_cbBufferLength - m_Pos;

		uint cbBytesToRead = cBytes;

		if (cbBytesToRead > cbBytesLeft)
			cbBytesToRead = cbBytesLeft;

		memcpy(pData, m_pData + m_Pos, cbBytesToRead);

		// Move file pointer on
		m_Pos += cbBytesToRead;

		// Tell caller how many bytes we read
		if (pcBytes != NULL)
		{
			*pcBytes = cbBytesToRead;
		}

		// Tell caller if it worked
		if (cbBytesToRead == cBytes)
		{
			// it worked
			return VNOERROR;
		}
		else
		{
			// Failed - EOF encountered
			return ERROR(Err_File_EOF, "MemoryFile::Read() encountered EOF.");
		}
	}

	/**************************************************************************
	
	>	Error MemoryFile::Write(const void* pData, uint cBytes, uint* pcBytes)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		pDate - the data to write to the file
					cBytes - the number of bytes to write to the file
		Outputs:	pcBytes - the number of bytes actually written to the file.
		Returns:	Err_File_WriteFail (always, because function does not work)
	
		Purpose:	Write a number of bytes to the MemoryFile.

					NB. This function is unimplemented!
	
	**************************************************************************/
	
	Error MemoryFile::Write(const void* pData, uint cBytes, uint* pcBytes)
	{
		// TODO: Implement this function
		COMPLAIN("MemoryFile::Write() not yet implemented!");
		return ERROR(Err_File_WriteFail, "MemoryFile::Write() not yet implemented!");
	}

	/**************************************************************************
	
	>	Error MemoryFile::SeekFromStart(uint Offset)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		Offset - absolute file position to seek to
		Returns:	Errors:	Err_File_NotOpen
	
		Purpose:	Seek to an absolute file position (i.e. relative to the 
					start of the file).
	
	**************************************************************************/
	
	Error MemoryFile::SeekFromStart(uint Offset)
	{
		// Check the file is open
		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, 
						 "An attempt was made to seek in a memory-based file that has not been opened.");
		}

		if (Offset >= m_cbBufferLength)
		{
			// Seek beyond end of data - this means eof. Set the file pointer to eof for consistency
			Offset = m_cbBufferLength;
		}

		// otherwise, process the valid seek request
		m_Pos = Offset;
		return VNOERROR;
	}

	/**************************************************************************
	
	>	Error MemoryFile::SeekFromEnd(uint Offset)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		Offset - the number of bytes before the end of the file
							 that we should seek to.  If this specifies a 
							 position before the start of the file, then the
							 function will seek to the start of the file.

		Returns:	Errors:	Err_File_NotOpen
	
		Purpose:	Seek to a file position that is relative to the end of
					the file.
	
	**************************************************************************/
	
	Error MemoryFile::SeekFromEnd(uint Offset)
	{
		// Check the file is open
		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, 
						 "An attempt was made to seek in a memory-based file that has not been opened.");
		}

		if (Offset >= m_cbBufferLength)
			m_Pos = 0;	// Clamp any seek beyond start of data to 0
		else
			m_Pos = (m_cbBufferLength - 1) - Offset;

		return VNOERROR;
	}

	/**************************************************************************
	
	>	Error MemoryFile::SeekRelative(int Offset)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		Offset - the number of bytes to move the file pointer;
							 positive numbers seek forward, negative numbers
							 seek backwards

		Returns:	Errors:	Err_File_NotOpen
	
		Purpose:	Seek to a position relative to the current position.

					If the offset specifies a position outside the file, the
					position is clamped to the start or the end of the file.
	
	**************************************************************************/
	
	Error MemoryFile::SeekRelative(int Offset)
	{
		// Check the file is open
		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, 
						 "An attempt was made to seek in a memory-based file that has not been opened.");
		}

		// Check bounds of seek request
		if (Offset < 0)
		{
			// Seeking backwards - is it seeking past the beginning of the file?
			if (abs(Offset) > m_Pos)
			{
				m_Pos = 0;					// Clamp to 0
			}
		}
		else
		{
			// Seeking forwards - is it seeking past the end of the file?
			if ((m_Pos + Offset) >= m_cbBufferLength)
			{
				m_Pos = m_cbBufferLength;	// Clamp to eof
			}
		}

		// Ok, it's a valid seek, so do it.
		m_Pos += Offset;
		return VNOERROR;
	}

	/**************************************************************************
	
	>	Error MemoryFile::GetPos(uint* pPos) const
	
		Author:		Tim Browse
		Date:		15/4/2002
		Outputs:	pPos - the current file read/write position
		Returns:	Errors:	Err_File_NotOpen
	
		Purpose:	Find out where the file read/write pointer is.
	
	**************************************************************************/
	
	Error MemoryFile::GetPos(uint* pPos) const
	{
		// Check the file is open
		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, 
						 "An attempt was made to seek in a memory-based file that has not been opened.");
		}

		// Check we're not at EOF
		if (eof())
		{
			*pPos = m_cbBufferLength;
		}
		else
		{
			// Otherwise, return position
			*pPos = m_Pos;
		}

		return VNOERROR;
	}

	/**************************************************************************
	
	>	bool MemoryFile::eof() const
	
		Author:		Tim Browse
		Date:		15/4/2002
		Returns:	true => EOF reached
					false => otherwise
	
		Purpose:	Check to see if EOF has been encountered yet.
	
	**************************************************************************/
	
	bool MemoryFile::eof() const
	{
		return (m_Pos >= m_cbBufferLength);
	}

	/**************************************************************************
	
	>	Error MemoryFile::ReadLine(String* pData)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Outputs:	pData - the line of text that was read from the file.

		Returns:	Errors:	Err_File_NotOpen
							Err_File_EOF
							Err_File_ReadFail
	
		Purpose:	Read a line of text from the file (i.e. read until the first
					line terminator has been found).
	
	**************************************************************************/
	
	Error MemoryFile::ReadLine(String* pData)
	{
		// Check the file is open
		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, 
						 "An attempt was made to read from a memory-based file that has not been opened.");
		}

		// Check we're not at EOF
		if (eof())
		{
			return ERROR(Err_File_EOF, "End of file reached in memory-based file.");
		}

		// Ok, find out how long the next line is - lines are terminated by CR, LF, or EOF
		uint cLineLength = 0;
		while (((m_Pos + cLineLength) < m_cbBufferLength) && 
			   (!IsLineTerminator(m_pData[m_Pos + cLineLength])))
		{
			// Move to next character
			cLineLength++;
		}

		// Ok, we have the end of the line - copy the data, excluding line terminators.
		pData->Clear();
		if (cLineLength > 0)
		{
			pData->AppendChars((char *) (m_pData + m_Pos), cLineLength);
		}

		// Move to end of line
		m_Pos += cLineLength;

		// If we didn't get to eof, skip the line terminator
		if (!eof())
		{
			// We must have a line terminator to get here
			byte EOL = m_pData[m_Pos];

			if (!IsLineTerminator(EOL))
			{
				COMPLAIN("Logic error in memoryFile::ReadLine()!");
				return ERROR(Err_File_ReadFail, "Logic error in memoryFile::ReadLine()!");
			}

			// Ok, so move past the terminator
			m_Pos++;

			if (!eof())
			{
				// Do we have another terminator, that isn't the same as the previous one?
				if (IsLineTerminator(m_pData[m_Pos]) && (m_pData[m_Pos] != EOL))
				{
					// Yes, so it's a double terminator like CRLF or LFCR etc, so skip it too.
					m_Pos++;
				}
			}
		}

		// All done
		return VNOERROR;
	}

	/**************************************************************************
	
	>	bool MemoryFile::IsLineTerminator(byte ch)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		ch - the character to test
		Returns:	true => ch is a line terminator
					false => otherwise
	
		Purpose:	Helper function to check to see if a character is classified
					as a line terminator character.  Currently this includes
					CR, LF and NUL.
	
	**************************************************************************/
	
	bool MemoryFile::IsLineTerminator(byte ch)
	{
		return ((ch == 0) || (ch == 0x0A) || (ch == 0x0D));
	}


	/**************************************************************************

	>	void MemoryFile::DeleteBuffer(void)

		Author:		Jason
		Date:		15/1/2001

		Purpose:	Internal helper

					Clears any memory buffer we are using, and resets the file
					position and buffer length to 0. If the data buffer was
					allocated by this object then the buffer memory is deallocated.

	**************************************************************************/

	void MemoryFile::DeleteBuffer(void)
	{
		if (m_pData != NULL && m_IOwnTheBuffer)
			delete(m_pData);

		m_pData				= NULL;
		m_cbBufferLength	= 0;
		m_Pos				= 0;
		m_IOwnTheBuffer		= false;
		m_FileOpen			= false;
	}

};

