//
// Preferences.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Preferences.h"
#include "PreferencesImp.h"
#include "String.h"
#include "Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	// Preference details
	String Preferences::s_AuthorName;
	String Preferences::s_ProgramName;
	bool   Preferences::s_bInitialised = false;

	const char kPrefList_Count[] = "Count";
	const char kPrefList_Item[]  = "Item%d";

	/**************************************************************************

	>	void Preferences::Init(const String& AuthorName,
							   const String& ProgramName)


		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		AuthorName - name of the author/developer, 
								 e.g. "Microsoft" or "Tim Browse"

					ProgramName - name of the program to store preferences
								  for, e.g. "Word" or "FastFind".

		Purpose:	Initialise the preferences system.  This is necessary on
					most platforms so the library knows where to store the 
					preference data.

	**************************************************************************/

	void Preferences::Init(const String& AuthorName,
						   const String& ProgramName)
	{
		// Remember strings for preference details
		s_AuthorName = AuthorName;
		s_ProgramName = ProgramName;

		if (s_ProgramName.GetLength() > 0)
		{
			// Minimum requirements met
			s_bInitialised = true;
		}
		else
		{
			// Must have a program name to run the preferences system with.
			COMPLAIN("You must pass a non-NULL program name to Preferences::Init()!");
			s_bInitialised = false;
		}
	}

	/**************************************************************************
	
	>	bool Preferences::IsInitialised()
	
		Author:		Tim Browse
		Date:		15/4/2002
		Returns:	true => preferences system has been initialised
					false => otherwise
	
		Purpose:	Find out if the preference system has been initialised.
	
	**************************************************************************/
	
	bool Preferences::IsInitialised()
	{
		return s_bInitialised;
	}

	Preferences::Preferences(const String& SectionName,
							 Preferences* pParent) : m_pImp(VOSFactory().MakePreferences()),
													 m_SectionName(SectionName)
	{
		// If we have a parent, add this object as a child
		if (pParent != NULL)
		{
			InsertLastChild(pParent);
		}

		// Link the implementation object back to this object, so it has
		// some context in which to operate.
		m_pImp->LinkToPreferences(this);
	}

	/**************************************************************************
	
	>	void Preferences::CleanUp()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Used internally to clean up the preferences system.
	
	**************************************************************************/
	
	void Preferences::CleanUp()
	{
		// Delete children of this node.
		while (GetFirstChild() != NULL)
		{
			GetFirstChild()->DeleteSubTree();
		}

		// Delete implementation object
		delete m_pImp;
		m_pImp = NULL;
	}

	/**************************************************************************
	
	>	Preferences::~Preferences()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	Preferences::~Preferences()
	{
		CleanUp();
	}

	/**************************************************************************
	
	>	Preferences& Preferences::GetPreferences()
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		-
		Outputs:	-
		Returns:	-
	
		Purpose:	Get the root preferences object - you should use this
					most of the time when using the preferences system, e.g.

					\code

						vos::GetPreferences().Read("MaxResults", &s_cMaxResults);

					\endcode
	
	**************************************************************************/
	
	Preferences& Preferences::GetPreferences()
	{
		// Singleton object.
		static Preferences ThePreferences("", NULL);

		return ThePreferences;
	}


	/**************************************************************************

	>	bool Preferences::SplitPreferencePathname(const vos::String& Pathname,
												  vos::String* pHead,
												  vos::String* pTail)


		Author:		Tim Browse
		Date:		5/5/2002
		Inputs:		Pathname - the preference pathname to split.
		Outputs:	pHead - the first component of the pathname.
					pTail - the remainder of the pathname (after *pHead)
		Returns:	true => a multi-level pathname was found.
					false => a single name was found (i.e. no path separator
							 characters were detected in the name).

		Purpose:	Split a preference pathname into the first component of
					the path, and the remainder.

					For example:

						Pathname					Head		Tail
						~~~~~~~~					~~~~		~~~~

						"Display.UI.Feedback"	=>	"Display"	"Feedback.UI"

						"Display.UI"			=>	"Display"	"UI"

						"Display"				=>	"Display"	""


					As you've probably guess, '.' is the pathname separator
					in preference pathnames.

	**************************************************************************/

	bool Preferences::SplitPreferencePathname(const String& Pathname,
											  String* pHead,
											  String* pTail)
	{
		uint iPos;
		if (Pathname.Find('.', &iPos))
		{
			// Split into first path component, and remainder of string
			*pHead = Pathname.Left(iPos);
			*pTail = Pathname.Mid(iPos+1);
			return true;
		}
		else
		{
			// It's just a single name with no path info, so 
			// just use that.
			*pHead = Pathname;
			pTail->Empty();
			return false;
		}
	}

	/**************************************************************************

	>	bool Preferences::SplitPathnameAtLeaf(const String& PrefName, 
											  String* pParentName,
											  String* pLeafName)


		Author:		Tim Browse
		Date:		5/5/2002
		Inputs:		PrefName - name of the preference

		Outputs:	pParentName - name of the parent of this preference
								  (can be NULL)
					pLeafname   - the leafname of this preference name
								  (can be NULL)

		Returns:	true => Preference has a parent, and it is in pParentName
					false => preference name is just a leafname - it has no
							 parent (if so, neither output parameter is updated)

		Purpose:	Given a preference pathname (e.g. "Display.UI.Feedback")
					decompose into the parent and leafname (in this example, 
					this results in "Display.UI" and "Feedback").

					If there is no parent, the function returns false.

	**************************************************************************/

	bool Preferences::SplitPathnameAtLeaf(const String& PrefName, 
										  String* pParentName,
										  String* pLeafname)
	{
		uint iPos;
		if (PrefName.ReverseFind('.', &iPos))
		{
			// Parent found
			if (pParentName != NULL)
				*pParentName = PrefName.Left(iPos);

			if (pLeafname != NULL)
				*pLeafname = PrefName.Mid(iPos+1);

			return true;
		}

		// No path separator found - it's just a preference name
		return false;
	}

	/**************************************************************************
	
	>	Preferences* Preferences::FindParentSection(const String& PrefName,
													const String& pLeafname = NULL)
	
		Author:		Tim Browse
		Date:		5/5/2002
		Inputs:		PrefName - name of the preference whose parent is to 
							   be found.

		Outputs:	pLeafname - the leafname from PrefName (which will be the
								same as PrefName is no parent is found).
								Pass in NULL if you don't need this.

		Returns:	The Preferences object that should contain the specified
					preference setting.
	
		Purpose:	Given a preference name, find the Preferences object
					that should hold this preference.   The name can be
					a preference pathname, such as "Display.UI.Feedback",
					and the sections will be traversed/created as necessary.
	
					NB. Asking for "Display.UI.Feedback" will return the
						section for "Display.UI", as it is assumed that
						"Feedback" (i.e. the leafname) describes a value
						contained in the "Display.UI" section.
		
						Similarly, asking for, e.g. "Feedback" will just
						return this same Preferences object (as there is
						no parent in the preference name).

	**************************************************************************/
	
	Preferences* Preferences::FindParentSection(const String& PrefName,
												String* pLeafname)
	{
		String Parent;
		if (SplitPathnameAtLeaf(PrefName, &Parent, pLeafname))
		{
			// Ok, find the parent section using this section as
			// the starting point
			return FindSection(Parent);
		}
		else
		{
			// There is no parent - just use this section
			*pLeafname = PrefName;
			return this;
		}
	}

	/**************************************************************************
	
	>	Preferences* Preferences::FindSection(const String& SectionName)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		SectionName - name of the section (can be either a
								  leafname, or a pathname such as
								  "Display.Performance").
		Returns:	The Preferences section corresponding to the named
					section you requested.
	
		Purpose:	Find/make the named preference section.  Once you have
					this Preferences object, you can read/write values with
					it, so you can divide up your preferences if you like.
	
	**************************************************************************/
	
	Preferences* Preferences::FindSection(const String& SectionName)
	{
		// Check for pathname and decompose it
		vos::String SectionNameHead, SectionNameTail;
		SplitPreferencePathname(SectionName, &SectionNameHead, &SectionNameTail);

		// Do we already have this section?
		Preferences *pSection = (Preferences *) GetFirstChild();
		
		while (pSection != NULL)
		{
			if (pSection->m_SectionName.CompareText(SectionNameHead))
			{
				// Found a match
				break;
			}

			// Try next section
			pSection = (Preferences *) pSection->GetNext();
		}

		// If we don't have a section, then make a new one.
		if (pSection == NULL)
		{
			pSection = new Preferences(SectionNameHead, this);
		}

		// Do we have to search deeper?
		if (!SectionNameTail.IsEmpty())
		{
			// Yes
			return pSection->FindSection(SectionNameTail);
		}
		else
		{
			// Otherwise, just return the found/new section to caller
			return pSection;
		}
	}

	/**************************************************************************
	
	>	bool Preferences::Read(const String& BooleanPrefName,
							   bool bDefaultValue)
	
		Author:		Tim Browse
		Date:		28/7/2002
		Inputs:		BooleanPrefName - name of the boolean preference to read
					bDefaultValue - the value to return if the preference
									cannot be read successfully.
		Returns:	named value of the preference (or the default value if
					it cannot be found).
	
		Purpose:	Special version of Read() for boolean preferences.
	
	**************************************************************************/
	
	bool Preferences::Read(const String& BooleanPrefName, bool bDefaultValue)
	{
		Read(BooleanPrefName, &bDefaultValue);
		return bDefaultValue;
	}

	/**************************************************************************
	
	>	bool Preferences::Read(const String& PrefName, String* pPrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
		Outputs:	pPrefValue - the value of the preference.
		Returns:	true => preference found
					false => preference did not exist.
	
		Purpose:	Read a String value from the named preference.
	
	**************************************************************************/
	
	bool Preferences::Read(const String& PrefName, String* pPrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Read(PrefLeafName, pPrefValue);
	}

	/**************************************************************************
	
	>	bool Preferences::Read(const String& PrefName, int* pPrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
		Outputs:	pPrefValue - the value of the preference.
		Returns:	true => preference found
					false => preference did not exist.
	
		Purpose:	Read an int value from the named preference.
	
	**************************************************************************/
	
	bool Preferences::Read(const String& PrefName, int* pPrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Read(PrefLeafName, pPrefValue);
	}

	/**************************************************************************
	
	>	bool Preferences::Read(const String& PrefName, bool* pPrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
		Outputs:	pPrefValue - the value of the preference.
		Returns:	true => preference found
					false => preference did not exist.
	
		Purpose:	Read a bool value from the named preference.
	
	**************************************************************************/
	
	bool Preferences::Read(const String& PrefName, bool* pPrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Read(PrefLeafName, pPrefValue);
	}

	/**************************************************************************
	
	>	bool Preferences::Read(const String& PrefName, FileSpec* pPrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
		Outputs:	pPrefValue - the value of the preference.
		Returns:	true => preference found
					false => preference did not exist.
	
		Purpose:	Read a FileSpec value from the named preference.
	
	**************************************************************************/
	
	bool Preferences::Read(const String& PrefName, FileSpec* pPrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Read(PrefLeafName, pPrefValue);
	}


	/**************************************************************************
	
	>	bool Preferences::Write(const String& PrefName, const String& PrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
					PrefValue - the value to write to the preference.
		Returns:	true => preference written successfully
					false => otherwise
	
		Purpose:	Write a String value to the named preference.
	
	**************************************************************************/
	
	bool Preferences::Write(const String& PrefName, const String& PrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Write(PrefLeafName, PrefValue);
	}

	/**************************************************************************
	
	>	bool Preferences::Write(const String& PrefName, const char* szPrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
					szPrefValue - the value to write to the preference.
		Returns:	true => preference written successfully
					false => otherwise
	
		Purpose:	Write a C string value to the named preference.
	
	**************************************************************************/
	
	bool Preferences::Write(const String& PrefName, const char* szPrefValue)
	{
		// Just a veneer to the proper String-based function.
		String PrefValue(szPrefValue);
		return Write(PrefName, PrefValue);
	}

	/**************************************************************************
	
	>	bool Preferences::Write(const String& PrefName, int PrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
					PrefValue - the value to write to the preference.
		Returns:	true => preference written successfully
					false => otherwise
	
		Purpose:	Write an int value to the named preference.
	
	**************************************************************************/
	
	bool Preferences::Write(const String& PrefName, int PrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Write(PrefLeafName, PrefValue);
	}

	/**************************************************************************
	
	>	bool Preferences::Write(const String& PrefName, bool PrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
					PrefValue - the value to write to the preference.
		Returns:	true => preference written successfully
					false => otherwise
	
		Purpose:	Write a bool value to the named preference.
	
	**************************************************************************/
	
	bool Preferences::Write(const String& PrefName, bool PrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Write(PrefLeafName, PrefValue);
	}

	/**************************************************************************
	
	>	bool Preferences::Write(const String& PrefName, const FileSpec& PrefValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		PrefName - name of the preference, e.g. "Username"
					PrefValue - the value to write to the preference.
		Returns:	true => preference written successfully
					false => otherwise
	
		Purpose:	Write a FileSpec value to the named preference.
	
	**************************************************************************/
	
	bool Preferences::Write(const String& PrefName, const FileSpec& PrefValue)
	{
		String PrefLeafName;
		Preferences* pSection = FindParentSection(PrefName, &PrefLeafName);
		return pSection->GetImp()->Write(PrefLeafName, PrefValue);
	}

	/**************************************************************************
	
	>	int Preferences::ReadListCount(const String& ListPrefName)
	
		Author:		Tim Browse
		Date:		15/7/2002
		Inputs:		ListPrefName - name of the preference to used to store
								   the list of preferences.
		Returns:	Number of items in list (possibly 0)
	
		Purpose:	Find out how many items are in the named list of preferences.
	
	**************************************************************************/
	
	int Preferences::ReadListCount(const String& ListPrefName)
	{
		int cItems = 0;
		vos::String PrefName(ListPrefName);
		PrefName += ".";
		PrefName += kPrefList_Count;
		Read(PrefName, &cItems);
		return cItems;
	}

	/**************************************************************************
	
	>	String Preferences::ReadListItem(const String& ListPrefName, int iItem)
	
		Author:		Tim Browse
		Date:		15/7/2002
		Inputs:		ListPrefName - name of the preference to used to store
								   the list of preferences.
					iItem - 0-based index of the item to read
							(i.e. first item is number 0)
		Returns:	The item's value (or empty, if iItem is not valid).
	
		Purpose:	Read the specified item from the list.  The iItem parameter
					must be within the limit returned by ReadListCount().
	
	**************************************************************************/
	
	String Preferences::ReadListItem(const String& ListPrefName, int iItem)
	{
		String strItem;

		// Construct preference name for this item
		vos::String PrefNameFormat(ListPrefName);
		PrefNameFormat += ".";
		PrefNameFormat += kPrefList_Item;
		vos::String PrefName;
		PrefName.Printf(PrefNameFormat, iItem);

		// Ok, now read the item's value
		Read(PrefName, &strItem);
		return strItem;
	}

	/**************************************************************************
	
	>	bool Preferences::WriteListCount(const String& ListPrefName, int cItems)
	
		Author:		Tim Browse
		Date:		15/7/2002
		Inputs:		ListPrefName - name of the preference to used to store
								   the list of preferences.
					cItems - the number of items in the list
		Returns:	true => write succeeded
					false => otherwise
	
		Purpose:	Specify how many items are in the named list, which is
					saved in the preferences.
	
	**************************************************************************/
	
	bool Preferences::WriteListCount(const String& ListPrefName, int cItems)
	{
		vos::String PrefName(ListPrefName);
		PrefName += ".";
		PrefName += kPrefList_Count;
		return Write(PrefName, cItems);
	}

	/**************************************************************************

	>	bool Preferences::WriteListItem(const String& ListPrefName, 
										int iItem, const String& strItem)


		Author:		Tim Browse
		Date:		15/7/2002
		Inputs:		ListPrefName - name of the preference to used to store
								   the list of preferences.
					iItem - index of the item to write (first item is number 0)
					strItem = the value of the item to save in the preferences.
		Returns:	true => write succeeded
					false => otherwise

		Purpose:	Write a list item to the preferences.

	**************************************************************************/

	bool Preferences::WriteListItem(const String& ListPrefName, 
									int iItem, const String& strItem)
	{
		// Construct preference name for this item
		vos::String PrefNameFormat(ListPrefName);
		PrefNameFormat += ".";
		PrefNameFormat += kPrefList_Item;
		vos::String PrefName;
		PrefName.Printf(PrefNameFormat, iItem);

		// Ok, now write the item's value
		return Write(PrefName, strItem);
	}
	
	/**************************************************************************
	
	>	void Preferences::DeleteAll()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Remove all preferences for this application.

					NB. This function currently does not work, as it could
						be quite dangerous, and would be used rarely if ever
						anyway.
	
	**************************************************************************/
	
	void Preferences::DeleteAll()
	{
		GetPreferences().GetImp()->DeleteAll();
	}

	/**************************************************************************
	
	>	Preferences& GetPreferences()
	
		Author:		Tim Browse
		Date:		15/4/2002
		Returns:	The main preferences object.
	
		Purpose:	Get the main preferences object for the application.

					Use this function whenever you want to work with preferences.
	
					Remember to call Init() before using the preferences system.

	**************************************************************************/
	
	Preferences& GetPreferences()
	{
		return Preferences::GetPreferences();
	}

};

