//
// Socket.cpp - vOS implementation of socket API
//

#include "StdHdr.h"

#include "Socket.h"


#ifdef VOS_MSWINDOWS
#undef ERROR
#include <winsock2.h>
#endif

#include "vos\Errors.h"

#include "Msgs.h"

#include "SocketAddress.h"

namespace vos
{
	//
	// Type of OS-dependent socket handle
	//
#ifdef VOS_MSWINDOWS
	typedef SOCKET SocketHandle;
	const SocketHandle kInvalidSocket = INVALID_SOCKET;
#else
#error vosSocket type needs porting.
#endif

	/**************************************************************************
	
	>	class SocketImp : public Socket
	
		Author:		Tim Browse
		Date:		8/2/2002
	
		Purpose:	Socket implementation.
	
	**************************************************************************/
	
	class SocketImp : public Socket
	{
		public:
			// Constructor
			SocketImp();
			SocketImp(SocketHandle hSocket);
	
			// Destructor
			~SocketImp();

			//
			// General API
			//

			// Create the socket object
			vos::Error SocketImp::Create(ProtocolFamily Family,
										 ProtocolType	Type);

			// Mark the socket as non-blocking
			virtual vos::Error MarkAsNonBlocking();

			// Binary read/write
			virtual vos::Error Read(void* pData, uint cBytes, uint* pcBytes = NULL);
			virtual vos::Error Write(const void* pData, uint cBytes, uint* pcBytes = NULL);

			// Close the socket
			virtual vos::Error Close();

			//
			// Server API
			//

			// Bind the socket to an address and start listening for
			// connections.
			virtual vos::Error StartServer(uint16 Port, int cBackLog = 5);

			// Accept a client connection
			virtual vos::Error AcceptClientConnection(Socket** ppNewConnection,
													  SocketAddress* pClientAddress);


			//
			// Client API
			//

			// Connect the socket to a server
			virtual vos::Error ConnectToServer(SocketAddress ServerAddress);

		private:
			// Init/shutdown the socket system
			static vos::Error Initialise();
			static void Shutdown();

		private:
			// Socket handle
			SocketHandle m_hSocket;

			// One time initialisation
			static bool s_bInitialised;

			// Let the Socket class init/shutdown the SocketImp system.
			friend class Socket;
	};

	/**************************************************************************
	
	>	vos::Error Socket::Initialise()
	
		Author:		Tim Browse
		Date:		19/2/2002
		Returns:	Error
	
		Purpose:	Do socket initialisation, including any platform dependent
					code.
	
	**************************************************************************/

	vos::Error Socket::Initialise()
	{
		return SocketImp::Initialise();
	}

	/**************************************************************************
	
	>	void Socket::Shutdown()
	
		Author:		Tim Browse
		Date:		19/2/2002
	
		Purpose:	Shutdown the socket system
	
	**************************************************************************/

	void Socket::Shutdown()
	{
		SocketImp::Shutdown();
	}

	//
	// Socket system not initialised by default
	//
	bool SocketImp::s_bInitialised = false;

	//
	// Translation arrays
	//
	static int s_ProtocolFamilies[] = 
	{
#ifdef VOS_MSWINDOWS
		// Windows uses address family constant
		AF_INET
#else
		// Unix uses PF constant
		PF_INET
#endif
	};

	static int s_ProtocolTypes[] = 
	{
		SOCK_STREAM,
		SOCK_DGRAM
	};

	//
	// Macro for checking to see if a socket has been initialised
	//
#define CHECK_SOCKET();									\
	if (m_hSocket == kInvalidSocket)					\
	{													\
		COMPLAIN("Socket not initialised!");			\
		return NORMAL_ERROR(Err_Socket_BadSocket,		\
							msgSocket_BadSocket);		\
	}

	/**************************************************************************
	
	>	SocketImp::SocketImp()
	
		Author:		Tim Browse
		Date:		8/2/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	SocketImp::SocketImp() : m_hSocket(kInvalidSocket)
	{
		// TCILB
	}

	/**************************************************************************
	
	>	SocketImp::SocketImp(SocketHandle hSocket)
	
		Author:		Tim Browse
		Date:		15/5/2002
		Inputs:		hSocket - the existing BSD socket
	
		Purpose:	Construct a socket using an existing BSD socket handle.
	
	**************************************************************************/
	
	SocketImp::SocketImp(SocketHandle hSocket) : m_hSocket(hSocket)
	{
		// TCILB
	}

	/**************************************************************************
	
	>	SocketImp::~SocketImp()
	
		Author:		Tim Browse
		Date:		15/5/2002
	
		Purpose:	Destructor - shuts down socket if necessary
	
	**************************************************************************/
	
	SocketImp::~SocketImp()
	{
		if (m_hSocket != kInvalidSocket)
		{
			// Close down socket, and ignore any errors (as we can't
			// propogate them)
			if (!Close())
				ERROR_HANDLED();
		}
	}

	/**************************************************************************

	>	vos::Error SocketImp::Create(ProtocolFamily Family,
									 ProtocolType	 Type)


		Author:		Tim Browse
		Date:		8/2/2002
		Inputs:		Family - the protocol family to use
					Type - the type of protocol to use
		Returns:	Error

		Purpose:	Create a socket to be used with a particular protocol.

	**************************************************************************/

	vos::Error SocketImp::Create(ProtocolFamily Family,
								 ProtocolType	Type)
	{
		// Create the socket
		// TODO: proper range checking/table lookup on protocol family/types.
		m_hSocket = (SocketHandle) socket(s_ProtocolFamilies[Family],
										  s_ProtocolTypes[Type], IPPROTO_IP);

		if (m_hSocket == kInvalidSocket)
		{
			return NORMAL_ERROR(Err_Socket_CreateFailed, msgSocket_CreateFailed);
		}

		return VNOERROR;
	}

	/**************************************************************************
	
	>	vos::Error SocketImp::Initialise()
	
		Author:		Tim Browse
		Date:		19/2/2002
		Returns:	Error
	
		Purpose:	Do socket initialisation, including any platform dependent
					code.
	
	**************************************************************************/
	
	vos::Error SocketImp::Initialise()
	{
#ifdef VOS_MSWINDOWS
		// Windows specific socket init code
		WSADATA	wsaData;

		const WORD kRequiredWinsockVersion = 0x0101;

		if (WSAStartup(kRequiredWinsockVersion, &wsaData) != 0)
		{
			// Failed to initialise winsock
			return NORMAL_ERROR(Err_Socket_InitFailed, msgSocket_CreateFailed);
		}
#else
#error SocketImp::Initialise() needs porting
#endif

		// If we get here, it worked ok
		s_bInitialised = true;
		return VNOERROR;
	}

	/**************************************************************************
	
	>	void SocketImp::Shutdown()
	
		Author:		Tim Browse
		Date:		19/2/2002
	
		Purpose:	Shutdown the socket system
	
	**************************************************************************/
	
	void SocketImp::Shutdown()
	{
#ifdef VOS_MSWINDOWS
		// Windows specific socket shutdown code
		WSACleanup();
#else
#error SocketImp::Shutdown() needs porting
#endif
		s_bInitialised = false;
	}

	/**************************************************************************

	>	vos::Error SocketImp::StartServer(uint16 Port, int cBackLog = 5)


		Author:		Tim Browse
		Date:		8/2/2002
		Inputs:		Port - the port number to bind the server to
					cBackLog - the number of pending connections to queue at once.
		Returns:	Error

		Purpose:	Set up a server socket - bind to a particular port on the
					local interface, and start listening on this socket for 
					client connections.

	**************************************************************************/

	vos::Error SocketImp::StartServer(uint16 Port, int cBackLog)
	{
		// Sanity check
		CHECK_SOCKET();

		// Ok, first bind this socket to the specified port on the
		// local interface, then put the socket into listening mode
		struct sockaddr_in tcpaddr;

		tcpaddr.sin_family = AF_INET;
		tcpaddr.sin_port = htons(Port);
		tcpaddr.sin_addr.s_addr = htonl(INADDR_ANY);

		if ((bind(m_hSocket, (struct sockaddr*) &tcpaddr, sizeof(tcpaddr)) != 0) ||
			(listen(m_hSocket, cBackLog) != 0))
		{
			// Failed to bind the socket to the local port, or to put it into
			// listen mode
			return NORMAL_ERROR(Err_Socket_CreateServerFailed, 
								msgSocket_CreateServerFailed);
		}

		// Ok, socket is set up for server mode
		return VNOERROR;
	}

	/**************************************************************************

	>	vos::Error SocketImp::AcceptClientConnection(Socket** ppNewConnection,
													 SocketAddress* pClientAddress)

		Author:		Tim Browse
		Date:		8/2/2002
		Outputs:	ppNewConnection - the Socket object to use for the new
									  client connection.  The caller should
									  delete this object when it is no longer
									  required.
									  If NULL, then no new connection was
									  available (e.g. if the socket is
									  non-blocking).

					pClientAddress - the address of the client that has just
									 been accepted.
		Returns:	Error

		Purpose:	Accept a new client connection on a server socket.

	**************************************************************************/

	vos::Error SocketImp::AcceptClientConnection(Socket** ppNewConnection,
												 SocketAddress* pClientAddress)
	{
		// Sanity check
		CHECK_SOCKET();

		// Try to accept a connection
		struct sockaddr_in Client;
		int cbAddressSize = sizeof(Client);

		SocketHandle hClientSocket = accept(m_hSocket, 
											(struct sockaddr*) &Client, 
											&cbAddressSize);

		// Did it work?
		if (hClientSocket == INVALID_SOCKET)
		{
			// Check why this failed - it's probably because there are
			// no new connections yet...
			int WSAError = WSAGetLastError();

			if (WSAError == WSAEWOULDBLOCK)
			{
				// Ok, no new connections - this is not classed as an error.
				*ppNewConnection = NULL;
				return VNOERROR;
			}

			// Otherwise, it's an error...
			return NORMAL_ERROR(Err_Socket_AcceptClientFailed,
								msgSocket_AcceptClientFailed);
		}

		// New client is available - make new socket and give it the
		// new handle
		*ppNewConnection = new SocketImp(hClientSocket);
		return VNOERROR;
	}

	/**************************************************************************

	>	vos::Error SocketImp::ConnectToServer(SocketAddress ServerAddress)

		Author:		Tim Browse
		Date:		8/2/2002
		Inputs:		ServerAddress - the server to connect to
		Returns:	Error

		Purpose:	Connect a client socket to a specified server.

	**************************************************************************/

	vos::Error SocketImp::ConnectToServer(SocketAddress ServerAddress)
	{
		// Sanity check
		CHECK_SOCKET();

		// Get the server's IP address (this may involve DNS lookup)
		RawIPAddress RawServerIP;
		if (!ServerAddress.GetAddress(&RawServerIP))
		{
			// Failed to get the raw IP address
			return LASTERROR;
		}

		// Fill in socket address of the server
		struct sockaddr_in Server;
		Server.sin_family = s_ProtocolFamilies[ProtocolFamily_Internet];
		Server.sin_port = htons(ServerAddress.GetPort());
		Server.sin_addr.s_addr = htonl(RawServerIP.m_Address);

		// Ok, try to connect to this server
		if (connect(m_hSocket, (struct sockaddr*) &Server, sizeof(Server)) != 0)
		{
			// Error occurred while connecting to server
			return NORMAL_ERROR(Err_Socket_ConnectToServerFailed,
								msgSocket_ConnectToServerFailed);
		}
		
		// Connection worked
		return VNOERROR;
	}


	/**************************************************************************

	>	vos::Error SocketImp::Read(void* pData, uint cBytes, uint* pcBytes)

		Author:		Tim Browse
		Date:		8/2/2002
		Inputs:		pData - the buffer to read data into
					cBytes - maximum number of bytes to read
		Outputs:	pcBytes - number of bytes actually read (may be NULL)
		Returns:	Error

		Purpose:	Read data from the socket.

	**************************************************************************/

	vos::Error SocketImp::Read(void* pData, uint cBytes, uint* pcBytes)
	{
		// Sanity check
		CHECK_SOCKET();

		int cBytesReceived = recv(m_hSocket, (char*) pData, cBytes, 0);

		if (cBytesReceived == SOCKET_ERROR)
		{
			// Failed to recevie the data (this is an actual error, rather
			// than there not being any data to receive at the moment)
			return NORMAL_ERROR(Err_Socket_ReadFailed, 
								msgSocket_ReadFailed);
		}

		// Ok, tell the caller how much data we received, and we're done
		*pcBytes = cBytesReceived;
		return VNOERROR;
	}

	/**************************************************************************

	>	vos::Error SocketImp::Write(const void* pData, uint cBytes, uint* pcBytes)

		Author:		Tim Browse
		Date:		8/2/2002
		Inputs:		pData - the buffer to read the outgoing data from
					cBytes - number of bytes to write
		Outputs:	pcBytes - number of bytes written (may be NULL)
		Returns:	Error

		Purpose:	Write a block of data to the socket.

	**************************************************************************/

	vos::Error SocketImp::Write(const void* pData, uint cBytes, uint* pcBytes)
	{
		// Sanity check
		CHECK_SOCKET();

		int cBytesSent = send(m_hSocket, (const char*) pData, cBytes, 0);

		if (cBytesSent == SOCKET_ERROR)
		{
			// Failed to send the data
			return NORMAL_ERROR(Err_Socket_WriteFailed, 
								msgSocket_WriteFailed);
		}

		// Ok, tell the caller how much data we sent, and we're done
		*pcBytes = cBytesSent;
		return VNOERROR;
	}

	/**************************************************************************

	>	void SocketImp::Close()

		Author:		Tim Browse
		Date:		8/2/2002
		Returns:	Error

		Purpose:	Close the socket - it is no longer needed.
					(The owner must still delete it.)

	**************************************************************************/

	vos::Error SocketImp::Close()
	{
		// Sanity check
		CHECK_SOCKET();

		// Close the socket down gracefully
		shutdown(m_hSocket, SD_BOTH);
		closesocket(m_hSocket);
		m_hSocket = kInvalidSocket;
		return VNOERROR;
	}

	/**************************************************************************
	
	>	vos::Error SocketImp::MarkAsNonBlocking()
	
		Author:		Tim Browse
		Date:		8/2/2002
	
		Purpose:	Mark this socket as non-blocking.  This means if no
					connections/data are immediately available, the socket
					API calls will return immediately.
	
	**************************************************************************/
	
	vos::Error SocketImp::MarkAsNonBlocking()
	{
		// Sanity check
		CHECK_SOCKET();

		// Ok, make it non-blocking
		u_long Parameter = 1;

		if (ioctlsocket(m_hSocket, FIONBIO, &Parameter) != 0)
		{
			// Error occurred
			return NORMAL_ERROR(Err_Socket_NonBlockFailed, 
								msgSocket_NonBlockFailed);
		}

		// All done
		return VNOERROR;
	}

	/**************************************************************************
	
	>	Socket::~Socket()
	
		Author:		Tim Browse
		Date:		8/2/2002
	
		Purpose:	Virtual Destructor
	
	**************************************************************************/
	
	Socket::~Socket()
	{
		// TDILB
	}

	/**************************************************************************

	>	vos::Error Socket::Create(ProtocolFamily Family,
								  ProtocolType	 Type,
								  Socket**		 ppSocket)


		Author:		Tim Browse
		Date:		8/2/2002
		Inputs:		Family - the protocol family to use
					Type - the type of protocol to use
		Outputs:	The new socket object.
		Returns:	Error

		Purpose:	Create a socket to be used with a particular protocol.

	**************************************************************************/

	vos::Error Socket::Create(ProtocolFamily Family,
							  ProtocolType	 Type,
							  Socket**		 ppSocket)
	{
		// Create the concrete socket object and initialise it
		SocketImp* pNewSocket = new SocketImp;

		if (pNewSocket->Create(Family, Type))
		{
			// Socket creation worked ok
			*ppSocket = pNewSocket;
			return VNOERROR;
		}
		else
		{
			// An error occurred while creating the socket.
			return LASTERROR;
		}
	}


}

