//
// SocketAddress.cpp - vOS implementation of socket address type
//

#include "StdHdr.h"

#ifdef VOS_MSWINDOWS
#include "Win32\Win32_Headers.h"
#include "winsock2.h"
#endif

#include "SocketAddress.h"

#include "Socket.h"
#include "Errors.h"
#include "StringParser.h"
#include "Msgs.h"

namespace vos
{

	/**************************************************************************
	
	>	RawIPAddress::RawIPAddress()
	
		Author:		Tim Browse
		Date:		19/2/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	RawIPAddress::RawIPAddress()
	{
		memset(&m_Field, 0, sizeof(m_Field));
	}

	/**************************************************************************

	>	SocketAddress::SocketAddress()

		Author:		Tim Browse
		Date:		19/2/2002

		Purpose:	Constructor

	**************************************************************************/

	SocketAddress::SocketAddress() : m_Port(0)
	{
		/// TCILB
	}

	/**************************************************************************

	>	SocketAddress::SocketAddress(const vos::String& Address, 
									 uint16 PortNumber = 0)

		Author:		Tim Browse
		Date:		19/2/2002
		Inputs:		Address - IP address; may be dotted IP number, or a DNS name.
							  If a DNS name, it will be resolved when used.
					PortNumber - port number to use on this address

		Purpose:	Constructor

	**************************************************************************/

	SocketAddress::SocketAddress(const vos::String& Address, 
								 uint16 PortNumber) : m_Port(PortNumber)
	{
		SetAddress(Address);
	}

	/**************************************************************************

	>	vos::Error SocketAddress::GetAddress(vos::String* pAddress)

		Author:		Tim Browse
		Date:		19/2/2002
		Returns:	Get the string version of the address, in dotted IP
					format, e.g. "192.137.52.21"

		Purpose:	Get the raw IP address in string form.

	**************************************************************************/

	vos::Error SocketAddress::GetAddress(vos::String* pAddress)
	{
		// Currently, if we have a DNS name, we resolve it whenever the
		// client calls this function.  This policy may change, if required.
		if (!ResolveDNSName())
		{
			// Problem resolving DNS name
			return LASTERROR;
		}

		// Ok, just convert to dotted IP address.
		pAddress->Printf("%d.%d.%d.%d", 
						 m_Address.m_Field[0],
						 m_Address.m_Field[1],
						 m_Address.m_Field[2],
						 m_Address.m_Field[3]);
		return VNOERROR;
	}

	/**************************************************************************

	>	vos::Error SocketAddress::GetAddress(RawIPAddress* pAddress)

		Author:		Tim Browse
		Date:		19/2/2002
		Outputs:	pAddress - the IP address
		Returns:	Error if DNS name could not be resolved
					(If the address was specified as dotted IP, then this
					function always succeeds.

		Purpose:	Prove that Tim is a JERK!

	**************************************************************************/

	vos::Error SocketAddress::GetAddress(RawIPAddress* pAddress)
	{
		// Currently, if we have a DNS name, we resolve it whenever the
		// client calls this function.  This policy may change, if required.
		if (!ResolveDNSName())
		{
			// Problem resolving DNS name
			return LASTERROR;
		}

		*pAddress = m_Address;
		return VNOERROR;
	}

	/**************************************************************************

	>	uint SocketAddress::GetPort() const

		Author:		Tim Browse
		Date:		19/2/2002
		Returns:	The port number

		Purpose:	Get the port number of this address

	**************************************************************************/

	uint SocketAddress::GetPort() const
	{
		return m_Port;
	}

	/**************************************************************************

	>	void SocketAddress::SetAddress(const vos::String& Address)

		Author:		Tim Browse
		Date:		19/2/2002
		Inputs:		Address - the address to use, either dotted IP or a
							  DNS name.

		Purpose:	Set the IP address to use for this object.

	**************************************************************************/

	void SocketAddress::SetAddress(const vos::String& Address)
	{
		if (!DecodeDottedIP(Address, &m_Address))
		{
			// Ok, not a dotted IP address, so we assume it's a DNS name
			m_DNSName = Address;
		}
	}

	/**************************************************************************

	>	void SocketAddress::SetAddress(const RawIPAddress& Address)

		Author:		Tim Browse
		Date:		19/2/2002
		Inputs:		Address - raw IP address to use.

		Purpose:	Set the address using a raw IP address

	**************************************************************************/

	void SocketAddress::SetAddress(const RawIPAddress& Address)
	{
		// Use raw IP address, and ensure DNS name is cleared so we don't
		// use it.
		m_Address = Address;
		m_DNSName.Empty();
	}

	/**************************************************************************

	>	void SocketAddress::SetPort(uint16 PortNumber)

		Author:		Tim Browse
		Date:		19/2/2002
		Inputs:		PortNumber - the port number to use with this address.

		Purpose:	Set the port number for this address.

	**************************************************************************/

	void SocketAddress::SetPort(uint16 PortNumber)
	{
		m_Port = PortNumber;
	}

	/**************************************************************************

	>	bool SocketAddress::DecodeDottedIP(const vos::String& DottedIP,
										   RawIPAddress* pRawIPAddress)


		Author:		Tim Browse
		Date:		19/2/2002
		Inputs:		DottedIP - possible string representation of a dotted
							   IP address.
		Outputs:	pRawIPAddress - the resulting address, if a valid IP
									address was found in the string.
		Returns:	true => valid IP address was decoded
					false => address is not a valid IP address.

		Purpose:	Decode a dotted IP address from a string.

					e.g. "192.34.23.88"

					Leading and trailing whitespace is ignored, as is
					whitespace between numbers and dots (but not between
					digits of same number).

					Address is not valid if:
					
						* any characters other than digits, whitespace or 
						  the period character appear.
						* the address is not of the form a.b.c.d
						* any number exceeds 255

	**************************************************************************/

	bool SocketAddress::DecodeDottedIP(const vos::String& DottedIP,
									   RawIPAddress* pRawIPAddress)
	{
		unsigned long Address = inet_addr(DottedIP.c_str());
		
		if (Address == INADDR_NONE)
		{
			// Address decode failed
			return false;
		}

		// Address decoded ok - convert to host byte order and store
		pRawIPAddress->Set(ntohl(Address));
		return true;
	}

	/**************************************************************************
	
	>	vos::Error ReportDNSFailure(const vos::String& DNSName)
	
		Author:		Tim Browse
		Date:		19/2/2002
		Inputs:		DNSName - the DNS name that could not be resolved.
		Returns:	The error to return to the caller of the caller :-)
	
		Purpose:	Convenience function for reporting DNS resolution errors.

					Registers and returns the appropriate error data.
	
	**************************************************************************/
	
	vos::Error ReportDNSFailure(const vos::String& DNSName)
	{
		String ErrorMessage;
		ErrorMessage.Printf(msgSocket_DNSResolutionFailed,
							DNSName.c_str());

		return NORMAL_ERROR(Err_Socket_DNSResolutionFailed,
							ErrorMessage.c_str());
	}

	/**************************************************************************
	
	>	vos::Error SocketAddress::ResolveDNSName()
	
		Author:		Tim Browse
		Date:		19/2/2002
		Returns:	Error (DNS related)
	
		Purpose:	Resolve this object's DNS name.

					NB. If no DNS name has been passed in, this function 
						completes silently with no error.
	
	**************************************************************************/
	
	vos::Error SocketAddress::ResolveDNSName()
	{
		if (m_DNSName.IsEmpty())
		{
			// We have no DNS name, so ignore the request
			return VNOERROR;
		}

		// Ok resolve name using standard sockets API...
		HOSTENT *pHostEntry = gethostbyname(m_DNSName.c_str());

		if (pHostEntry == NULL)
		{
			// DNS could not resolve this name
			return ReportDNSFailure(m_DNSName);
		}

		if (pHostEntry->h_length != sizeof(uint32))
		{
			// IP Address is unexpected length
			return NORMAL_ERROR(Err_LogicError, "Socket API failure");
		}

		// Ok, count the number of addresses
		int cAddresses = 0;

		while (pHostEntry->h_addr_list[cAddresses] != NULL)
		{
			cAddresses++;
		}

		if (cAddresses > 0)
		{
			// Choose random address
			int iAddress = rand() % cAddresses;
			memcpy(&m_Address, pHostEntry->h_addr_list[iAddress], pHostEntry->h_length);

			m_Address.m_Address = ntohl(m_Address.m_Address);
			return VNOERROR;
		}

		// No addresses returned - what's going on?
		COMPLAIN("gethostbyname() succeeded by returned no addresses!");
		return NORMAL_ERROR(Err_LogicError, "Socket API failure");
	}

}

