//
// String.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_String_h
#define vos_String_h

#include "vos.h"

//
// Resolve naming clash with MacOS Universal Interfaces
//

#if (defined VOS_MACOS) & (defined __STRINGCOMPARE__)
	#undef CompareText
#endif


//
// We use the standard C++ library to do the basic string implementation.
//

#include <string>
#include <stdio.h>
#include <stdarg.h>

#include "ResourceID.h"

namespace vos
{
	class StringParser;

	//
	// This character typedef defines the basic character used for this string class.
	// All characters are referred to as 'tchar' type by this class.
	//
#ifdef VOS_UNICODE
	typedef wchar_t tchar;
#else
	typedef char tchar;
#endif

	/******************************************************************************

	>	class String

		Author:		Tim Browse
		Created:	22/7/99
		Purpose:	Implement a platform-independent string class.

	******************************************************************************/

	class String
	{
		public:
			// Constructor
			String();

			// Copy constructor
			String(const String& Src);

			// Constructor to initialise from a standard C-style string.
			String(const tchar* pString);

			// Constructor to initialise from te first n characters of standard C-style string.
			String(const tchar* pString, uint nChars);

			// Construct a String object containing a repeated character sequence.
			String(tchar ch, uint nRepeat = 1);

			// Construct String objects containing string representations of integers.
			String(int n);
			String(uint n);
			
			// Construct string from a Pascal string
			String(uint8* pPStr);
	
			// Destructor
			~String();

			// Load string from a string resource
			bool Load(const StringResourceID& ID);

		public:
			// Find the length of the string, in characters.
			int GetLength() const
			{
				return m_StringLength;
			}

			// Check to see if a string is empty, i.e. zero characters long.
			bool IsEmpty() const
			{
				return (m_StringLength == 0);
			}

			// Clear the String/make it empty
			void Clear();
			void Empty()
			{
				Clear();
			}

			// Read a character at a particular position (0 based index)
			tchar at(uint nIndex) const
			{
				return m_pData[nIndex];
			}

			// Synonym for above.
			tchar GetAt(uint nIndex) const
			{
				return m_pData[nIndex];
			}

			// Array access operator
			tchar operator[](int nIndex) const
			{
				if (m_pData == NULL)
				{
					return 0;
				}
				else
				{
					return m_pData[nIndex];
				}
			}

			// Set a character at a particular position (0 based index)
			void SetAt(uint nIndex, tchar ch);

			// overloaded assignment
			const String& operator=(const String& Src);
			const String& operator=(tchar Ch);
			const String& operator=(tchar *pString);
			const String& operator=(const tchar* pString);

			// Comparison operators (case sensitive)
			bool operator==(const String& Other) const;
			bool operator==(const tchar *pString) const;
			bool operator!=(const String& Other) const;
			bool operator!=(const tchar *pString) const;

			bool operator<(const String& Other) const;

			// String order comparison:
			//
			// -1 => this < Other
			//  0 => this = Other
			//  1 => this > Other
			int CompareOrder(const String& Other) const;

			// Comparison functions (case insensitive)
			bool CompareText(const String &Other) const;
			bool CompareText(const tchar *pOther) const;

			// Test to see if a string starts with a particular prefix
			bool StartsWith(const String &Prefix) const;
			bool StartsWith(const tchar *szPrefix) const;

			// Test if string ends with a particular suffix
			bool EndsWith(const String &Suffix) const;
			bool EndsWith(const tchar *szSuffix) const;

			// string concatenation
			const String& operator+=(const String& String);
			const String& operator+=(const tchar *pString);
			const String& operator+=(tchar ch);

			friend String operator+(const String& String1, const String& String2);
			friend String operator+(const String& String, tchar Ch);
			friend String operator+(tchar Ch, const String& String);
			friend String operator+(const String& String, const tchar *pString);
			friend String operator+(const tchar *pString, const String& String);

			// Generic append operation, using C-style string as source to append.
			void AppendChars(const tchar *pSrc, int nCharsToCopy);

			// If the specified character is at the end of the string, remove it.
			void RemoveTrailingChar(tchar Ch);

			// Simple sub-string extraction
			String Mid(int nFirst, int nCount) const;
			String Mid(int nFirst) const;
			String Left(int nCount) const;
			String Right(int nCount) const;
			String TruncateRight(int nCount) const;
			void Truncate(int NewLength);

			// upper/lower/reverse conversion
			void MakeUpper();
			void MakeLower();
			void MakeReverse();

			//
			// Searching functions.
			//

			// Look for a single character match
			bool Find(tchar ch, uint *pPosition = NULL) const;
			bool ReverseFind(tchar ch, uint *pPosition = NULL) const;

			// Look for a specific sub-string
			bool Find(const tchar *pSubString, uint *pPosition = NULL) const;

			// Replacing parts of strings
			void Replace(tchar Old, tchar New);
			void Replace(const tchar* szOld, const tchar* szNew);
			void Replace(const String& Old, const String& New);
			void Replace(const tchar* szOld, const String& New);

			// String formatting
			void Printf(const tchar *szFormat, ...);
			void Printf_va(const tchar *szFormat, va_list);

			// Convert to quoted string
			//
			// e.g. Hello "Tim" is converted to "Hello \"Tim\""
			//
			String QuoteString(tchar QuoteChar='"') const;

			// Remove leading and trailing whitespace from the string.
			void Trim();

			// Access the String as a C-style string (read-only access though).
			const tchar* GetData() const { return c_str(); }
			const tchar* c_str() const;

			// Access the String as an array of not less than MinSizeChars characters.
			tchar *LockBuffer(uint32 MinSizeChars);
			void UnlockBuffer(void);

			// Copy the string into a length-prefixed Pascal string (uint8[256])
			const uint8* p_str(uint8* pPStr) const;

			// Handy aliases
			void assign(const tchar *szString) { *this = szString; }
			void assign(const tchar *szString, uint nChars);

		public:
			// Cast operators...
			operator const tchar *() const;

		protected:
			// Helper functions
			void Copy(const tchar *);
			void CopyChars(const tchar *pSrc, int nFirst, int nCharsToCopy);
			void Concat(const tchar *);

			void ReserveCapacity(uint32 MinSizeChars);

		protected:
			// String buffer
			tchar *m_pData;

			// Length of buffer (in tchars)
			int   m_BufferLength;

			// Length of actual string (in tchars)
			int   m_StringLength;

			static const tchar* m_EmptyString;
	};

};// namespace vos

#endif //vos_String_h
