//
// StringParser.h - class to parse String objects in a simple way
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_StringParser_h
#define vos_StringParser_h

#include "String.h"

namespace vos
{
	/**************************************************************************
	
	>	class StringParser : public String
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Parsing string objects to extract strings, numbers, 
					tokens etc.

					Example usage:

					\code

						// Parse a dotted IP address
						String DottedIP("192.134.56.98");
  						StringParser Parser(DottedIP);

						Parser.SkipWS();

						const int kNumIPFields = 4;

						int Value[kNumIPFields];

						int iField;

						for (iField = 0; iField < kNumIPFields; iField++)
						{
							if (!Parser.ReadCardinal(Value[iField]))
								return false;

							if (Value[iField] > 255)
							{
								// Outside 8-bit limit
								return false;
							}

							// Last number doesn't have trailing period character
							if (iField < (kNumIPFields - 1))
							{
								Parser.SkipWS();

								if (!Parser.SkipDelim("."))
									return false;
							}
						}

						Parser.SkipWS();

						// Check there isn't any junk at the end
						if (!Parser.EOS())
							return false;


					\endcode
	
	**************************************************************************/
	
	class StringParser : public String
	{
		public:
			StringParser();
			~StringParser();
			StringParser(const String&);

		public:
			void SetContents(const String&);
			void ResetParsing();

			bool ReadInt(int &outValue);
			bool ReadCardinal(int &outValue);
			bool ReadCardinalString(String &outValue);
			bool ReadDouble(double &outValue);
			bool ReadHex(int &outValue);

			bool ReadString(String &outValue, String Delimiters,
							bool bUseQuotes = false);

			bool ReadQuotedString(String& outValue, tchar EscapeChar = '\\');

			tchar ReadChar();
			tchar NextChar();

			// Always returns true (makes it easier to be part of
			// a conditional expression)
			bool SkipWS();

			bool SkipDelim(String Delimiters);
			bool SkipDelims(String Delimiters);

			bool SkipDelim(const tchar* szDelimiters);
			bool SkipDelims(const tchar* szDelimiters);

			void SkipNChars(int nChars);

			bool EOS();

		protected:
			// The current parse point
			int m_ParsePos;
	};
};


#endif // !vos_StringParser_h

