//
// Time.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_Time_h
#define vos_Time_h

#include <time.h>

#include "vos.h"

// For month/day names
#include "Date.h"

#ifdef VOS_MSWINDOWS
typedef struct _FILETIME FILETIME;
typedef struct _SYSTEMTIME SYSTEMTIME;
#undef GetCurrentTime
#endif

namespace vos
{
	class TimeInfo;
	class TimeSpan;
	class String;

	/**************************************************************************
	
	>	class Time
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Represents a specific point in time, to a resolution of
					one second.

					This is implemented using time_t, but provides a (hopefully)
					less annoying API.

					As well as the usual features, the class allows you to 
					subtract one Time object from another to produce a 
					TimeSpan object.

	**************************************************************************/
	
	class Time
	{
		public:

		public:
			//
			// Constructors
			//
			Time();
			Time(int TheYear, MonthName Month, int DayOfMonth,
				 int Hour, int Minutes, int Seconds);

			Time(int TheYear, int Month, int DayOfMonth,
				 int Hour, int Minutes, int Seconds);
			
			Time(const Time& rhs);
			Time& operator=(const Time& rhs);

			Time(time_t time);

			// Time creation functions
			void Set(int TheYear, MonthName Month, int DayOfMonth,
					 int Hour = 0, int Minutes = 0, int Seconds = 0);

			void Set(int TheYear, int Month, int DayOfMonth,
					 int Hour = 0, int Minutes = 0, int Seconds = 0);

			void Set(time_t time);

			void SetDate(int TheYear, MonthName Month, int DayOfMonth);
			void SetTime(int Hour, int Minutes, int Seconds);

			// Access functions
			int			GetYear() const;
			MonthName	GetMonth() const;
			int			GetDayOfMonth() const;
			DayName		GetDayOfWeek() const;
			int			GetHour() const;
			int			GetMinute() const;
			int			GetSeconds() const;

			static Time GetCurrentTime();

			// Time comparisons
			bool operator==(const Time& rhs);
			bool operator!=(const Time& rhs);

			bool operator>(const Time& rhs);
			bool operator>=(const Time& rhs);
			bool operator<(const Time& rhs);
			bool operator<=(const Time& rhs);

			// Time arithmetic
			Time operator+(const TimeSpan& rhs);
			Time operator-(const TimeSpan& rhs);
			Time& operator+=(const TimeSpan& rhs);
			Time& operator-=(const TimeSpan& rhs);

			TimeSpan operator-(const Time& rhs);

			//
			// Time to string conversion
			//
			// (szFormat is in the same format as used by the
			//  ANSI function strftime() - see your CRT docs
			//  for details)
			//
			String FormatUTCTime(const char* szFormat);
			String FormatLocalTime(const char* szFormat);

			// Locale functions
			static String GetTimeSeparator();

			// Platform dependent routines
#ifdef VOS_MSWINDOWS
			void ToSystemTime(SYSTEMTIME *pSysTime) const;
			void SetFromSystemTime(const SYSTEMTIME& SysTime);

			void ToFileTime(FILETIME *pFileTime) const;
			void SetFromFileTime(const FILETIME& FileTime);
#endif

		private:
			// Implementation
			//
			// Number of seconds elapsed since 
			// midnight (00:00:00), January 1, 1970.
			time_t m_Time;

			// Other time classes can access this field
			friend class TimeInfo;
			friend class TimeSpan;
	};

	/**************************************************************************
	
	>	class TimeInfo
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Provides a simple structure to hold a breakdown of the
					date/time represented by a vos::Time object.

					If you're doing a lot of Time manipulation, it will
					be more efficient than calling GetDay(), GetMonth(), 
					etc. on the same Time object - it does all the calculations
					in one operation.

					e.g.

					\code

						Time MyTime;

						TimeInfo MyInfo(MyTime);

						int DayNumber = MyInfo.m_DayOfMonth;

					\endcode

	**************************************************************************/
	
	class TimeInfo
	{
		public:
			// Constructor
			TimeInfo(const Time& TheTime);

		public:
			// Data members
			int				m_Year;
			MonthName		m_Month;
			int				m_DayOfMonth;
			DayName			m_DayOfWeek;
			int				m_Hour;
			int				m_Minute;
			int				m_Seconds;
	};

	/**************************************************************************
	
	>	class TimeSpan
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Provide a representation for a length of time (to a
					resolution of one second).

					This class is usually obtained when subtracting one
					vos::Time value from another.
	
	**************************************************************************/
	
	class TimeSpan
	{
		public:
			TimeSpan();

		public:
			// Set length of the timespan
			void SetWeeks(int cWeeks);
			void SetDays(int cDays);
			void SetHours(int cHours);
			void SetMinutes(int cMinutes);
			void SetSeconds(int cSeconds);

			// Get length of the timespan
			int GetYears() const;
			int GetWeeks() const;
			int GetDays() const;
			int GetHours() const;
			int GetMinutes() const;
			int GetSeconds() const;

		public:
			// TimeSpan arithmetic
			TimeSpan operator+(const TimeSpan& rhs);
			TimeSpan operator-(const TimeSpan& rhs);
			TimeSpan& operator+=(const TimeSpan& rhs);
			TimeSpan& operator-=(const TimeSpan& rhs);

		private:
			// Implementation - count of seconds
			time_t m_Span;

			// Other time classes can access this field
			friend class Time;
	};

};

#endif // !vos_Time_h
