//
// TimeOfDay.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "TimeOfDay.h"

#include "Time.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/**************************************************************************
	
	>	TimeOfDay::TimeOfDay()
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	TimeOfDay::TimeOfDay() : m_Hours(0), 
							 m_Minutes(0), 
							 m_Seconds(0)
	{
	}

	/**************************************************************************

	>	TimeOfDay::TimeOfDay(uint8 Hours, 
							 uint8 Minutes, 
							 uint8 Seconds)


		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		Hours/Minutes/Seconds - the time of day to initialise this
											Time object with.

		Purpose:	Constructor

	**************************************************************************/

	TimeOfDay::TimeOfDay(uint8 Hours, 
						 uint8 Minutes, 
						 uint8 Seconds) : m_Hours(Hours),
										  m_Minutes(Minutes),
										  m_Seconds(Seconds)
	{
	}

	/**************************************************************************

	>	bool TimeOfDay::Set(uint8 Hours, 
							uint8 Minutes, 
							uint8 Seconds)


		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		Hours/Minutes/Seconds - the time of day to initialise this
											Time object with.

		Purpose:	Set this TimeOfDay object to a particular time of day.

	**************************************************************************/

	bool TimeOfDay::Set(uint8 Hours, 
						uint8 Minutes, 
						uint8 Seconds)
	{
		// Validate arguments
		if ((Hours > 23) ||	(Minutes > 59) || (Seconds > 59))
		{
			// Illegal (out of range) time arguments
			vosERROR();
			return false;
		}

		m_Hours = Hours;
		m_Minutes = Minutes;
		m_Seconds = Seconds;

		// All ok
		return true;
	}

	/**************************************************************************
	
	>	void TimeOfDay::AddHours(int nHours)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		nHours - number of hours to add (may be negative)
	
		Purpose:	Add a number of hours to the TimeOfDay object.
	
	**************************************************************************/
	
	void TimeOfDay::AddHours(int nHours)
	{
		m_Hours += (nHours % 24);
		m_Hours %= 24;

		// Time should be self-consistent
		ASSERT(IsValid());
	}

	/**************************************************************************
	
	>	void TimeOfDay::AddMinutes(int nMinutes)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		nMinutes - number of minutes to add (may be negative)
	
		Purpose:	Add a number of minutes to the TimeOfDay object.
	
	**************************************************************************/
	
	void TimeOfDay::AddMinutes(int nMinutes)
	{
		while (nMinutes > 0)
		{
			if (nMinutes >= 60)
			{
				AddHours(1);
				nMinutes -= 60;
			}
			else
			{
				m_Minutes += nMinutes;
				nMinutes = 0;

				// Check for wrap
				if (m_Minutes >= 60)
				{
					AddHours(1);
					m_Minutes -= 60;

					// Minutes should now be legal
					ASSERT(m_Minutes < 60);
				}
			}
		}

		// Time should be self-consistent
		ASSERT(IsValid());
	}

	/**************************************************************************
	
	>	void TimeOfDay::AddSeconds(int nSeconds)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		nSeconds - number of seconds to add (may be negative)
	
		Purpose:	Add a number of seconds to the TimeOfDay object.
	
	**************************************************************************/
	
	void TimeOfDay::AddSeconds(int nSeconds)
	{
		while (nSeconds > 0)
		{
			if (nSeconds >= 60)
			{
				AddMinutes(1);
				nSeconds -= 60;
			}
			else
			{
				m_Seconds += nSeconds;
				nSeconds = 0;

				// Check for wrap
				if (m_Seconds >= 60)
				{
					AddMinutes(1);
					m_Seconds -= 60;

					// Seconds should now be legal
					ASSERT(m_Seconds < 60);
				}
			}
		}

		// Time should be self-consistent
		ASSERT(IsValid());
	}

	bool TimeOfDay::operator==(const TimeOfDay& rhs) const
	{
		return (m_Hours == rhs.m_Hours) &&
			   (m_Minutes == rhs.m_Minutes) &&
			   (m_Seconds == rhs.m_Seconds);
	}

	bool TimeOfDay::operator!=(const TimeOfDay& rhs) const
	{
		return (m_Hours != rhs.m_Hours) ||
			   (m_Minutes != rhs.m_Minutes) ||
			   (m_Seconds != rhs.m_Seconds);
	}

	bool TimeOfDay::operator<(const TimeOfDay& rhs) const
	{
		return MakeTimeValue() < rhs.MakeTimeValue();
	}

	bool TimeOfDay::operator<=(const TimeOfDay& rhs) const
	{
		return MakeTimeValue() <= rhs.MakeTimeValue();
	}

	bool TimeOfDay::operator>=(const TimeOfDay& rhs) const
	{
		return MakeTimeValue() >= rhs.MakeTimeValue();
	}

	bool TimeOfDay::operator>(const TimeOfDay& rhs) const
	{
		return MakeTimeValue() > rhs.MakeTimeValue();
	}

	/**************************************************************************
	
	>	uint32 TimeOfDay::MakeTimeValue() const
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	The time value.
	
		Purpose:	Make a time value from this object that can be used
					to provide ordering.  Used by the inequality operators.
	
	**************************************************************************/
	
	uint32 TimeOfDay::MakeTimeValue() const
	{
		return ((((uint32) m_Hours) << 16) |
				(((uint32) m_Minutes) << 8) |
				((uint32) m_Seconds));
	}

	/**************************************************************************
	
	>	bool TimeOfDay::IsValid() const
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	true => object is valid
					false => otherwise
	
		Purpose:	Check the object to make sure the hours/minutes/seconds are
					within valid ranges.
	
	**************************************************************************/
	
	bool TimeOfDay::IsValid() const
	{
		return (m_Hours < 24) && (m_Minutes < 60) && (m_Minutes < 60);
	}

	/**************************************************************************
	
	>	TimeOfDay TimeOfDay::GetCurrentTime()
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	The current time of day.
	
		Purpose:	Get the current time of day.
	
	**************************************************************************/
	
	TimeOfDay TimeOfDay::GetCurrentTime()
	{
		// Get current time/date
		Time Now = Time::GetCurrentTime();

		// Decode into fields
		TimeInfo Info(Now);

		// Construct time representing today and return to caller
		TimeOfDay CurrentTime(Info.m_Hour, Info.m_Minute, Info.m_Seconds);
		return CurrentTime;
	}

};

