//
// VFS.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_VFS_h
#define vos_VFS_h

#include "vos.h"

#include "FileSpec.h"
#include "File.h"
#include "FileMatches.h"
#include "ErrorManager.h"
#include "Lists.h"
#include "RefCount.h"

namespace vos
{
	class VFSFactory;

	/**************************************************************************

		class VFS : public ListItem, public RefCount

		Author:		Tim Browse

		Date:		13/2/00

		Purpose:	The base class for Virtual Filing System objects.

					VFS objects provide a unified interface to a hierarchical
					collection of files.  The intention is that you can 
					perform the same operations on a set of files and folders
					using the same code regardless of whether those files
					and folders and stored natively in the host filing system
					or in some kind of image filing system (such as a .zip file).

					Files can be accessed using the standard vOS File class
					interface, and files are specified using the standard
					vOS FileSpec class.

	**************************************************************************/

	class VFS : public ListItem, public RefCount
	{
		public:
			// Use this function to open a VFS object (virtual constructor)
			static Error Open(const FileSpec& VFSObject,
							  VFS** ppVFS);

		public:
			// Open a file within the VFS
			virtual Error OpenFile(const FileSpec& RelativeFS, 
								   FileMode Mode,
								   File** ppFile) = 0;

			// Enumerate the catalogue of a VFS
			virtual FileMatches* EnumFiles(const FileSpec& RelativeFS) = 0;

			// Close the VFS when you have finished with it - do not 'delete' it.
			virtual void Close();

		public:
			// Registering VFS factories
			static void RegisterVFSFactory(VFSFactory* pFactory);

		protected:
			// Do not construct this object directly.
			VFS();

			// Do not delete this object directly - call Close() when done with it.
			~VFS();

		protected:
			// List of registered VFS factories
			static List m_Factories;
	};
};

#endif // !vos_VFS_h
