//
// ZipLayout.h
//

#ifndef vos_ZipLayout_h
#define vos_ZipLayout_h

#include "vos.h"
#include "Errors.h"

namespace vos
{
	/*************************************************************************/
	/*! \file

		This header file defines the various file structures used in ZIP
		format files, including file descriptors, indices, etc.

		See the official PKWARE ZIP file format document:
		
			vOS\docs\ZIP File Format.txt
		  
		...for details.

	**************************************************************************/

	//
	// ZIP file signatures (essentially magic numbers)
	//
	enum 
	{
		kLocalFileHeaderSignature	= 0x04034b50,
		kCentralFileHeaderSignature = 0x02014b50,
		kEndOfCentralDirSignature	= 0x06054b50
	};

	//
	// ZIP structures must be packed (no padding)
	//

#include "StructPackingOn.h"

	/**************************************************************************

	>	class RawZipDataDescriptor

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Provide a subset of information about a particular file
					in a ZIP file.

					This data is either provided in the file header before
					the file (see ZipLocalFileHeader below), or is provided
					after the file data.  The second option is provided so that
					ZIP files can be created without the need for seeking within
					the ZIP file.

	**************************************************************************/

	class RawZipDataDescriptor
	{
		public:
			// CRC for this file's data.
			uint32 m_CRC32;

			// Size of the file, both compressed and uncompressed.
			uint32 m_CompressedSize;
			uint32 m_UncompressedSize;

		public:
			// If host endian-ness is not little-endian like ZIP files, then
			// swap the fields.
			void EndianNormalise();
	};


	
	/**************************************************************************

	>	class RawZipLocalFileHeader

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Provide full information about a particular file in a 
					ZIP file.  See the ZIP file format document for more 
					details.

	**************************************************************************/

	class RawZipLocalFileHeader
	{
		public:
			// Signature number - must be kLocalFileHeaderSignature
			uint32 m_Signature;

			// Version number of PKZIP needed to extract this file.
			uint16 m_VersionNeededToExtract;

			// General purpose bit flags.
			uint16 m_GPBits;

			// The compression used to store this file, if any.
			uint16 m_CompressionMethod;

			// Modification time/data for this file.
			uint16 m_ModifiedFileTime;
			uint16 m_ModifiedFileDate;

			// CRC for this file's data.
			uint32 m_CRC32;

			// Size of the file, both compressed and uncompressed.
			uint32 m_CompressedSize;
			uint32 m_UncompressedSize;

			// Length of the filename field/extra field, both of which follow
			// this structure.
			uint16 m_FilenameLength;
			uint16 m_ExtraFieldLength;

		public:
			// If host endian-ness is not little-endian like ZIP files, then
			// swap the fields.  Called automatically by Load().
			void EndianNormalise();
	};

	
	/**************************************************************************

	>	class RawZipFileHeader

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Provide information about a file in a ZIP file, with
					some extra information that is only stored in the 
					central directory of the ZIP file.

	**************************************************************************/

	class RawZipFileHeader
	{
		public:
			// Signature number - must be kCentralFileHeaderSignature
			uint32 m_Signature;

			// Version number of PKZIP used to make this file
			uint16 m_VersionMadeBy;

			// Version number of PKZIP needed to extract this file.
			uint16 m_VersionNeededToExtract;

			// General purpose bit flags.
			uint16 m_GPBits;

			// The compression used to store this file, if any.
			uint16 m_CompressionMethod;

			// Modification time/data for this file.
			uint16 m_ModifiedFileTime;
			uint16 m_ModifiedFileDate;

			// CRC for this file's data.
			uint32 m_CRC32;

			// Size of the file, both compressed and uncompressed.
			uint32 m_CompressedSize;
			uint32 m_UncompressedSize;

			// Length of the filename field/extra field, both of which follow
			// this structure.
			uint16 m_FilenameLength;
			uint16 m_ExtraFieldLength;

			// Extra file info contained in the central directory records
			uint16 m_FileCommentLength;
			uint16 m_DiskNumberStart;
			uint16 m_InternalFileAttributes;
			uint32 m_ExternalFileAttributes;
			uint32 m_LocalHeaderRelativeOffset;

		public:
			// If host endian-ness is not little-endian like ZIP files, then
			// swap the fields.  Called automatically by Load().
			void EndianNormalise();
	};

	/**************************************************************************

	>	class RawZipEndOfCentralDir

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Mark the end of the central directory in the ZIP file.

					This is a special type of central directory record - it
					has its own signature number, and signifies that there
					are no more records in the file.

	**************************************************************************/

	class RawZipEndOfCentralDir
	{
		public:
			// Signature number - must be kEndOfCentralDirSignature
			uint32 m_Signature;

			// Number of disk that this ZIP file is on (for multi-part ZIPs)
			uint16 m_DiskNumber;

			// Number of the disk where the central directory starts.
			uint16 m_CentralDirDiskNumber;

			// Number of central directory entries in this ZIP file (for multi-part ZIPs).
			uint16 m_NumCentralDirEntries;

			// Total number of entries in the central directory.
			// (i.e. the total number of files in the zip file)
			uint16 m_TotalNumCentralDirEntries;

			// Size (in bytes) of the entire central directory
			uint32 m_CentralDirSize;

			// Offset to the start of the central directory 
			uint32 m_CentralDirOffset;

			// Length of the zip file comment field that follows this structure.
			uint16 m_ZipFileCommentLength;

		public:
			// If host endian-ness is not little-endian like ZIP files, then
			// swap the fields.  Called automatically by Load().
			void EndianNormalise();
	};

#include "StructPackingOff.h"


	/**************************************************************************

	>	class ZipDataDescriptor : public RawZipDataDescriptor

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Smart version of RawZipDataDescriptor, that knows how
					to load itself from a ZIP file.

	**************************************************************************/

	class ZipDataDescriptor : public RawZipDataDescriptor
	{
		public:
			// Load (and endian-swap, if necessary) the structure from a file.
			vos::Error Load(vos::File* pFile);
	};

	/**************************************************************************

	>	class ZipLocalFileHeader : public RawZipLocalFileHeader

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Smart version of RawZipLocalFileHeader, that knows how
					to load itself from a ZIP file, and how to write itself
					out to a ZIP file correctly.

	**************************************************************************/

	class ZipLocalFileHeader : public RawZipLocalFileHeader
	{
		public:
			ZipLocalFileHeader();
			~ZipLocalFileHeader();

			void SetFilename(const vos::String& Filename);

			void CleanUp();

		public:
			// Variable-length field data
			char *m_pFilename;
			byte *m_pExtraField;

		public:
			// Load (and endian-swap, if necessary) the structure from a file.
			vos::Error Load(vos::File* pFile);

			// Write structure to a file
			//
			// (NB. only works on little-endian hosts at the moment)
			vos::Error Write(vos::File* pFile);
	};

	/**************************************************************************

	>	class ZipFileHeader : public RawZipFileHeader

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Smart version of RawZipFileHeader, that knows how
					to load itself from a ZIP file, and how to write itself
					out to a ZIP file correctly.

	**************************************************************************/

	class ZipFileHeader : public RawZipFileHeader
	{
		public:
			ZipFileHeader();
			~ZipFileHeader();

			void Init(const ZipLocalFileHeader&);

			void SetFilename(const vos::String& Filename);

			void CleanUp();

		public:
			// Variable-length field data
			char *m_pFilename;
			byte *m_pExtraField;
			char *m_pFileComment;

		public:
			// Load (and endian-swap, if necessary) the structure from a file.
			vos::Error Load(vos::File* pFile);

			// Write structure to a file
			//
			// (NB. only works on little-endian hosts at the moment)
			vos::Error Write(vos::File* pFile);
	};

	/**************************************************************************

	>	class ZipEndOfCentralDir : public RawZipEndOfCentralDir

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Smart version of RawZipEndOfCentralDir, that knows how
					to load itself from a ZIP file, and how to write itself
					out to a ZIP file correctly.

	**************************************************************************/

	class ZipEndOfCentralDir : public RawZipEndOfCentralDir
	{
		public:
			ZipEndOfCentralDir();
			~ZipEndOfCentralDir();

			void CleanUp();

		public:
			// Variable-length field data
			char *m_pZipFileComment;

		public:
			// Load (and endian-swap, if necessary) the structure from a file.
			vos::Error Load(vos::File* pFile);

			// Write structure to a file
			//
			// (NB. only works on little-endian hosts at the moment)
			vos::Error Write(vos::File* pFile);
	};

}

#endif // !vos_ZipLayout_h
