//
// ZipReader.cpp - class to allow reading of ZIP files.
//

#include "StdHdr.h"

#include "ZipReader.h"

#include "ZipCatalogueReader.h"
#include "ZipLayout.h"
#include "FileInAFile.h"

#include "DiskFile.h"
#include "Errors.h"
#include "Msgs.h"

namespace vos
{
	/**************************************************************************

	>	ZipReader::ZipReader(const vos::FileSpec& fsZipFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		fsZipFile - the ZIP file to open.

		Purpose:	Constructor.  Attaches the object to a ZIP file on disk.

	**************************************************************************/

	ZipReader::ZipReader(const vos::FileSpec& fsZipFile) : m_fsZipFile(fsZipFile),
														   m_pZipFile(NULL)
	{
		// TCILB
	}

	/**************************************************************************

	>	ZipReader::~ZipReader()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Destructor.

	**************************************************************************/

	ZipReader::~ZipReader()
	{
		Close();
	}

	/**************************************************************************

	>	vos::Error ZipReader::Open()

		Author:		Tim Browse
		Date:		10/06/2001
		Returns:	Error

		Purpose:	Attempt to open the ZIP disk file for reading.

	**************************************************************************/

	vos::Error ZipReader::Open()
	{
		// Try to open the zip disk file for reading
		vos::DiskFile* pDiskFile = new vos::DiskFile;

		if (pDiskFile->Open(m_fsZipFile, vos::fam_Read))
		{
			// ZIP opened ok - try to read catalogue into RAM.
			m_pZipFile = pDiskFile;
			if (ReadCatalogue())
			{
				return VNOERROR;
			}
			else
			{
				// Failed to read the catalogue - bail out
				Close();
				return LASTERROR;
			}
		}
		else
		{
			// Error opening zip file
			return LASTERROR;
		}
	}

	/**************************************************************************

	>	void ZipReader::Close()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Close the ZIP file we are reading data from.

					Destructor automatically closes the file if it is left open.

	**************************************************************************/

	void ZipReader::Close()
	{
		if (m_pZipFile != NULL)
		{
			// Close the zipfile
			m_pZipFile->Close();
			delete m_pZipFile;
			m_pZipFile = NULL;

			// Clean up the catalogue.
			delete m_pCatalogue;
			m_pCatalogue = NULL;
		}
	}

	/**************************************************************************

	>	uint ZipReader::GetNumFiles() const

		Author:		Tim Browse
		Date:		10/06/2001
		Returns:	Number of files in the ZIP file.

		Purpose:	Find out how many files are stored in this ZIP file.

	**************************************************************************/

	uint ZipReader::GetNumFiles() const
	{
		return m_pCatalogue->GetNumFiles();
	}

	/**************************************************************************

	>	void ZipReader::GetFilename(uint iFile, vos::FileSpec* pFileSpec)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		iFile - index of the file to query
		Outputs:	pFileSpec - filename of the specified file.

		Purpose:	Find the filename of a specific file in the ZIP file.

					Used when you are iterating through the ZIP catalogue
					rather than just searching using OpenFile() and a
					FileSpec.

	**************************************************************************/

	void ZipReader::GetFilename(uint iFile, vos::FileSpec* pFileSpec)
	{
		m_pCatalogue->GetFilename(iFile, pFileSpec);
	}

	/**************************************************************************

	>	bool ZipReader::FindFile(const vos::FileSpec& fsFile, uint* piFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		fsFile - the filename to search for in the ZIP file.
		Outputs:	piFile - index of the file in question (if found)
		Returns:	true => file found;
					false => file not found.

		Purpose:	Search for a named file in the ZIP file, and get the
					catalogue index of the file (which can later be used to
					get the filename back, or open the file, etc).

	**************************************************************************/

	bool ZipReader::FindFile(const vos::FileSpec& fsFile, uint* piFile)
	{
		return m_pCatalogue->FindFile(fsFile, piFile);
	}

	/**************************************************************************

	>	vos::Error ZipReader::OpenFile(uint iFile, vos::File** ppFile, uint32* pCRC32)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		iFile - catalogue index of the file to open.
		Outputs:	ppFile - the new file object, which can be used to read 
							 data from the requested file if the search was
							 successful.

							 NB. The caller must dispose of this File object
								 when they have finished with it.

					pCRC32 - the CRC32 value for the file in question
							 (pass in NULL if you don't want the CRC value).
		Returns:	Error

		Purpose:	Open the specified file from the ZIP file, and obtain a
					File object which can be used to read the file's data.

	**************************************************************************/

	vos::Error ZipReader::OpenFile(uint iFile, vos::File** ppFile, uint32* pCRC32)
	{
		if (m_pCatalogue == NULL)
		{
			// No ZIP catalogue available!
			return ERROR(vos::Err_File_NotOpen, msgFile_Read_NotOpen);
		}

		// Ok, find the specified file
		uint FileOffset, FileLength;
		if (m_pCatalogue->GetFile(iFile, &FileOffset, &FileLength, pCRC32))
		{
			// Found the file - set up a FileInAFile to access it...
			return OpenFile(FileOffset, FileLength, ppFile);
		}
		else
		{
			// File not found - should not happen with indexed access!
			COMPLAIN("Invalid catalogue index used in ZipReader::OpenFile()");
			return ERROR(vos::Err_File_NotFound, msgZIP_FileNotInZIP);
		}
	}

	/**************************************************************************

	>	vos::Error ZipReader::OpenFile(const vos::FileSpec& fsFile, 
									   vos::File** ppFile, 
									   uint32* pCRC32)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		fsFile - the name of the file to open in the ZIP file.
		Outputs:	ppFile - the new file object, which can be used to read 
							 data from the requested file if the search was
							 successful.

							 NB. The caller must dispose of this File object
								 when they have finished with it.

					pCRC32 - the CRC32 value for the file in question
							 (pass in NULL if you don't want the CRC value).
		Returns:	Error

		Purpose:	Open the specified file from the ZIP file, and obtain a
					File object which can be used to read the file's data.

	**************************************************************************/

	vos::Error ZipReader::OpenFile(const vos::FileSpec& fsFile, 
								   vos::File** ppFile, 
								   uint32* pCRC32)
	{
		if (m_pCatalogue == NULL)
		{
			// No ZIP catalogue available!
			return ERROR(vos::Err_File_NotOpen, msgFile_Read_NotOpen);
		}

		// Ok, find the specified file
		uint FileOffset, FileLength;
		if (m_pCatalogue->FindFile(fsFile, &FileOffset, &FileLength, pCRC32))
		{
			// Found the file - set up a FileInAFile to access it...
			return OpenFile(FileOffset, FileLength, ppFile);
		}
		else
		{
			// File not found
			return ERROR(vos::Err_File_NotFound, msgZIP_FileNotInZIP);
		}
	}

	/**************************************************************************

	>	vos::Error ZipReader::OpenFile(uint FileOffset, uint FileLength, 
									   vos::File** ppFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		FileOffset - the offset of the file to open, within the
								 ZIP file (in bytes).
					FileLength - length of the file to open (in bytes).
		Outputs:	ppFile - the File object to use to read the specified
							 file's data.

							 NB. The caller must dispose of this File object
								 when they have finished with it.
		Returns:	Error

		Purpose:	Low-level file opening routine, for accessing the data in
					a file contained in a ZIP file.

					This object is used by the other higher-level OpenFile()
					functions to do the nitty gritty work.

	**************************************************************************/

	vos::Error ZipReader::OpenFile(uint FileOffset, uint FileLength, vos::File** ppFile)
	{
		// Set up a FileInAFile to access the file...

		// First we create a disk file to access the ZIP file
		vos::DiskFile* pRawDiskFile = new vos::DiskFile;
		pRawDiskFile->Create(m_fsZipFile);

		// Now we create a FileInAFile
		vos::FileInAFile* pSubFile = new FileInAFile;

		// After this function call, the DiskFile object is now the
		// responsibility of the FileInAFile object.
		pSubFile->Create(pRawDiskFile, FileOffset, FileLength);

		// And now we try to open this file object.
		if (pSubFile->Open())
		{
			// Success
			*ppFile = pSubFile;
			return VNOERROR;
		}
		else
		{
			// Error opening file...
			delete pSubFile;
			return LASTERROR;
		}
	}


	/**************************************************************************

	>	vos::Error ZipReader::ReadCatalogue()

		Author:		Tim Browse
		Date:		10/06/2001
		Returns:	Error

		Purpose:	Load the ZIP catalogue and make a compact catalogue 
					containing the information we need.

					The catalogue is read as soon as the ZIP file is first
					opened for reading (and hence before we try to read any
					of the files inside the ZIP archive).

	**************************************************************************/

	vos::Error ZipReader::ReadCatalogue()
	{
		uint32 cbZIPFileSize = m_fsZipFile.GetFileSize();

		// Have to check for 0, as GetFileSize() does not return an error
		// code - gah!
		if (cbZIPFileSize == 0)
		{
			return ERROR(vos::Err_File_InvalidFileFormat, msgZIP_Invalid);
		}

		// Ok, we need to find the end of central directory record, which
		// has to be done by seeking backwards from the end of the ZIP
		// file (ok, a slightly dumb bit of design in the ZIP format there...)
		uint cFilesInZIP;
		if (!FindCentralDirectory(cbZIPFileSize, &cFilesInZIP))
		{
			// Unable to find the central directory - bail out!
			return ERROR(vos::Err_File_InvalidFileFormat, msgZIP_Invalid);
		}

		// Ok, read in each entry in the central directory, and create a 
		// compact catalogue.
		m_pCatalogue = new ZipCatalogueReader();

		if (!m_pCatalogue->AllocateCatalogue(cFilesInZIP))
		{
			// Out of memory
			return ERROR_NOMEMORY();
		}

		ZipFileHeader FileHeader;

		for (int iFile = 0; iFile < cFilesInZIP; iFile++)
		{
			// Read in the next file info entry from the central directory
			if (!FileHeader.Load(m_pZipFile))
			{
				// Failed to load catalogue entry.
				delete m_pCatalogue;
				m_pCatalogue = NULL;
				return LASTERROR;
			}

			// We only support uncompressed files at the moment
			COMPLAIN_IF(FileHeader.m_UncompressedSize != 
						FileHeader.m_CompressedSize,
						"Compressed files not supported in ZIP files yet!");

			// Add this file's details to our catalogue
			m_pCatalogue->AddFile(FileHeader.m_pFilename, FileHeader);
		}

		// Ok, the catalogue has been read successfully.
		return VNOERROR;
	}

	
	/**************************************************************************

	>	vos::Error ZipReader::FindCentralDirectory(uint32 cbFileSize, uint* pNumFiles)

		Author:		Tim Browse
		Date:		07/06/2001
		Inputs:		cbFileSize - size of the ZIP file, in bytes.
		Outputs:	pNumFiles - number of files in the ZIP file.
		Returns:	No error => the file pointer is currently at the start of
								the central directory.

					File-related errors
							 => unable to find the ZIP file's central directory.

		Purpose:	Seek to the start of the ZIP file's central directory.

	**************************************************************************/

#ifndef max
#define max(x, y) ((x) > (y)) ? (x) : (y)
#endif

#ifndef min
#define min(x, y) ((x) < (y)) ? (x) : (y)
#endif

	vos::Error ZipReader::FindCentralDirectory(uint32 cbFileSize, uint* pNumFiles)
	{
		// Sanity checks.
		COMPLAIN_IF(m_pZipFile == NULL, "FindCentralDirectory() - zip file has not been created!");
		COMPLAIN_IF(!m_pZipFile->IsOpen(), "FindCentralDirectory() - zip file has not been opened!");

		// Seek to end of file, and start reading in 1k blocks to find the 
		// end of central directory record.
		const int kBlockSize = 1024;

		// NB. 1k is large enough to find the end-marker in 99.9%, as the
		//     only uncertainty is the length of the zipfile-comment field,
		//	   and most zip files have no comment anyway.  It's small enough
		//	   to present no read I/O overhead.
		//
		//	   In other words, most ZIP files on the planet will be recognised on
		//	   the first read op, so don't stress about performance issues.

		// Work out how far back we will seek before giving up...
		const uint32 kMaxZipFileCommentSize = 0xFFFF;
		const uint32 kMaxFinalRecordSize = sizeof(ZipEndOfCentralDir) + kMaxZipFileCommentSize;

		int MinFileSeekPos = cbFileSize - kMaxFinalRecordSize;
		MinFileSeekPos = max(0, MinFileSeekPos);

		// Work out where to start looking...
		int SeekPos = cbFileSize - kBlockSize;
		SeekPos = max(0, SeekPos);

		byte ZipBlock[kBlockSize];

		int cbBlockSize = min(kBlockSize, cbFileSize);

		while (SeekPos >= MinFileSeekPos)
		{
			// Read the 1k block
			if (!m_pZipFile->SeekFromStart(SeekPos) ||
				!m_pZipFile->Read(ZipBlock, cbBlockSize))
			{
				// Unable to read the block...
				return LASTERROR;
			}

			// Look for signature - if found, then seek to the start of the final central 
			// directory record, and return success.
			uint iOffset;
			if (FindSignature(ZipBlock, cbBlockSize, kEndOfCentralDirSignature, &iOffset))
			{
				// Ok, try to read the final record.
				ZipEndOfCentralDir FinalRecord;

				if (!m_pZipFile->SeekFromStart(SeekPos + iOffset) ||
					!FinalRecord.Load(m_pZipFile))
				{
					// Central directory record could not be read
					return LASTERROR;
				}

				// Let the caller know how many files are in this ZIP file.
				*pNumFiles = FinalRecord.m_TotalNumCentralDirEntries;

				// Ok, now we know where the central record starts - seek to it.
				return m_pZipFile->SeekFromStart(FinalRecord.m_CentralDirOffset);
			}

			// Signature not found - move back 1k and try again.
			// (Actually we overlap by 4 bytes, to cope with the signature
			// number straddling our block boundary.)
			SeekPos -= (kBlockSize - 4);
		}

		// We've gone back past our seek limit => this is not a ZIP file!
		return ERROR(vos::Err_File_InvalidFileFormat, msgZIP_Invalid);
	}

	
	/**************************************************************************

	>	bool ZipReader::FindSignature(byte *pBlock, uint cbBlockSize, 
									  uint32 Signature, 
									  uint* piOffset)

		Author:		Tim Browse
		Date:		07/06/2001
		Inputs:		pBlock - the block of (little-endian) data to search.
					cbBlockSize - size (in bytes) of data pointed to by pBlock.
					Signature - the 32-bit word to search for.
		Outputs:	piOffset - offset into the block where the Signature word
							   was found (if it was).
		Returns:	true => Signature found; offset in piOffset
					false => otherwise

		Purpose:	Search a block of bytes for a 4-byte signature.  The word
					is searched for on a byte-boundary granularity.
		
					NB. The data is assumed to be in little-endian format, so 
						the signature is endian-swapped if necessary before the 
						search.

	**************************************************************************/

	bool ZipReader::FindSignature(byte *pBlock, uint cbBlockSize, 
								  uint32 Signature, 
								  uint* piOffset)
	{
		// Set up a little-endian version of the signature.
		byte SigBytes[4];
		SigBytes[0] = Signature & 0xFF;
		Signature >>= 8;
		SigBytes[1] = Signature & 0xFF;
		Signature >>= 8;
		SigBytes[2] = Signature & 0xFF;
		Signature >>= 8;
		SigBytes[3] = Signature & 0xFF;

		// Ok, search block for the signature
		//
		// NB. We subtract 4 from block size in case signature straddles block boundary.
		//
		for (int i = 0; i <= (cbBlockSize - 4); i++)
		{
			if (pBlock[i] == SigBytes[0])
			{
				// Possible match - check it
				if ((pBlock[i+1] == SigBytes[1]) &&
					(pBlock[i+2] == SigBytes[2]) &&
					(pBlock[i+3] == SigBytes[3]))
				{
					// Signature found
					*piOffset = i;
					return true;
				}
			}
		}

		// Signature not found
		return false;
	}

}
