//
// ZipReader.h - class to allow reading of ZIP files.
//

#ifndef util_ZipReader_h
#define util_ZipReader_h

#include "FileSpec.h"

namespace vos
{
	typedef int ZipHandle;

	class ZipCatalogueReader;
	
	/**************************************************************************

	>	class ZipReader

		Author:		Tim Browse

		Date:		03/06/2001

		Purpose:	Provide a way of accessing files contained in a standard
					ZIP file.

					Compressed files are not yet supported.

					Once a file has been opened, you can delete the ZipReader
					object if you no longer need it; a separate File object
					is used to access the embedded file.

					Example code:
					
					\code

						// Open the ZIP file...
						vos::FileSpec FS("mydata.zip");
						ZipReader ZIP(FS);

						// Open each file and read the first 8 bytes of data
						// from each file
						if (ZIP.Open())
						{
							int cFiles = ZIP.GetNumFiles();

							for (int iFile = 0; iFile < cFiles; iFile++)
							{
								vos::File* pFile;

								// Open the n'th file
								if (ZIP.OpenFile(iFile, &pFile))
								{
									// Ok, read the first 8 bytes from this file
									byte Buf[8];
									pFile->Read(Buf, 8);

									// Close the file and delete it
									pFile->Close();
									delete pFile;
								}
							}
						}

						// Find a particular file, and read the first 4 bytes.
						vos::FileSpec SearchFS("model.max");
						vos::File* pModelFile;

						if (ZIP.OpenFile(SearchFS, &pModelFile))
						{
							byte Buf[4];
							pModelFile->Read(Buf, 4);

							pModelFile->Close();
							delete pModelFile;
						}

					\endcode

	**************************************************************************/

	class ZipReader
	{
		public:
			// Constructor/destructor
			ZipReader(const vos::FileSpec& fsZipFile);
			~ZipReader();

			// Open/Close the ZIP file
			vos::Error Open();
			void Close();

			// Catalogue access routines
			uint GetNumFiles() const;
			void GetFilename(uint iFile, vos::FileSpec* pFileSpec);
			bool FindFile(const vos::FileSpec& fsFile, uint* piFile);

			// File access routines - once you open a file with these
			// functions, *you* own the new File object, and must 
			// dispose of it yourself.
			vos::Error OpenFile(uint iFile, vos::File** ppFile, uint32* pCRC32 = NULL);
			vos::Error OpenFile(const vos::FileSpec& fsFile, vos::File** ppFile, uint32* pCRC32 = NULL);

			//////////////////////////////////////////////////////////////////////

		private:
			// Load the ZIP catalogue and make a compact catalogue containing
			// the information we need.
			vos::Error ReadCatalogue();

			// Find the start of the ZIP's central directory.
			vos::Error FindCentralDirectory(uint32 cbFileSize, uint* pNumFiles);

			// Search for a 4-byte signature in a block of bytes.
			bool FindSignature(byte *pBlock, uint cbBlockSize, 
							   uint32 Signature, 
							   uint* piOffset);

			// Auxiliary function to open a file from a specific section of the
			// ZIP file (used by the other OpenFile() functions).
			vos::Error OpenFile(uint FileOffset, uint FileLength, vos::File** ppFile);

		private:
			// Name of the ZIP file
			vos::FileSpec m_fsZipFile;

			// ZIP file, if open
			vos::File* m_pZipFile;

			// The catalogue of files in the ZIP file
			ZipCatalogueReader* m_pCatalogue;
	};
}

#endif // !util_ZipReader_h
