//
// ZipWriter.cpp - a class to create ZIP files
//

#include "StdHdr.h"

#include "ZipWriter.h"
#include "ZipUtils.h"

#include "Msgs.h"
#include "Time.h"

namespace vos
{
	
	/**************************************************************************

	>	ZipWriter::ZipWriter()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Constructor

	**************************************************************************/

	ZipWriter::ZipWriter()
	{
		// TCILB
	}

	
	/**************************************************************************

	>	ZipWriter::~ZipWriter()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Destructor

	**************************************************************************/

	ZipWriter::~ZipWriter()
	{
		// Delete all our catalogue entries.
		CatalogueEntries::iterator it = m_Catalogue.begin();

		while (it != m_Catalogue.end())
		{
			delete *it;

			// do next entry
			++it;
		}
	}

	
	/**************************************************************************

	>	vos::Error ZipWriter::Create(const vos::FileSpec& fsNewZIPFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		fsNewZIPFile - name of the new ZIP file to create.
		Returns:	Standard error
						- Err_File_OpenFailed

		Purpose:	Create a new ZIP file from scratch.  The function will fail
					if the specified file already exists.

	**************************************************************************/

	vos::Error ZipWriter::Create(const vos::FileSpec& fsNewZIPFile)
	{
		if (fsNewZIPFile.GetFileType() != vos::FileType_NotFound)
		{
			// File already exists...
			return ERROR(vos::Err_File_OpenFailed, msgZIP_CreateZIPFailed);
		}

		// Ok, try to create the ZIP file.
		return m_ZipFile.Open(fsNewZIPFile, vos::fam_Write);
	}


	
	/**************************************************************************

	>	vos::Error ZipWriter::Open(const vos::FileSpec& fsExistingZIPFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		fsExistingZIPFile - name of the ZIP file to open.
		Returns:	Standard error
						- Err_File_OpenFailed

		Purpose:	Open an existing file, so that new files can be added.
					The function will fail if ZIP file does not exist.

					NB. Currently unimplemented!

	**************************************************************************/

	vos::Error ZipWriter::Open(const vos::FileSpec& fsExistingZIPFile)
	{
		COMPLAIN("UNIMPLEMENTED");
		return VNOERROR;
	}

	
	/**************************************************************************

	>	vos::Error ZipWriter::Close()

		Author:		Tim Browse
		Date:		10/06/2001
		Returns:	Standard error:

						Err_File_WriteFail - an error occurred when writing the ZIP 
											 catalogue out to the ZIP file.

		Purpose:	Close the ZIP file.  This must be done, as it updates the
					ZIP file's catalogue.

	**************************************************************************/

	vos::Error ZipWriter::Close()
	{
		if (m_ZipFile.IsOpen())
		{
			if (!WriteCatalogue())
			{
				// Could not write the ZIP catalogue - we close the ZIP
				// file anyway, to keep things tidy.
				m_ZipFile.Close();
				return LASTERROR;
			}

			m_ZipFile.Close();
		}

		// All done
		return VNOERROR;
	}

	/**************************************************************************

	>	vos::Error ZipWriter::WriteCatalogue()

		Author:		Tim Browse
		Date:		10/06/2001
		Returns:	Standard error:

						Err_File_WriteFail - an error occurred when writing the ZIP 
											 catalogue out to the ZIP file.

		Purpose:	Write the ZIP format catalogue out to the ZIP file.

	**************************************************************************/

	vos::Error ZipWriter::WriteCatalogue()
	{
		// We need to know where the central catalogue is, and how big it is.
		uint CatalogueStartPos;
		if (!m_ZipFile.GetPos(&CatalogueStartPos))
			return LASTERROR;

		// Write out the catalogue
		CatalogueEntries::iterator it = m_Catalogue.begin();

		while (it != m_Catalogue.end())
		{
			// Write this entry
			if (!(*it)->Write(&m_ZipFile))
			{
				return ERROR(vos::Err_File_WriteFail, msgFile_Write_Fail);
			}
			
			// Do the next one
			++it;
		}

		//
		// Fill in the end record...
		//

		// Only for multi-part zips, but some unzippers may barf if this is wrong.
		m_CatalogueEnd.m_NumCentralDirEntries = m_CatalogueEnd.m_TotalNumCentralDirEntries;

		// Fill in where the central directory starts
		m_CatalogueEnd.m_CentralDirOffset = CatalogueStartPos;

		// Find out how big the central dir will be
		uint EndRecordPos;
		if (!m_ZipFile.GetPos(&EndRecordPos))
		{
			return LASTERROR;
		}

		m_CatalogueEnd.m_CentralDirSize = EndRecordPos - CatalogueStartPos;

		// Ok, final record is correctly, filled in, so write it out (and we're done!)
		return m_CatalogueEnd.Write(&m_ZipFile);
	}

	/**************************************************************************

	>	vos::Error ZipWriter::CopyFileToZIP(const vos::FileSpec& fsFileToCopy, 
											uint cbFileSize, uint32* pCRC32)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		fsFileToCopy - the file to copy into the ZIP file.
					cbFileSize - size of this file, in bytes.
		Outputs:	pCRC32 - the CRC value for this file
		Returns:	Error
		Purpose:	Copy the specified file into the ZIP file we are making.

	**************************************************************************/

	vos::Error ZipWriter::CopyFileToZIP(const vos::FileSpec& fsFileToCopy, 
										uint cbFileSize, uint32* pCRC32)
	{
		// Ok, try to open the file
		vos::DiskFile m_DataFile;
		if (!m_DataFile.Open(fsFileToCopy, vos::fam_Read))
		{
			// Unable to open the file to read it!
			return LASTERROR;
		}

		// Initialise CRC value
		if (pCRC32 != NULL)
			*pCRC32 = 0;

		// Ok, allocate a copying buffer - we use 64k for now.
		const int kBufferSize = 64 * 1024;
		byte *pBuffer = new byte[kBufferSize];

		uint cbBlockSize = kBufferSize;
		uint cbBytesLeft = cbFileSize;

		while (cbBytesLeft > 0)
		{
			if (cbBlockSize > cbBytesLeft)
				cbBlockSize = cbBytesLeft;

			if (!m_DataFile.Read(pBuffer, cbBlockSize) ||
				!m_ZipFile.Write(pBuffer, cbBlockSize))
			{
				// Could not copy all the data - error!
				m_DataFile.Close();
				delete [] pBuffer;
				return LASTERROR;
			}

			// Update CRC value
			if (pCRC32 != NULL)
			{
				*pCRC32 = crc32(*pCRC32, pBuffer, cbBlockSize);
			}

			// Adjust bytes left to copy
			cbBytesLeft -= cbBlockSize;
		}

		// Copy done - clean up and return
		m_DataFile.Close();
		delete [] pBuffer;
		return VNOERROR;
	}


	
	/**************************************************************************

	>	vos::Error ZipWriter::AddFile(const vos::FileSpec& fsFileToAdd,
									  const vos::FileSpec& fsFilenameToUseInZIP)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		fsFileToAdd - the name of the disk file to add to the	
								  ZIP file.

					fsFilenameToUseInZIP 
								- the filename to use in the ZIP catalogue.
								  You will usually not want to use the full
								  pathname specified in fsFileToAdd, as it will
								  contain drive letters/volume names etc.
								  You may pass in a relative pathname, or just
								  a leafname.
											
		Returns:	Standard error

		Purpose:	Add a file to the ZIP file, using an existing disk file
					as the source.

	**************************************************************************/

	vos::Error ZipWriter::AddFile(const vos::FileSpec& fsFileToAdd,
								  const vos::FileSpec& fsFilenameToUseInZIP)
	{
		// Get information about the file.
		ZipLocalFileHeader ZLFH;
		
		// Set this as v1.0, MSDOS mode
		ZLFH.m_VersionNeededToExtract = 1 * 10;

		// Modification time/data for this file.
		vos::Time Timestamp;
		if (fsFileToAdd.GetFileTime(&Timestamp))
		{
			ConvertToDOSTimeStamp(Timestamp,
					 			  &ZLFH.m_ModifiedFileDate,
								  &ZLFH.m_ModifiedFileTime);
		}
		else
		{
			// Can't get timestamp - shouldn't really happen
			COMPLAIN("ZipWriter::AddFile() - can't get timestamp of file.");
			return LASTERROR;
		}

		// Size of the file, both compressed and uncompressed.
		ZLFH.m_UncompressedSize = fsFileToAdd.GetFileSize();
		ZLFH.m_CompressedSize = ZLFH.m_UncompressedSize;

		// Filename encoding - forward slashes are always used as folder
		// separators in ZIPs.
		vos::String strZIPFilename = fsFilenameToUseInZIP.GetString('/');
		ZLFH.SetFilename(strZIPFilename);

		// Ok, so now we write out the file header, and then copy the file data
		// into the ZIP file.
		//
		// The only twist is that we have to seek back to the header to fill in
		// the correct CRC value for the file.

		uint FileHeaderPos;
		if (!m_ZipFile.GetPos(&FileHeaderPos))
		{
			COMPLAIN("GetPos() failed while writing ZIP file!");
			return LASTERROR;
		}

		if (!ZLFH.Write(&m_ZipFile))
		{
			COMPLAIN("Failed to write LocalFileHeader to ZIP file.");
			return LASTERROR;
		}

		// Copy the file's data across
		if (!CopyFileToZIP(fsFileToAdd, ZLFH.m_UncompressedSize, &ZLFH.m_CRC32))
		{
			COMPLAIN("Failed to copy file's data to ZIP file.");
			return LASTERROR;
		}

		// OK, remember where we are
		uint EndOfFileData;
		if (!m_ZipFile.GetPos(&EndOfFileData))
		{
			COMPLAIN("GetPos() failed while writing ZIP file!");
			return LASTERROR;
		}

		// Right, seek back, write out the header with the new CRC value, 
		// and seek back to the end of the file's data, ready for the next
		// file to be added.
		if (!m_ZipFile.SeekFromStart(FileHeaderPos) ||
			!ZLFH.Write(&m_ZipFile) ||
			!m_ZipFile.SeekFromStart(EndOfFileData))
		{
			COMPLAIN("Failed to fix the CRC value in the file header!");
			return LASTERROR;
		}

		// Ok, so we've saved the file's header data and the file's data
		// to the ZIP file, so all we need to do is add an entry to the
		// ZIP catalogue, and we're done.
		ZipFileHeader *pZFH = new ZipFileHeader;
		pZFH->Init(ZLFH);
		pZFH->m_LocalHeaderRelativeOffset = FileHeaderPos;

		m_Catalogue.push_back(pZFH);

		// Update catalogue end marker
		m_CatalogueEnd.m_TotalNumCentralDirEntries++;

		// All done!
		return VNOERROR;
	}

	
	/**************************************************************************

	>	vos::Error ZipWriter::AddFile(const byte* pFileData, uint cbFileData, 
									  const vos::FileSpec& fsFilenameToUseInZIP)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFileData - the block of data to put in the file.
					cbFileData - size of the data in pFileData (in bytes)

					fsFilenameToUseInZIP 
								- the filename to use in the ZIP catalogue.
								  You may pass in a relative pathname, or just
								  a leafname.
											
		Returns:	Standard error

		Purpose:	Add a file to the ZIP file, using a block of memory data
					as the data to use for the file.

					NB. Currently UNIMPLEMENTED!  (Sorry!)

						(Should not be hard to do though)

	**************************************************************************/

	vos::Error ZipWriter::AddFile(const byte* pFileData, uint cbFileData, 
								  const vos::FileSpec& fsFilenameToUseInZIP)
	{
		COMPLAIN("UNIMPLEMENTED");
		return VNOERROR;
	}
}
