//
// ZipWriter.h - a class to create ZIP files
//

#ifndef inc_ZipWriter_h
#define inc_ZipWriter_h

#include "Errors.h"
#include "DiskFile.h"

#include "ZipLayout.h"

#include <vector>

namespace vos
{
	// Convenient name for STL type.
	typedef std::vector<ZipFileHeader*> CatalogueEntries;

	/**************************************************************************

	>	class ZipWriter

		Author:		Tim Browse

		Date:		10/06/2001

		Purpose:	Provide the facility to create ZIP files.

					You can either add new files to an existing ZIP file, or
					create a new ZIP file from scratch.

					At the moment, only uncompressed ZIP files are handled.

					For example:

					\code
						// Error checking removed for clarity!				

						ZipWriter MyZip;

						// MYFILE.ZIP will be created, and the files added to this ZIP file.
						MyZip.Create("myfile.zip");

						// Note that when adding files, you can control how they are named
						// in the resulting ZIP file.  For example, as seen below, you can 
						// put them in a different folder hierarchy, or rename the file,
						// or both.
						MyZip.AddFile("c:\\data\\mylist.txt", "mylist.txt");
						MyZip.AddFile("data1.dat",			  "data\\data1.dat");
						MyZip.AddFile("data2.dat",			  "data\\data2.dat");
						MyZip.AddFile("index2.dat",			  "data\\index.idx");

						// Add a file constructed in memory
						byte FileBuf[1024];
						memset(FileBuf, 42, 1024);
						MyZip.AddFile(FileBuf, 1024, "data\fortytwo.dat");

						// Close the ZIP file - you must do this, as it updates the
						// ZIP catalogue.
						MyZip.Close();

					\\endcode

	/************************************************************************/

	class ZipWriter
	{
		public:
			ZipWriter();
			~ZipWriter();

		public:
			// Create a new ZIP file, ready for adding files
			vos::Error Create(const vos::FileSpec& fsNewZIPFile);

			// Open an existing ZIP file, ready for adding files
			vos::Error Open(const vos::FileSpec& fsExistingZIPFile);

			// Close a ZIP file - should be done when you have finished adding files.
			vos::Error Close();

			//
			// Adding files to the ZIP...
			//

			// Add an existing disk file
			vos::Error AddFile(const vos::FileSpec& fsFileToAdd,
							   const vos::FileSpec& fsFilenameToUseInZIP);

			// Add a file using data from RAM.
			vos::Error AddFile(const byte* pFileData, uint cbFileData, 
							   const vos::FileSpec& fsFilenameToUseInZIP);

		private:
			// Add an entry to our soft catalogue
			void AddToCatalogue(const char* szFilename,
								uint32 HeaderOffset,
								uint cbFileSize,
								vos::Time TimeStamp,
								uint32 CRC);

			// Copy the file's data to the current position in the ZIP file.
			vos::Error CopyFileToZIP(const vos::FileSpec& fsFileToCopy, 
									 uint cbFileSize, uint32* pCRC32);

			// Write the central directory out to the ZIP file
			vos::Error WriteCatalogue();

		private:
			// The main ZIP file object
			vos::DiskFile m_ZipFile;

			// The zip catalogue entries.
			CatalogueEntries m_Catalogue;

			// The final ZIP catalogue entry
			vos::ZipEndOfCentralDir m_CatalogueEnd;
	};
}

#endif // !inc_ZipWriter_h
