//
// vos.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_h
#define vos_h


// Basic stuff needed by library

#include <stddef.h>			// For NULL etc.

#include <limits.h>			// For checking type sizes


#ifdef _MSC_VER
//
// Disable annoying MSVC warning
//
#pragma warning(disable:4786)
#endif // _MSC_VER


//
// Set vOS build type.
//

#ifdef WIN32

// Set basic vos build as Microsoft Windows.
#define VOS_MSWINDOWS	(1)

// The Windows flavour is Win32
#define VOS_WIN32		(1)

#elif (defined macintosh) | (defined __APPLE__)

// Set basic vos build as MacOS.
#define VOS_MACOS	(1)

// The Macintosh flavour is Carbon-compatible MacOS Classic.
#define VOS_CARBON		(1)

#endif


//
// Set up debug new
//

#if defined VOS_WIN32

#ifdef _DEBUG

// We have to include this here otherwise crtdbg.h causes problems.
#include <string>

#include <crtdbg.h>

#define vosDEBUG_NEW   new( _NORMAL_BLOCK, __FILE__, __LINE__)
#else
#define vosDEBUG_NEW

#endif // _DEBUG

#elif defined VOS_MACOS

#define vosDEBUG_NEW   new

#endif // defined VOS_WIN32


//
// NLS codes
//

#define VOS_LANG_EN (1)		// US English
#define VOS_LANG_UK (44)	// British English


// Set NLS to EN (US English) if not specified by make process
#ifndef VOS_LANG
#define VOS_LANG VOS_LANG_EN
#endif

// Asserts

#include "Assert.h"

	//////////////////////////////////////////////////////////////////////
	//
	//	Basic integer types
	//		
	//////////////////////////////////////////////////////////////////////

	/*!	Typedef: byte

		An unsigned 8 bit quantity.
	*/
		
	/*!	Typedef: uint

		An unsigned quantity that is the architecture's native word size.
	*/

	/*!	Typedef: uint8

		An unsigned 8 bit quantity.
	*/

	/*!	Typedef: uint16

		An unsigned 16 bit quantity.
	*/

	/*!	Typedef: uint32

		An unsigned 32 bit quantity.
	*/

	/*!	Typedef: uint64

		An unsigned 64 bit quantity.
	*/

	/*!	Typedef: int8

		A signed 8 bit quantity.
	*/

	/*!	Typedef: int16

		A signed 16 bit quantity.
	*/

	/*!	Typedef: int32

		A signed 32 bit quantity.
	*/

	/*!	Typedef: int64

		A signed 64 bit quantity.
	*/

	//////////////////////////////////////////////////////////////////////
	//
	//	Basic floating point types
	//		
	//////////////////////////////////////////////////////////////////////

	/*!	Typedef: float32

		A 32-bit floating point value
	*/
		
	/*!	Typedef: float64

		A 64-bit floating point value
	*/
		

#ifdef VOS_MSWINDOWS

	typedef unsigned char byte;

/* Define 8-bit types */
#if UCHAR_MAX == 0xff

	typedef unsigned char   uint8;
	typedef signed char      int8;

#else
#error This machine has no 8-bit type
#endif /* UCHAR_MAX == 0xff */

/* Define 16-bit types */
#if UINT_MAX == 0xffff

	typedef unsigned int    uint16;
	typedef int              int16;

#elif USHRT_MAX == 0xffff

	typedef unsigned short  uint16;
	typedef short            int16;

#else
#error This machine has no 16-bit type
#endif /* UINT_MAX == 0xffff */

/* Define 32-bit types */
#if UINT_MAX == 0xfffffffful

	typedef unsigned int    uint32;
	typedef int              int32;

#elif ULONG_MAX == 0xfffffffful
	typedef unsigned long   uint32;
	typedef long             int32;

#elif USHRT_MAX == 0xfffffffful

	typedef unsigned short  uint32;
	typedef short            int32;

#else
#error This machine has no 32-bit type
#endif /* UINT_MAX == 0xfffffffful */

	typedef unsigned int   uint;

	typedef unsigned __int64 uint64;
	typedef __int64 int64;


	typedef float	float32;
	typedef double	float64;

#elif (defined VOS_MACOS)

	typedef unsigned char byte;

	typedef unsigned int   uint;
	typedef unsigned char  uint8;
	typedef unsigned short uint16;
	typedef unsigned long  uint32;
	typedef unsigned long long uint64;

	typedef char  int8;
	typedef short int16;
	typedef long  int32;
	typedef long long int64;

	typedef float	float32;
	typedef double	float64;

#else

#error "Needs porting"

#endif



namespace vos
{
	//
	// Forward declarations of common vos objects.
	//
	class String;
	class StringParser;
	class Date;
	class File;
	class FileSpec;
	class FileInfo;
	class List;
	class ListItem;
	class Log;
	class LogFile;
	class Preferences;
	class Time;
	class TimeOfDay;
	class TreeNode;
	class ResourceID;
	class StringResourceID;

	// 
	// Basic integral types

	/*! Typedef: vchar

		Character type to be used in string objects (signed).
	*/

#ifdef VOS_MSWINDOWS

	typedef char vchar;

	#define VOS_LITTLE_ENDIAN (1)

#elif (defined VOS_MACOS)

	typedef char vchar;

	#define VOS_LITTLE_ENDIAN (0)

#else

#error "Needs porting"

#endif

	//
	// Bit twiddling
	//

	/*! Macro: BIT(n)

		\arg n the bit to set (0 is LSB)

		This macro gives an int value representing an int with the
		specified bit set, e.g.

		BIT(4) returns 16 (i.e. 10000 in binary)
	*/
#ifndef BIT
#define BIT(n) (1 << (n))
#endif

	/*! Macro: VOS_SETBIT(Value, Bit)

		\arg Value the value to change
		\arg Bit the bit to set (0 is LSB)

		Set a particular bit in the specified value.

		e.g.

		\code

			int Value = 96;			// Value is 1100000 in binary
			VOS_SETBIT(Value, 2);	// Value is now 100 (1100100)

		\endcode
	*/
#ifndef VOS_SETBIT
#define VOS_SETBIT(x, n) (x) |= BIT(n)
#endif

	/*! Macro: VOS_CLRBIT(Value, Bit)

		\arg Value the value to change
		\arg Bit the bit to clear (0 is LSB)

		Clear a particular bit in the specified value.

		e.g.

		\code

			int Value = 96;			// Value is 1100000 in binary
			VOS_CLRBIT(Value, 5);	// Value is now 64 (1000000)

		\endcode
	*/
#ifndef VOS_CLRBIT
#define VOS_CLRBIT(x, n) (x) &= (~BIT(n))
#endif

	//
	//
	// Abstract factory for vos components
	//

	// Implementation classes
	class PreferencesImp;

	/**************************************************************************
	
	>	class vosFactory
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Class factory for vOS components that use the Bridge
					pattern (it builds implementation objects).
	
	**************************************************************************/
	
	class vosFactory
	{
		public:
			// Virtual Constructors
			PreferencesImp		*MakePreferences();
	};

	vosFactory& VOSFactory();

};

namespace vos
{
	//
	//	File typedefs
	//
	
#ifdef VOS_MSWINDOWS		

	const char * const kFolderSeparator		= "\\";
	const char * const kThisFolder			= ".";
	const char * const kParentFolder		= "..";

#elif (defined VOS_MACOS)

	// Metrowerks hates const type const included in pch++ files!!!
	#define kFolderSeparator				":"
	#define kThisFolder						""	// So THIS + SEPARATOR = ":", Mac "This Dir"
	#define kParentFolder					":"	// So PARENT + SEPARATOR = "::", Mac "Parent Dir"

#else

	const char * const kFolderSeparator		= "/";
	const char * const kThisFolder			= ".";
	const char * const kParentFolder		= "..";
	
#endif
};

// 
// Some useful macros
//

#define VOS_SAFE_DELETE(pX)  { delete (pX); (pX) = NULL; }
#define VOS_INIT_AUTOVAR(Object) { memset(&Object, 0, sizeof(Object); }
#define VOS_INIT_SIZEDAUTOVAR(Object, SizeField) { memset(&Object, 0, sizeof(Object); Object.SizeField = sizeof(Object); }

// Include Win32 types that we use in headers
#ifdef VOS_MSWINDOWS
#include "Win32\Win32_Types.h"
#endif

#endif // !vos_h
