//
// Win32Error.cpp - Error decoding/reporting.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32Error.h"

#include <windows.h>

#include "..\Errors.h"
#include "..\ErrorCodes.h"
#include "..\File.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

//
// Linefeed/Carriage Return character codes
//
#define LF (10)
#define CR (13)


namespace vos
{
	/**************************************************************************
	
	>	String Win32Error::GetLastError()
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	The error message
	
		Purpose:	Get the textual error message for the last Win32
					error that occurred.

					Relies on the Win32 functions GetLastError() and
					FormatMessage().
	
	**************************************************************************/
	
	String Win32Error::GetLastError()
	{
		LPVOID lpMsgBuf;

		// Get the error message.
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					  FORMAT_MESSAGE_FROM_SYSTEM | 
					  FORMAT_MESSAGE_IGNORE_INSERTS,
					  NULL,
					  ::GetLastError(),
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					  (LPTSTR) &lpMsgBuf,
					  0,
					  NULL);

		// Create a String object.
		String ErrorMessage((const char *) lpMsgBuf);

		// Free the buffer.
		::LocalFree(lpMsgBuf);

		// Strip trailing CR/LF from end of message.
		ErrorMessage.RemoveTrailingChar(CR);
		ErrorMessage.RemoveTrailingChar(LF);
		ErrorMessage.RemoveTrailingChar(CR);

		// Return error message to caller
		return ErrorMessage;
	}

	/**************************************************************************
	
	>	Error Win32Error::TranslateError()
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	The synthesised vOS Error object
	
		Purpose:	Make a vOS error from the last Win32 error that occurred, 
					e.g.

					\code

						if (Win32Function() != 0)
						{
							return Win32Error::TranslateError();
						}
						else
						{
							return VNOERROR;
						}

					\endcode
	
	**************************************************************************/
	
	Error Win32Error::TranslateError()
	{
		// Try to determine specific error code...
		int ErrorCode = Err_GenericError;

		DWORD dwErr = ::GetLastError();

		switch (dwErr)
		{
			case ERROR_FILE_NOT_FOUND:
			case ERROR_PATH_NOT_FOUND:
				ErrorCode = Err_File_NotFound;
				break;

			case ERROR_ACCESS_DENIED:
				ErrorCode = Err_File_AccessDenied;
				break;

			case ERROR_FILE_EXISTS:
			case ERROR_ALREADY_EXISTS:
				ErrorCode = Err_File_AlreadyExists;
				break;

			case ERROR_CANNOT_MAKE:
				ErrorCode = Err_File_CreateFailed;
				break;
		}

		// Construct vOS error...
		String ErrMsg = Win32Error::GetLastError();

		return ERROR(ErrorCode, ErrMsg.c_str());
	}
};

