//
// Win32Registry.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32Registry.h"

#include "..\String.h"
#include "..\Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	// Maximum length for a single key or value name.
	const int kMaxNameSize = _MAX_PATH + 1;

	/**************************************************************************

	>	bool Win32Registry::ReadStringValue(HKEY hKey, 
											const String& ValueName,
											String* pValue)


		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hKey - the key to examine
					ValueName - name of the value under hKey to read
		Outputs:	pValue - the value of the key that was read from the
							 registry.
		Returns:	true => key found, and was a string

		Purpose:	Read a string from a registry key

	**************************************************************************/

	bool Win32Registry::ReadStringValue(HKEY hKey, 
										const String& ValueName,
										String* pValue)
	{
		if (hKey == NULL)
		{
			// Error - invalid key.
			vosERROR();
			return false;
		}

		// Find out how big the string is.
		DWORD cbData = 0;
		DWORD ValueType;
		LONG Err = ::RegQueryValueEx(hKey, ValueName.c_str(), 0, 
									 &ValueType, NULL, &cbData);

		if (Err == ERROR_SUCCESS)
		{
			// Is this the correct type, i.e. a string?
			if ((ValueType != REG_SZ) &&
				(ValueType != REG_EXPAND_SZ) &&
				(ValueType != REG_MULTI_SZ))
			{
				// Not a string
				return false;
			}

			// Ok, now we know how big the data is, how should be create
			// a buffer?
			const int kBufSize = 64;
			if (cbData < kBufSize)
			{
				// Static buffer is big enough, so use that
				BYTE StrData[kBufSize];

				// Read the string from the registry
				Err = ::RegQueryValueEx(hKey, ValueName.c_str(), 0, 
										&ValueType, StrData, &cbData);

				// if that worked, copy string value
				if (Err == ERROR_SUCCESS)
				{
					*pValue = (char *) StrData;
				}
				else
				{
					// Could not read value - complain in debug builds
					vosERROR();
				}
			}
			else
			{
				// Static buffer not big enough - allocate one on the heap
				BYTE *pStrData = new BYTE[cbData];

				// Read the string from the registry
				Err = ::RegQueryValueEx(hKey, ValueName.c_str(), 0, 
										&ValueType, pStrData, &cbData);

				// if that worked, copy string value
				if (Err == ERROR_SUCCESS)
				{
					*pValue = (char *) pStrData;
				}
				else
				{
					// Could not read value - complain in debug builds
					vosERROR();
				}

				// Clean up temporary buffer
				delete [] pStrData;
			}
		}
		else
		{
			// Failure is probably because preference value is not
			// there yet - in which case, we leave the caller's variable
			// alone so they can initialise it to a default value before
			// calling us.
		}

		return (Err == ERROR_SUCCESS);
	}

	/**************************************************************************
	
	>	bool Win32Registry::KeyExists(HKEY hParent, const String& KeyName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hParent - parent of the key to check for
					KeyName - name of the key to check for
		Returns:	true => key exists
					false => otherwise
	
		Purpose:	Find out if the specified key exists
	
	**************************************************************************/
	
	bool Win32Registry::KeyExists(HKEY hParent, const String& KeyName)
	{
		// Is it a key?
		HKEY hKey;
		if (::RegOpenKey(hParent, KeyName.c_str(), &hKey) == ERROR_SUCCESS)
		{
			// Yes, it's a key
			::RegCloseKey(hKey);

			return true;
		}

		// otherwise, it doesn't exist
		return false;
	}

	/**************************************************************************
	
	>	bool Win32Registry::ValueExists(HKEY hParent, const String& ValueName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hParent - parent of the value to check for
					ValueName - name of the value to look for
		Returns:	true => value exists
					false => otherwise
	
		Purpose:	Find out if the specified registry value exists.
	
	**************************************************************************/
	
	bool Win32Registry::ValueExists(HKEY hParent, const String& ValueName)
	{
		// Is it a value?
		return (::RegQueryValueEx(hParent, ValueName.c_str(), NULL, NULL, NULL, NULL) == ERROR_SUCCESS);
	}

	/**************************************************************************
	
	>	HKEY Win32Registry::CreateKey(HKEY hParent, const String& KeyName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hParent - where to create the key
					KeyName - name of the key to create.
		Returns:	Handle to the newly created key, or NULL if it did 
					not work (e.g. key already exists, permission denied).
	
		Purpose:	Create the named registry key
	
	**************************************************************************/
	
	HKEY Win32Registry::CreateKey(HKEY hParent, const String& KeyName)
	{
		// If it already exists, we can't create it
		if (KeyExists(hParent, KeyName))
		{
			// Cannot create key
			return NULL;
		}

		// The way is clear, so create the key.
		HKEY hKey;
		if (::RegCreateKey(hParent, KeyName.c_str(), &hKey) != ERROR_SUCCESS)
		{
			// Key creation failed
			return NULL;
		}

		// Key was created ok
		return hKey;
	}

	/**************************************************************************
	
	>	HKEY Win32Registry::OpenKey(HKEY hParent, const String& KeyName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hParent - parent of key to access
					KeyName - name of key to access
		Returns:	The handle of the opened key, or 
					NULL if it did not work (e.g. does not exist, permission
					denied etc)
	
		Purpose:	Open/access the named registry key.

					You should close they key with CloseKey() when you no
					longer need it.
	
	**************************************************************************/
	
	HKEY Win32Registry::OpenKey(HKEY hParent, const String& KeyName)
	{
		HKEY hKey;
		if (::RegOpenKey(hParent, KeyName.c_str(), &hKey) == ERROR_SUCCESS)
		{
			// Yes, found the key
			return hKey;
		}

		// No such key
		return NULL;
	}

	/**************************************************************************
	
	>	void Win32Registry::CloseKey(HKEY hKey)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hKey - handle of the key to close (as returned by OpenKey()).
	
		Purpose:	Close a registry key that was opened with OpenKey().
	
	**************************************************************************/
	
	void Win32Registry::CloseKey(HKEY hKey)
	{
		::RegCloseKey(hKey);
	}

	/**************************************************************************

	>	bool Win32Registry::CreateStringValue(HKEY hParent, 
											  const String& Name,
											  const String& Value)


		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hParent - the parent of the value to create
					Name - the name of the value to create
					Value - the string to place in the new registry value
		Returns:	true => value created successfully
					false => otherwise (e.g. value already exists)

		Purpose:	Create the named string value in the registry.

	**************************************************************************/

	bool Win32Registry::CreateStringValue(HKEY hParent, 
										  const String& Name,
										  const String& Value)
	{
		// If it already exists, we can't create it
		if (ValueExists(hParent, Name))
		{
			// Cannot create value
			return false;
		}

		return SetStringValue(hParent, Name, Value);
	}


	/**************************************************************************

	>	bool Win32Registry::SetStringValue(HKEY hParent, 
										   const String& Name,
										   const String& Value)


		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hParent - the parent of the value to set
					Name - the name of the value to set
					Value - the string to place in the registry value
		Returns:	true => registry value set ok
					false => otherwise (e.g. current value is not a string,
										permission denied, etc.)

		Purpose:	Set a registry value to hold a particular string.

	**************************************************************************/

	bool Win32Registry::SetStringValue(HKEY hParent, 
									   const String& Name,
									   const String& Value)
	{
		// Work out description of value data to satisfy registry API.
		const BYTE* pValueData = (const BYTE*) Value.c_str();
		DWORD cbData = Value.GetLength() + 1;

		// Write the specified string
		return (::RegSetValueEx(hParent, Name.c_str(), 0,
								REG_SZ, pValueData, cbData) == ERROR_SUCCESS);
	}

	/**************************************************************************
	
	>	bool Win32Registry::DeleteKey(HKEY hKey, const String& SubKeyName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hKey - the parent of the key to delete
					SubKeyName - name of the key to delete
		Returns:	false - always (function is disabled)
	
		Purpose:	Delete a key and all its children.
		
					NB. currently DISABLED as it is a bit of a scary 
						function to have lying around.  Use DeleteEmptyKey()
						instead.
	
	**************************************************************************/
	
	bool Win32Registry::DeleteKey(HKEY hKey, const String& SubKeyName)
	{
#if 1
		COMPLAIN("DeleteKey() not implemented - too scary!  Use DeleteEmptyKey() instead.");
		return false;
#else
		// Access the specified key
		HKEY hKeyToDelete;
		if (::RegOpenKey(hKey, SubKeyName.c_str(), &hKeyToDelete) != ERROR_SUCCESS)
		{
			// Can't find the key, so we can't delete it
			return false;
		}

		// Delete all its child keys
		char szChildName[kMaxNameSize];
		while (::RegEnumKey(hKeyToDelete, 0, szChildName, kMaxNameSize) == ERROR_SUCCESS)
		{
			if (!DeleteKey(hKeyToDelete, szChildName))
			{
				// Unable to delete this key - abort
				return false;
			}
		}

		// And now delete all its child values
		DWORD BufSize = kMaxNameSize;
		while (::RegEnumValue(hKeyToDelete, 0, szChildName, &BufSize, 0, NULL, NULL, NULL) == ERROR_SUCCESS)
		{
			if (::RegDeleteValue(hKeyToDelete, szChildName) != ERROR_SUCCESS)
			{
				// Unable to delete this child - abort
				return false;
			}
		}

		// Delete the key, now that we know it has no children
		return (::RegDeleteKey(hKey, SubKeyName.c_str()) == ERROR_SUCCESS);
#endif
	}

	/**************************************************************************
	
	>	bool Win32Registry::DeleteEmptyKey(HKEY hKey, const String& SubKeyName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hKey - the parent of the key to delete
					SubKeyName - name of the key to delete
		Returns:	true => key deleted ok
					false => otherwise (e.g. it has children)
	
		Purpose:	Delete a key (if it has no children).
	
	**************************************************************************/
	
	bool Win32Registry::DeleteEmptyKey(HKEY hKey, const String& SubKeyName)
	{
		// Access the specified key
		HKEY hKeyToDelete;
		if (::RegOpenKey(hKey, SubKeyName.c_str(), &hKeyToDelete) != ERROR_SUCCESS)
		{
			// Can't find the key, so we can't delete it
			return false;
		}

		// Find out if this key has any children
		if (KeyHasChildren(hKeyToDelete))
		{
			// Key has children - we can't delete it
			::RegCloseKey(hKeyToDelete);
			return false;
		}
		else		
		{
			// Delete the key, now that we know it has no children
			::RegCloseKey(hKeyToDelete);
			return (::RegDeleteKey(hKey, SubKeyName.c_str()) == ERROR_SUCCESS);
		}
	}

	/**************************************************************************

	>	void Win32Registry::SplitRegistryName(const String& ValueName, 
											  String* pParentKey, 
											  String* pLeafName)


		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		ValueName - the full pathname of the value to split
		Outputs:	pParentKey - pathname of the parent key
					pLeafName - the name of just the leaf node from the pathname.

		Purpose:	Split a registry pathname into a leafname and its
					parent.

					For example, split:

						"HKEY_CURRENT_USER\Software\Microsoft\Microsoft Office"

					into: 

							"HKEY_CURRENT_USER\Software\Microsoft"

						and

  							"Microsoft Office"


	**************************************************************************/

	void Win32Registry::SplitRegistryName(const String& ValueName, 
										  String* pParentKey, 
										  String* pLeafName)
	{
		// Search for a backslash
		unsigned int SlashPos;
		if (ValueName.ReverseFind('\\', &SlashPos) && (SlashPos > 0))
		{
			// Found a slash - separate the strings
			*pParentKey = ValueName.Left(SlashPos);
			*pLeafName = ValueName.Mid(SlashPos + 1);
		}
		else
		{
			// ValueName is just a leaf name
			pParentKey->Clear();
			*pLeafName = ValueName;
		}
	}

	/**************************************************************************
	
	>	bool Win32Registry::DeleteValue(HKEY hKey, const String& ValueName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hKey - parent key of the value to delete
					ValueName - name of the value to delete
		Returns:	true => value deleted ok
					false => otherwise (e.g. value does not exist, permission
										denied etc)
	
		Purpose:	Delete a specific registry value.
	
	**************************************************************************/
	
	bool Win32Registry::DeleteValue(HKEY hKey, const String& ValueName)
	{
		// Split valuename into parent path and leafname.
		String ParentKey, LeafName;
		SplitRegistryName(ValueName, &ParentKey, &LeafName);

		// If ValueName is a leafname, just delete it.
		if (ParentKey.IsEmpty())
		{
			return (::RegDeleteValue(hKey, ValueName.c_str()) == ERROR_SUCCESS);
		}

		// Otherwise, open the parent key so we can delete the value.
		HKEY hParent;
		if (::RegOpenKey(hKey, ParentKey.c_str(), &hParent) == ERROR_SUCCESS)
		{
			// Ok, let's delete the value and go
			bool bSuccess = (::RegDeleteValue(hParent, LeafName.c_str()) == ERROR_SUCCESS);
			::RegCloseKey(hParent);
			return bSuccess;
		}
		else
		{
			// Could not find parent key - give up
			return false;
		}
	}

	/**************************************************************************
	
	>	bool Win32Registry::KeyHasChildren(HKEY hKey, const String& SubKeyName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hKey - the parent of the key to check
					SubKeyName - name of the key to check
		Returns:	true => the named key has children
					false => otherwise
	
		Purpose:	Determine if a key has children or not.
	
	**************************************************************************/
	
	bool Win32Registry::KeyHasChildren(HKEY hKey, const String& SubKeyName)
	{
		// Open the key...
		HKEY hSubKey;
		if (::RegOpenKey(hKey, SubKeyName.c_str(), &hSubKey))
		{
			// Find out if it has children
			bool bHasChildren = KeyHasChildren(hSubKey);

			// Close key and return answer to caller
			::RegCloseKey(hSubKey);
			return bHasChildren;
		}
		else
		{
			// Unable to open key - assume this means it has no children
			return false;
		}
	}

	/**************************************************************************
	
	>	bool Win32Registry::KeyHasChildren(HKEY hKey)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		hKey - the handle of the key to check.
		Returns:	true => the key has children
					false => otherwise
	
		Purpose:	Determine if a key has children or not.
	
	**************************************************************************/
	
	bool Win32Registry::KeyHasChildren(HKEY hKey)
	{
		DWORD cSubKeys;
		DWORD cValues;

		// We're only interested in the number of subkeys and values in this key...
		if (::RegQueryInfoKey(hKey, NULL, NULL, NULL, 
							  &cSubKeys, NULL, NULL, 
							  &cValues, NULL, NULL, 
							  NULL, NULL) == ERROR_SUCCESS)
		{
			// Tell caller if the key has children of some sort
			return ((cSubKeys + cValues) > 0);
		}
		else
		{
			// Unable to find key - assume this means it has no children
			return false;
		}
	}

	/**************************************************************************

		bool Win32Registry::SplitRegistryString(const String& RegistryKey,
												HKEY* phKey, String* pKeyName)

		Author:		Tim Browse
		Date:		12/10/99
		Inputs:		RegistryKey - full name of a registry key.
		Outputs:	phKey - one of the six main registry keys.
					pKeyName - path to the registry key starting from phKey.

		Purpose:	Split a registry key name into a system HKEY and a 
					key pathname.

					For example:

						HKEY_CURRENT_USER\Software\Microsoft\Microsoft Office

					would be split into:

						HKEY_CURRENT_USER

						and 

						"Software\Microsoft\Microsoft Office"

	**************************************************************************/

	bool Win32Registry::SplitRegistryString(const String& RegistryKey,
											HKEY* phKey, String* pKeyName)
	{
		if (RegistryKey.StartsWith("HKEY_CLASSES_ROOT\\"))
		{
			// Decompose string accordingly
			*phKey = HKEY_CLASSES_ROOT;
			*pKeyName = RegistryKey.Mid(strlen("HKEY_CLASSES_ROOT\\"));
			return true;
		}
		else if (RegistryKey.StartsWith("HKEY_CURRENT_USER\\"))
		{
			// Decompose string accordingly
			*phKey = HKEY_CURRENT_USER;
			*pKeyName = RegistryKey.Mid(strlen("HKEY_CURRENT_USER\\"));
			return true;
		}
		else if (RegistryKey.StartsWith("HKEY_USERS\\"))
		{
			// Decompose string accordingly
			*phKey = HKEY_USERS;
			*pKeyName = RegistryKey.Mid(strlen("HKEY_USERS\\"));
			return true;
		}
		else if (RegistryKey.StartsWith("HKEY_CURRENT_CONFIG\\"))
		{
			// Decompose string accordingly
			*phKey = HKEY_CURRENT_CONFIG;
			*pKeyName = RegistryKey.Mid(strlen("HKEY_CURRENT_CONFIG\\"));
			return true;
		}
		else if (RegistryKey.StartsWith("HKEY_LOCAL_MACHINE\\"))
		{
			// Decompose string accordingly
			*phKey = HKEY_LOCAL_MACHINE;
			*pKeyName = RegistryKey.Mid(strlen("HKEY_LOCAL_MACHINE\\"));
			return true;
		}
		else if (RegistryKey.StartsWith("HKEY_DYN_DATA\\"))
		{
			// Decompose string accordingly
			*phKey = HKEY_DYN_DATA;
			*pKeyName = RegistryKey.Mid(strlen("HKEY_DYN_DATA\\"));
			return true;
		}
		else
		{
			// Unrecognised system key
			return false;
		}
	}
};

