//
// Win32RegistryKey.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32RegistryKey.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/**************************************************************************

	>	Win32RegistryKey::Win32RegistryKey()

		Author:		Tim Browse
		Date:		29/11/99

		Purpose:	Create a null key object (the object is not attached
					to any registry key - you must use Open() or Create()
					to do so).

	**************************************************************************/

	Win32RegistryKey::Win32RegistryKey() : m_hKey(NULL), m_hParentKey(NULL)
	{
	}

	/**************************************************************************

	>	Win32RegistryKey::Win32RegistryKey(HKEY hParentKey,
										   const String& KeyName)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		hParentKey - HKEY of the parent to look for the key in.
					KeyName - name of the child key to open

		Purpose:	Open a registry key given the above inputs.

	**************************************************************************/

	Win32RegistryKey::Win32RegistryKey(HKEY hParentKey,
									   const String& KeyName) : m_hParentKey(hParentKey),
															    m_KeyName(KeyName),
																m_hKey(NULL)
	{
		if (::RegOpenKey(m_hParentKey, m_KeyName.c_str(), &m_hKey) != ERROR_SUCCESS)
		{
			// Failed to open key - make sure m_hKey is still NULL
			m_hKey = NULL;
		}
	}

	/**************************************************************************

	>	Win32RegistryKey::Win32RegistryKey(const Win32RegistryKey& ParentKey,
										   const String& KeyName)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		hParentKey - HKEY of the parent to look for the key in.
					KeyName - name of the child key to open

		Purpose:	Open a registry key given the above inputs.

	**************************************************************************/

	Win32RegistryKey::Win32RegistryKey(const Win32RegistryKey& ParentKey,
									   const String& KeyName) : m_hParentKey(ParentKey.GetHKEY()),
															    m_KeyName(KeyName),
																m_hKey(NULL)
	{
	}


	/**************************************************************************

	>	Win32RegistryKey::~Win32RegistryKey()

		Author:		Tim Browse
		Date:		29/11/99

		Purpose:	Clean up the key object (any open registry key is closed).

	**************************************************************************/

	Win32RegistryKey::~Win32RegistryKey()
	{
		CleanUp();
	}


	/**************************************************************************

	>	void Win32RegistryKey::CleanUp()

		Author:		Tim Browse
		Date:		29/11/99

		Purpose:	Clean up the key object (any open registry key is closed).

	**************************************************************************/

	void Win32RegistryKey::CleanUp()
	{
		if (m_hKey != NULL)
		{
			// Clean up open key.
			::RegCloseKey(m_hKey);
			m_hKey = NULL;
		}
	}

	/**************************************************************************

	>	const Win32RegistryKey& Win32RegistryKey::operator=(const Win32RegistryKey& rhs)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		rhs - the object to copy

		Purpose:	Copy the registry key object into this one.
					(Both objects end up with their own HKEY handle to the key)

	**************************************************************************/

	const Win32RegistryKey& Win32RegistryKey::operator=(const Win32RegistryKey& rhs)
	{
		// Ignore self-assignment
		if (this != &rhs)
		{
			// Free up our own key
			CleanUp();

			// Try to open key of rhs
			Open(rhs.m_hParentKey, rhs.m_KeyName);
		}

		// All done
		return *this;
	}

	/**************************************************************************

	>	Win32RegistryKey::Win32RegistryKey(const Win32RegistryKey& rhs)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		rhs - the object to copy

		Purpose:	Copy constructor
					(Both objects end up with their own HKEY handle to the key)

	**************************************************************************/

	Win32RegistryKey::Win32RegistryKey(const Win32RegistryKey& rhs) : m_hKey(NULL), m_hParentKey(NULL)
	{
		// Try to open key of rhs
		Open(rhs.m_hParentKey, rhs.m_KeyName);
	}

	/**************************************************************************

	>	bool Win32RegistryKey::Open(HKEY hParentKey, const String& KeyName)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		hParentKey - HKEY of the parent to look for the key in.
					KeyName - name of the child key to open
		Returns:	true => the key was opened ok
					false => key does not exist or could not be opened.

		Purpose:	Open a registry key given the above inputs.

	**************************************************************************/

	bool Win32RegistryKey::Open(HKEY hParentKey, const String& KeyName)
	{
		// Free any open key we might have
		CleanUp();

		// Remember parent/name info
		m_hParentKey = hParentKey;
		m_KeyName = KeyName;

		if (::RegOpenKey(m_hParentKey, m_KeyName.c_str(), &m_hKey) != ERROR_SUCCESS)
		{
			// Failed to open key - make sure m_hKey is still NULL
			m_hKey = NULL;
			return false;
		}

		// Worked ok
		return true;
	}

	/**************************************************************************

	>	bool Win32RegistryKey::Open(const Win32RegistryKey& ParentKey, 
									const String& KeyName)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		ParentKey - the parent registry key to look in
					KeyName - name of the child key to open
		Returns:	true => the key was opened ok
					false => key does not exist or could not be opened.

		Purpose:	Open a registry key given the above inputs.

	**************************************************************************/

	bool Win32RegistryKey::Open(const Win32RegistryKey& ParentKey, 
								const String& KeyName)
	{
		return Open(ParentKey.GetHKEY(), KeyName);
	}

	/**************************************************************************

	>	bool Win32RegistryKey::Create(HKEY hParentKey, const String& KeyName)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		hParentKey - HKEY of the parent to look for the key in.
					KeyName - name of the child key to open
		Returns:	true => the key was created ok, or it already existed and
							was opened.
					false => key could not be created or found.

		Purpose:	Create a registry key given the above inputs.
					If the key already exists, it is just opened.

	**************************************************************************/

	bool Win32RegistryKey::Create(HKEY hParentKey, const String& KeyName)
	{
		// Free any open key we might have
		CleanUp();

		// Remember parent/name info
		m_hParentKey = hParentKey;
		m_KeyName = KeyName;

		if (::RegCreateKey(m_hParentKey, m_KeyName, &m_hKey) != ERROR_SUCCESS)
		{
			// Failed to create key - make sure m_hKey is still NULL
			m_hKey = NULL;
			return false;
		}

		// Worked ok
		return true;
	}

	/**************************************************************************

	>	bool Win32RegistryKey::Create(const Win32RegistryKey& ParentKey, 
									  const String& KeyName)

		Author:		Tim Browse
		Date:		29/11/99
		Inputs:		ParentKey - the parent registry key to look in
					KeyName - name of the child key to open
		Returns:	true => the key was created ok, or it already existed and
							was opened.
					false => key could not be created or found.

		Purpose:	Create a registry key given the above inputs.
					If the key already exists, it is just opened.

	**************************************************************************/

	bool Win32RegistryKey::Create(const Win32RegistryKey& ParentKey, 
								  const String& KeyName)
	{
		return Create(ParentKey.GetHKEY(), KeyName);
	}

	/**************************************************************************

	>	bool Win32RegistryKey::IsValid() const

		Author:		Tim Browse
		Date:		29/11/99
		Returns:	true => key is valid and open
					false => key is null

		Purpose:	Find out if the object is linked to a valid registry key,
					or if it is a null object.

	**************************************************************************/

	bool Win32RegistryKey::IsValid() const
	{
		return (m_hKey != NULL);
	}

	/**************************************************************************

	>	HKEY Win32RegistryKey::GetHKEY() const

		Author:		Tim Browse
		Date:		29/11/99
		Returns:	The HKEY of this object, or NULL if it is not linked to
					a valid registry key.

		Purpose:	Access the HKEY of this object.

	**************************************************************************/

	HKEY Win32RegistryKey::GetHKEY() const
	{
		return m_hKey;
	}

	/**************************************************************************

	>	HKEY Win32RegistryKey::DetachHKEY()

		Author:		Tim Browse
		Date:		29/11/99
		Returns:	The HKEY of this object, or NULL if it is not linked to
					a valid registry key.

		Purpose:	Detach the HKEY of this object and return to the caller.

					The caller then has the responsibility to call 
					::RegCloseKey() on this key.

					After this, this object then acts as a null key, until
					Open() or Create() is called successfully again.

	**************************************************************************/

	HKEY Win32RegistryKey::DetachHKEY()
	{
		// Detach HKEy from this object
		HKEY hKey = m_hKey;
		m_hKey = NULL;

		// Give the HKEY handle to the caller, who now owns it.
		return hKey;
	}

};

