//
// Win32_PreferencesImp.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_PreferencesImp.h"

#include "..\Preferences.h"

#include "Win32Registry.h"

#include "..\Errors.h"

#include "..\FileSpec.h"

namespace vos
{
	/**************************************************************************
	
	>	Win32_PreferencesImp::Win32_PreferencesImp()
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	Win32_PreferencesImp::Win32_PreferencesImp() : m_hKey(NULL)
	{
	}

	/**************************************************************************
	
	>	Win32_PreferencesImp::~Win32_PreferencesImp()
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	Win32_PreferencesImp::~Win32_PreferencesImp()
	{
		// Clean up registry key
		CloseRegKey();
	}

	/**************************************************************************
	
	>	void Win32_PreferencesImp::CloseRegKey()
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Close the registry associated with this preferences object.
	
	**************************************************************************/
	
	void Win32_PreferencesImp::CloseRegKey()
	{
		// Clean up registry key
		if (m_hKey != NULL)
		{
			// We don't need this key any more
			LONG Err = ::RegCloseKey(m_hKey);

			// Complain in debug builds if closing the key causes an error
			ASSERT(Err == ERROR_SUCCESS);

			// Lose key handle
			m_hKey = NULL;
		}
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::IsRoot()
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	true => this object is the root preferences object
					false => otherwise
	
		Purpose:	Check to see if this is the root preferences object.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::IsRoot()
	{
		return (m_pPrefs->GetParent() == NULL);
	}

	/**************************************************************************
	
	>	HKEY Win32_PreferencesImp::GetKey()
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	HKEY for this object (or NULL if failed)
	
		Purpose:	Get the HKEY associated with this preferences object.
	
	**************************************************************************/
	
	HKEY Win32_PreferencesImp::GetKey()
	{
		// If we already have a legal key, use it
		if (m_hKey != NULL)
		{
			return m_hKey;
		}

		// If this is the root preferences object, then we need
		// to create the root object.
		if (IsRoot())
		{
			String AuthorName = Preferences::GetAuthorName();
			String ProgramName = Preferences::GetProgramName();

			// We must at least have a program name
			if (ProgramName.GetLength() == 0)
			{
				COMPLAIN("You must call Preferences::Init() before using the Preferences module.");
				return NULL;
			}

			// Ok, open the key from the registry.
			String RootKeyName = "Software\\";

			// AuthorName (Company) is optional.
			if (AuthorName.GetLength() > 0)
			{
				RootKeyName += AuthorName + "\\";
			}

			RootKeyName += ProgramName + "\\";
			RootKeyName += m_pPrefs->GetSectionName();

			if (::RegCreateKey(HKEY_CURRENT_USER, RootKeyName.c_str(), &m_hKey) != ERROR_SUCCESS)
			{
				// Key creation failed - critical error
				m_hKey = NULL;
				vosERROR();
			}
		}
		else
		{
			// Find HKEY of parent section
			HKEY hParentKey = GetParentImp()->GetKey();

			if (hParentKey == NULL)
			{
				// Cannot find parent key
				vosERROR();
			}
			else
			{
				if (::RegCreateKey(hParentKey, m_pPrefs->GetSectionName().c_str(), &m_hKey) != ERROR_SUCCESS)
				{
					// Key creation failed - critical error
					m_hKey = NULL;
					vosERROR();
				}
			}
		}

		// Return whatever key we created (or NULL if it failed)
		return m_hKey;
	}

	/**************************************************************************
	
	>	Win32_PreferencesImp* Win32_PreferencesImp::GetParentImp()
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	Parent implementation object.
	
		Purpose:	Get parent implementation object of this object.
	
	**************************************************************************/
	
	Win32_PreferencesImp* Win32_PreferencesImp::GetParentImp()
	{
		Preferences* pParentPrefs = (Preferences*) m_pPrefs->GetParent();

		// If no parent, then don't dereference it.
		if (pParentPrefs == NULL)
		{
			return NULL;
		}

		Win32_PreferencesImp* pParentImp = (Win32_PreferencesImp*) pParentPrefs->GetImp();

		return pParentImp;
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Read(const String& PrefName, String* pPrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
		Outputs:	pPrefValue - contents of the preference.
		Returns:	true => preference read ok
					false => otherwise
	
		Purpose:	Read a String preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Read(const String& PrefName, String* pPrefValue)
	{
		// Get the HKEY for this object
		HKEY hKey = GetKey();

		if (hKey == NULL)
		{
			// Error - can't find HKEY.
			vosERROR();
			return false;
		}

		return Win32Registry::ReadStringValue(hKey, PrefName, pPrefValue);
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Read(const String& PrefName, int* pPrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
		Outputs:	pPrefValue - contents of the preference.
		Returns:	true => preference read ok
					false => otherwise
	
		Purpose:	Read an integer preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Read(const String& PrefName, int* pPrefValue)
	{
		// Get the HKEY for this object
		HKEY hKey = GetKey();

		if (hKey == NULL)
		{
			// Error - can't find HKEY.
			vosERROR();
			return false;
		}

		// Work out description of value data to satisfy registry API.
		DWORD dwValue;
		BYTE* pValueData = (BYTE*) &dwValue;
		DWORD cbData = sizeof(DWORD);

		DWORD ValueType;

		// Read the specified value
		LONG Err = ::RegQueryValueEx(hKey, PrefName.c_str(), 0, 
									 &ValueType, pValueData, &cbData);

		// If we read a value of the correct type, then copy the
		// result into caller's variable.
		if (Err == ERROR_SUCCESS)
		{
			if ((ValueType == REG_DWORD) && (cbData == sizeof(DWORD)))
			{
				// Correct data type/size - transfer value.
				*pPrefValue = (int) dwValue;
			}
			else
			{
				// Data type mismatch - complain in debug builds
				vosERROR();
			}
		}
		else
		{
			// Failure is probably because preference value is not
			// there yet - in which case, we leave the caller's variable
			// alone so they can initialise it to a default value before
			// calling us.
		}

		return (Err == ERROR_SUCCESS);
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Read(const String& PrefName, bool*   pPrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
		Outputs:	pPrefValue - contents of the preference.
		Returns:	true => preference read ok
					false => otherwise
	
		Purpose:	Read a boolean preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Read(const String& PrefName, bool*   pPrefValue)
	{
		// Defer to more general int preference routine
		int Boolean;
		bool bSuccess = Read(PrefName, &Boolean);

		// If it worked, copy the bool value read from preferences into the
		// caller's variable.
		if (bSuccess)
		{
			*pPrefValue = (Boolean != 0);
		}

		return bSuccess;
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Read(const String& PrefName, FileSpec* pPrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
		Outputs:	pPrefValue - contents of the preference.
		Returns:	true => preference read ok
					false => otherwise
	
		Purpose:	Read a FileSpec preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Read(const String& PrefName, FileSpec* pPrefValue)
	{
		// Defer to more general string preference routine
		vos::String strPrefValue;
		bool bSuccess = Read(PrefName, &strPrefValue);

		// If it worked, copy the string value read from preferences into the
		// caller's variable.
		if (bSuccess)
		{
			pPrefValue->SetFromString(strPrefValue);
		}

		return bSuccess;
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Write(const String& PrefName, const String& PrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
					PrefValue - contents of the preference.
		Returns:	true => preference written ok
					false => otherwise
	
		Purpose:	Write a String preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Write(const String& PrefName, const String& PrefValue)
	{
		// Get the HKEY for this object
		HKEY hKey = GetKey();

		if (hKey == NULL)
		{
			// Error - can't find HKEY.
			vosERROR();
			return false;
		}

		// Work out description of value data to satisfy registry API.
		const BYTE* pValueData = (const BYTE*) PrefValue.c_str();
		DWORD cbData = PrefValue.GetLength() + 1;

		// Write the specified string
		LONG Err = ::RegSetValueEx(hKey, PrefName.c_str(), 
								   0, // Reserved
								   REG_SZ, pValueData, cbData);

		return (Err == ERROR_SUCCESS);
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Write(const String& PrefName, int PrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
					PrefValue - contents of the preference.
		Returns:	true => preference written ok
					false => otherwise
	
		Purpose:	Write an integer preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Write(const String& PrefName, int PrefValue)
	{
		// Get the HKEY for this object
		HKEY hKey = GetKey();

		if (hKey == NULL)
		{
			// Error - can't find HKEY.
			vosERROR();
			return false;
		}

		// Work out description of value data to satisfy registry API.
		const BYTE* pValueData = (const BYTE*) &PrefValue;
		DWORD cbData = sizeof(DWORD);

		// Write the specified string
		LONG Err = ::RegSetValueEx(hKey, PrefName.c_str(), 
								   0, // Reserved
								   REG_DWORD, pValueData, cbData);

		return (Err == ERROR_SUCCESS);
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Write(const String& PrefName, bool PrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
					PrefValue - contents of the preference.
		Returns:	true => preference written ok
					false => otherwise
	
		Purpose:	Write a boolean preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Write(const String& PrefName, bool PrefValue)
	{
		return Write(PrefName, (int) PrefValue);
	}

	/**************************************************************************
	
	>	bool Win32_PreferencesImp::Write(const String& PrefName, 
										 const FileSpec& PrefValue)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		PrefName - name of the preference
					PrefValue - contents of the preference.
		Returns:	true => preference written ok
					false => otherwise
	
		Purpose:	Write a FileSpec preference.
	
	**************************************************************************/
	
	bool Win32_PreferencesImp::Write(const String& PrefName, 
									 const FileSpec& PrefValue)
	{
		// Convert FileSpec to flat string, and store that
		String strPrefValue = PrefValue.GetString();

		return Write(PrefName, strPrefValue);
	}

	/**************************************************************************
	
	>	void Win32_PreferencesImp::DeleteAll()
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Clean up preferences objects when application exits.
	
	**************************************************************************/
	
	void Win32_PreferencesImp::DeleteAll()
	{
		// Make sure we've been initialised.
		if (!Preferences::IsInitialised())
		{
			COMPLAIN("Preferences system not initialised!");
			return;
		}

		if (IsRoot())
		{
			// Delete all the preferences
			String AuthorName = Preferences::GetAuthorName();
			String ProgramName = Preferences::GetProgramName();

			// We must at least have a program name
			if (ProgramName.GetLength() == 0)
			{
				COMPLAIN("You must call Preferences::Init() before using the Preferences module.");
				return;
			}

			// Ok, open the key from the registry.
			String RootKeyName = "Software\\";

			// AuthorName (Company) is optional.
			if (AuthorName.GetLength() > 0)
			{
				RootKeyName += AuthorName + "\\";
			}

			RootKeyName += ProgramName;
			if (m_pPrefs->GetSectionName().GetLength() > 0)
			{
				RootKeyName += "\\" + m_pPrefs->GetSectionName();
			}

			if (Win32Registry::DeleteKey(HKEY_CURRENT_USER, RootKeyName))
			{
				// Deleted the program key and all its children - delete the
				// author key if we have one, and if it's empty.
				if (AuthorName.GetLength() > 0)
				{
					RootKeyName = "Software\\" + AuthorName;
					Win32Registry::DeleteEmptyKey(HKEY_CURRENT_USER, RootKeyName);
				}
			}
		}
	}

};

