//
// Win32_VolumeSpec.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "..\vos.h"

#include "Win32_Headers.h"
#include "Win32Error.h"

#include "..\VolumeSpec.h"
#include "..\Errors.h"
#include "..\StringParser.h"
#include "..\FileSpec.h"
#include "..\File.h"


//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	//
	// Win32 support
	//
	// We have to dynamically check for the GetDiskFreeSpaceEx() function,
	// as it might not always be available (e.g. Windows 95, prior to OSR2).
	typedef BOOL WINAPI GetDiskFreeSpaceExFn(LPCTSTR,
											 PULARGE_INTEGER,
											 PULARGE_INTEGER,
											 PULARGE_INTEGER);

	// Default to NULL pointer until we locate the routine.
	static GetDiskFreeSpaceExFn *pGetDiskFreeSpaceEx = NULL;
	
	// By default, we don't know that it's not available (we only know
	// when we've checked).
	static bool bGetDiskFreeSpaceExNotAvailable = false;


	/**************************************************************************
	
	>	static bool FindGetDiskFreeSpaceEx()
	
		Author:		Tim Browse
		Date:		18/4/2002
		Returns:	true => GetDiskFreeSpaceEx() is available
					false => otherwise
	
		Purpose:	Find out if the GetDiskFreeSpaceEx() function is available
					on the host machine (it is not available before Win95
					OSR2).
	
	**************************************************************************/
	
	static bool FindGetDiskFreeSpaceEx()
	{
		//
		// Ok, we need to know if we can use GetDiskFreeSpaceEx() or not - it
		// doesn't exist on Windows 95 prior to OSR2.
		//

		// Do we already have the function?
		if (pGetDiskFreeSpaceEx != NULL)
		{
			// Yes, so we can use it.
			return true;
		}

		// Have we already tried and failed?
		if (bGetDiskFreeSpaceExNotAvailable)
		{
			// Yes, so don't try again
			return false;
		}

		//
		// Ok, first time we're trying, so try to find the function.
		//
		// (We don't need to load Kernel32.dll - it is always loaded for Win32 apps)
		//
		HMODULE hKernel32 = ::GetModuleHandle("kernel32.dll");

		pGetDiskFreeSpaceEx = (GetDiskFreeSpaceExFn *) ::GetProcAddress(hKernel32, "GetDiskFreeSpaceExA");

		// Did it work?
		if (pGetDiskFreeSpaceEx != NULL)
		{
			// yes, we found the function
			return true;
		}

		// Otherwise, we've tried and failed, so we know it doesn't
		// exist...
		bGetDiskFreeSpaceExNotAvailable = true;
		return false;
	}

	/**************************************************************************

	>	bool VolumeSpec::DecodeVolumeName(const String& Pathname)

		Author:		Tim Browse
		Date:		21/11/99
		Inputs:		Pathname - the pathname string to extract the volume name
							   from.
		Returns:	true => valid volume name was extracted
					false => valid volume name could not be found.

		Purpose:	Given the pathname, extract the volume name and store it 
					in this object.

	**************************************************************************/

	bool VolumeSpec::DecodeVolumeName(const String& Pathname)
	{
		//
		// Extract the volume name, which under Win32 can either be a drive 
		// letter followed by a colon, or a UNC mount point, e.g.
		//
		//            C:
		//
		//			  \\server\foldername
		//
		String Volume = Pathname;
		Volume.Trim();
		if (Volume.Left(2) == "\\\\")
		{
			// We have a UNC...extract it
			StringParser UNC(Volume);
			UNC.SkipDelims("\\");

			// Try to read the server name
			String ServerName;
			if (UNC.ReadString(ServerName, "\\"))
			{
				// Read the server name ok, so skip the backslash
				UNC.SkipDelim("\\");

				// Ok, now try to read the mount point
				String MountPoint;
				if (UNC.ReadString(MountPoint, "\\"))
				{
					// Read the whole UNC volume correctly - build the string.
					m_VolumeName = "\\\\";
					m_VolumeName += ServerName;
					m_VolumeName += '\\';
					m_VolumeName += MountPoint;
					m_VolumeName += '\\';

					// All done
					return true;
				}
			}
		}
		else
		{
			// Check for a drive letter.
			if (isalpha(Volume[0]) && (Volume[1] == ':'))
			{
				// It's a drive letter volume spec, so copy it, and add
				// the standard folder separator.
				m_VolumeName = Volume.Left(2);
				m_VolumeName += '\\';

				// All done
				return true;
			}
			else
			{
				// Don't recognise the volume spec; do nothing
			}
		}

		// Unable to decode volume name successfully
		return false;
	}

	/**************************************************************************

	>	VolumeSpec::VolumeSpec(const String& Pathname)

		Author:		Tim Browse
		Date:		21/11/99
		Inputs:		Pathname - the pathname containing the volume name to 
							   initialise this object with,

		Purpose:	Initialise a volume spec object from a pathname.

	**************************************************************************/

	VolumeSpec::VolumeSpec(const String& Pathname)
	{
		DecodeVolumeName(Pathname);
	}

	/**************************************************************************

	>	VolumeSpec::VolumeSpec(const FileSpec& aFileSpec)

		Author:		Tim Browse
		Date:		21/11/99
		Inputs:		aFileSpec - the object to extract the volume spec from.

		Purpose:	Initialise a VolumeSpec object from the specified FileSpec.
					Only the volume name is stored in this object; not the 
					whole FileSpec.

	**************************************************************************/

	VolumeSpec::VolumeSpec(const FileSpec& aFileSpec)
	{
		DecodeVolumeName(aFileSpec.GetString());
	}

	/**************************************************************************

	>	bool VolumeSpec::operator==(const VolumeSpec& rhs) const

		Author:		Tim Browse
		Date:		21/11/99

		Purpose:	Standard equality operator.

					NB. Copes with the different case-dependencies of
					    different OSs - i.e. Unix derivatives care about
						the case of filenames; Windows, MacOS etc. don't.

	**************************************************************************/

	bool VolumeSpec::operator==(const VolumeSpec& rhs) const
	{
		// Other OS's ignore the case of filenames
		return (m_VolumeName.CompareText(rhs.m_VolumeName));
	}

	/**************************************************************************

	>	bool VolumeSpec::operator!=(const VolumeSpec& rhs) const

		Author:		Tim Browse
		Date:		21/11/99

		Purpose:	Standard inequality operator.

					NB. Copes with the different case-dependencies of
					    different OSs - i.e. Unix derivatives care about
						the case of filenames; Windows, MacOS etc. don't.

	**************************************************************************/

	bool VolumeSpec::operator!=(const VolumeSpec& rhs) const
	{
		return !(*this == rhs);
	}

	/**************************************************************************

	>	bool VolumeSpec::IsEmpty() const

		Author:		Tim Browse
		Date:		21/11/99
		Returns:	true => VolumeSpec object does not contain a volume name.
					false => VolumeSpec has a valid volume name.

		Purpose:	Find out if th object specifies a valid volume name.
					In this scope, valid means non-null - this function will
					always return false if a non-null name is held in the
					object, even if the volume does not actually exist.

	**************************************************************************/

	bool VolumeSpec::IsEmpty() const
	{
		return m_VolumeName.IsEmpty();
	}

	/**************************************************************************

	>	String VolumeSpec::GetString() const

		Author:		Tim Browse
		Date:		21/11/99

		Purpose:	Get a string representation of this volume spec object.

	**************************************************************************/

	String VolumeSpec::GetString() const
	{
		return m_VolumeName;
	}

	/**************************************************************************

	>	void VolumeSpec::SetFromString(const String& Filename)

		Author:		Tim Browse
		Date:		21/11/99
		Inputs:		Filename - the filename to extract the volume name from.
							   This can be just a volume name or a full
							   pathname - either way, just the volume name
							   will be stored.

		Purpose:	Set the VolumeSpec object to hold the volume name
					specified in the Filename passed in.

	**************************************************************************/

	void VolumeSpec::SetFromString(const String& Filename)
	{
		DecodeVolumeName(Filename);
	}

	/**************************************************************************

	>	bool VolumeSpec::IsUNC() const

		Author:		Tim Browse
		Date:		21/11/99
		Returns:	true => volume specifies a Windows UNC mount point.
					false => null volume, or it's a drive letter volume.

		Purpose:	Find out if the VolumeSpec object describes a UNC-based
					pathname.

	**************************************************************************/

	bool VolumeSpec::IsUNC() const
	{
		return (m_VolumeName.Left(2) == "\\\\");
	}

	/**************************************************************************

	>	Error VolumeSpec::GetWin32FreeDiskSpace(uint *pBytesFree,
												uint *pClusterSize)

		Author:		Tim Browse
		Date:		21/11/99
		Returns:	Error

		Purpose:	Use the standard GetDiskFreeSpace() function from the 
					Win32 API to calculate the free space and/or cluster size
					of the volume specified by this object.

					NB.

					On versions of Windows 95 prior to OSR2, this function
					automatically temporarily maps UNC paths to a network
					drive so that GetFreeDiskSpace() will return the
					correct answer (see API documentation for the 
					GetDiskFreeSpace() function for details).

					Or at least it would, if I could find out how to do it.

	**************************************************************************/

	Error VolumeSpec::GetWin32FreeDiskSpace(uint *pBytesFree,
											uint *pClusterSize)
	{
		//
		// BADBAD - should use 64 bit math to avoid overflow...
		//

		if (IsEmpty())
		{
			// Volume name is null - no correct answer exists...
			return ERROR(Err_File_NotFound, 
						 "VolumeSpec::GetWin32FreeDiskSpace - Volume has not been initialised");
		}

		//
		// Check to see if GetDiskFreeSpaceEx() is available on this
		// platform - we don't want to use it; we just want to know if it
		// exists, because if it doesn't, then GetDiskFreeSpace() won't work
		// on UNC paths, so must temporarily map them to a network drive.
		//
		bool bMapNetworkDrive = IsUNC() && !FindGetDiskFreeSpaceEx();

		if (bMapNetworkDrive)
		{
			// Map the network drive...
			// BADBAD: Fix this!
			COMPLAIN("Mapping of network drive not implemented!");
		}

		// Call the Win32 function to get the disk free space/cluster size.
		DWORD dwSectorsPerCluster,
			  dwBytesPerSector,
			  dwNumberOfFreeClusters,
			  dwTotalNumberOfClusters;

		BOOL bSuccess = ::GetDiskFreeSpace(m_VolumeName.c_str(),
										   &dwSectorsPerCluster,
										   &dwBytesPerSector,
										   &dwNumberOfFreeClusters,
										   &dwTotalNumberOfClusters);

		if (bMapNetworkDrive)
		{
			// Unmap the temporary network drive.
			COMPLAIN("Unmapping of network drive not implemented!");
		}
 
		// Did the API call work?
		if (bSuccess)
		{
			// Yes, so calculate the size of each cluster.
			DWORD dwClusterSize = dwBytesPerSector * dwSectorsPerCluster;

			// Work out the free disk space according to how many
			// clusters are free
			DWORD dwBytesFree = dwNumberOfFreeClusters * dwClusterSize;

			// It all worked - return answer(s) to caller that they asked for
			if (pBytesFree != NULL)
			{
				*pBytesFree = (uint) dwBytesFree;
			}

			if (pClusterSize != NULL)
			{
				*pClusterSize = (uint) dwClusterSize;
			}

			return VNOERROR;
		}
		else
		{
			// Error getting free space info; report to caller.
			return Win32Error::TranslateError();
		}
	}

	/**************************************************************************

	>	Error VolumeSpec::GetFreeSpace(uint *pBytesFree)

		Author:		Tim Browse
		Date:		21/11/99
		Ouptuts:	The number of bytes free on this

		Purpose:	Find out how many bytes of free disk space there are on
					the disk volume specified by this object.

					NB. On OSs that implement per-user disk quotas, this
						figure may be less than the actual number of free
						bytes available on the volume.

	**************************************************************************/

	Error VolumeSpec::GetFreeSpace(uint *pBytesFree)
	{
		COMPLAIN_IF(pBytesFree == NULL, "Illegal NULL param");

		if (IsEmpty())
		{
			// Volume name is null - no correct answer exists...
			return ERROR(Err_File_NotFound, 
						 "VolumeSpec::GetFreeSpace - Volume has not been initialised");
		}

		// Otherwise, find out how much space is free...

		if (FindGetDiskFreeSpaceEx())
		{
			//
			// It's ok to use GetDiskFreeSpaceEx...
			//
			ULARGE_INTEGER BytesFree;
			ULARGE_INTEGER TotalBytes;

			if ((*pGetDiskFreeSpaceEx)(m_VolumeName.c_str(), &BytesFree, &TotalBytes, NULL))
			{
				// Got the free space ok...
				if (BytesFree.HighPart != 0)
				{
					// More than (2^32)-1 bytes free - just return (2^32)-1
					*pBytesFree = UINT_MAX;
				}
				else
				{
					// Bytes free can be represented in 32 bit number - so
					// just pass it back to the caller.
					*pBytesFree = BytesFree.LowPart;
				}

				// All done
				return VNOERROR;
			}
			else
			{
				// Error getting free space info; report to caller.
				return Win32Error::TranslateError();
			}
		}
		else
		{
			// We have to use GetDiskFreeSpace instead, as GetDiskFreeSpaceEx
			// is unavailable.
			uint BytesFree;
			if (GetWin32FreeDiskSpace(&BytesFree, NULL))
			{
				// It worked - return answer to caller
				*pBytesFree = BytesFree;
				return VNOERROR;
			}
			else
			{
				// Error getting free space info; report to caller.
				return Win32Error::TranslateError();
			}
		}

		// We should never get here...
		COMPLAIN("Unexpected condition met in VolumeSpec::GetFreeSpace().");
	}

	/**************************************************************************

	>	Error VolumeSpec::GetAllocationUnitSize(uint *pClusterSize)

		Author:		Tim Browse
		Date:		21/11/99
		Outputs:	pClusterSize - the size of clusters on this volume, in bytes.
		Returns:	Error

		Purpose:	Find out how large the allocation units (clusters) are on
					the disk volume specified by this object.

	**************************************************************************/

	Error VolumeSpec::GetAllocationUnitSize(uint *pClusterSize)
	{
		if (IsEmpty())
		{
			// Volume name is null - no correct answer exists...
			return ERROR(Err_File_NotFound, 
						 "VolumeSpec::GetAllocationUnitSize - Volume has not been initialised");
		}

		// We have to use GetDiskFreeSpace, because GetDiskFreeSpaceEx does not
		// tell us the cluster size.
		// NB. According to the documentation, the cluster size may be
		//     reported incorrectly for large drives, so this is just
		//     an estimate.
		uint ClusterSize;
		if (GetWin32FreeDiskSpace(NULL, &ClusterSize))
		{
			// It worked - return answer to caller
			*pClusterSize = ClusterSize;
			return VNOERROR;
		}
		else
		{
			// Error getting free space info; report to caller.
			return Win32Error::TranslateError();
		}
	}

	/**************************************************************************
	
	>	void VolumeSpec::GetDisplayName(String* pDisplayName)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Outputs:	pDisplayName - the user friendly display name.
	
		Purpose:	Get the name of the volume in a format that is appropriate
					to show to the user.
	
	**************************************************************************/
	
	void VolumeSpec::GetDisplayName(String* pDisplayName)
	{
		if (IsEmpty())
		{
			// Volume name is null
			pDisplayName->Clear();
		}

		// Return volume name, but remove trailing backslash.
		*pDisplayName = m_VolumeName;
		if (pDisplayName->Right(1) == "\\")
		{
			pDisplayName->Truncate(pDisplayName->GetLength() - 1);
		}
	}

};
